Package: R.devices
Version: 2.17.4
Depends: R (>= 2.14.0)
Imports: grDevices, R.methodsS3 (>= 1.8.1), R.oo (>= 1.24.0), R.utils
        (>= 2.10.1), base64enc (>= 0.1-2)
Suggests: digest (>= 0.6.13), Cairo (>= 1.5-9), R.rsp
VignetteBuilder: R.rsp
Authors@R: c(person("Henrik", "Bengtsson", role=c("aut", "cre", "cph"),
                                          email = "henrikb@braju.com"))
Author: Henrik Bengtsson [aut, cre, cph]
Maintainer: Henrik Bengtsson <henrikb@braju.com>
Title: Unified Handling of Graphics Devices
Description: Functions for creating plots and image files in a unified way
    regardless of output format (EPS, PDF, PNG, SVG, TIFF, WMF, etc.). Default
    device options as well as scales and aspect ratios are controlled in a uniform
    way across all device types. Switching output format requires minimal changes
    in code. This package is ideal for large-scale batch processing, because it
    will never leave open graphics devices or incomplete image files behind, even on
    errors or user interrupts.
License: LGPL (>= 2.1)
URL: https://henrikbengtsson.github.io/R.devices/,
        https://github.com/HenrikBengtsson/R.devices
BugReports: https://github.com/HenrikBengtsson/R.devices/issues
LazyLoad: TRUE
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-07 19:52:16 UTC; hb
Repository: CRAN
Date/Publication: 2026-05-09 14:42:11 UTC
Built: R 4.6.0; ; 2026-05-09 17:42:55 UTC; unix

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# IMPORTS
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
importFrom("R.methodsS3", "setMethodS3")
importFrom("R.methodsS3", "appendVarArgs")
importFrom("R.oo", "setConstructorS3")
importFrom("R.oo", "extend")
importFrom("R.oo", "Package")
importFrom("R.oo", "startupMessage")
importFrom("R.oo", "throw")

## Import Class:es
importFrom("R.oo", "BasicObject")
importFrom("R.utils", "Arguments")

## importFrom("R.oo", "[[.BasicObject")

## Importing functions
importFrom("R.oo", "trim")
importFrom("R.utils", "capitalize")
importFrom("R.utils", "getOption") ## Multi-sources: R.utils, base
importFrom("R.utils", "setOption")
importFrom("R.utils", "getAbsolutePath")
importFrom("R.utils", "getRelativePath")
importFrom("R.utils", "hpaste")
importFrom("R.utils", "isFile")
importFrom("R.utils", "isPackageInstalled")
importFrom("R.utils", "isPackageLoaded")
importFrom("base64enc", "dataURI")

importFrom("graphics", "par")
importFrom("utils", "browseURL")

# Import from grDevices
import("grDevices")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# EXPORTS
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
export("R.devices")

export("withPar")

export("capabilitiesX11")
export("capturePlot")

export("devDump")
export("devEval")
export("devIsOpen")
export("devIsInteractive")
export("devList")
export("devGetLabel")
export("devSetLabel")
export("devSet")
export("devOff")
export("devDone")
export("devNew")
export("devOptions")
export("getDevOption")
export("eps")
export("favicon")
export("jpeg2")
export("png2")
export("nulldev")

export("toBMP")
export("toEPS")
export("toEMF")
export("toPDF")
export("toPNG")
export("toSVG")
export("toTIFF")
export("toWMF")
##export("toCairoPNG")
export("toFavicon")
export("toNullDev")

export("toDefault")
export("toQuartz")
export("toWindows")
export("toX11")
export("toCairoWin")
export("toCairoX11")
##export("toJavaGD")
export("toRStudioGD")

export("asDataURI")

export("DevEvalProduct")
export("DevEvalFileProduct")

export("suppressGraphics")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# DECLARATIONS
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# DevEvalFileProduct
S3method("as.character", "DevEvalFileProduct")
S3method("getData", "DevEvalFileProduct")
S3method("getDataURI", "DevEvalFileProduct")
S3method("getExtension", "DevEvalFileProduct")
S3method("getFilename", "DevEvalFileProduct")
S3method("getFullname", "DevEvalFileProduct")
S3method("getMime", "DevEvalFileProduct")
S3method("getMimeType", "DevEvalFileProduct")
S3method("getPath", "DevEvalFileProduct")
S3method("getPathname", "DevEvalFileProduct")
S3method("view", "DevEvalFileProduct")

# DevEvalProduct
S3method("$", "DevEvalProduct")
S3method("[[", "DevEvalProduct")
S3method("as.character", "DevEvalProduct")
S3method("getFullname", "DevEvalProduct")
S3method("getName", "DevEvalProduct")
S3method("getTags", "DevEvalProduct")
S3method("getType", "DevEvalProduct")
S3method("view", "DevEvalProduct")
S3method("!", "DevEvalProduct")


S3method("architecture", "recordedplot")
S3method("architecture", "RecordedPlot")
S3method("as.architecture", "recordedplot")
S3method("as.architecture", "RecordedPlot")
export("architecture")
export("as.architecture")
S3method("print", "RecordedPlot")

<div id="badges"><!-- pkgdown markup -->
<a href="https://CRAN.R-project.org/web/checks/check_results_R.devices.html"><img border="0" src="https://www.r-pkg.org/badges/version/R.devices" alt="CRAN check status"/></a> <a href="https://github.com/HenrikBengtsson/R.devices/actions?query=workflow%3AR-CMD-check"><img border="0" src="https://github.com/HenrikBengtsson/R.devices/actions/workflows/R-CMD-check.yaml/badge.svg?branch=develop" alt="R CMD check status"/></a>     <a href="https://app.codecov.io/gh/HenrikBengtsson/R.devices"><img border="0" src="https://codecov.io/gh/HenrikBengtsson/R.devices/branch/develop/graph/badge.svg" alt="Coverage Status"/></a> 
</div>

# R.devices: Unified Handling of Graphics Devices 


## Installation
R package R.devices is available on [CRAN](https://cran.r-project.org/package=R.devices) and can be installed in R as:
```r
install.packages("R.devices")
```


### Pre-release version

To install the pre-release version that is available in Git branch `develop` on GitHub, use:
```r
remotes::install_github("HenrikBengtsson/R.devices", ref="develop")
```
This will install the package from source.  

<!-- pkgdown-drop-below -->

library(R.devices)

args <- commandArgs(trailingOnly = TRUE)
engine <- args[1]

g <- capturePlot({plot(1:10)})
system <- attr(g, "system")
tags <- sprintf("%s=%s", names(system), system)
pathname <- sprintf("inst/exdata/capturePlot,engine=%s,%s.rds",
                    engine, paste(tags, collapse=","))
cat(sprintf("generating %s ...\n", pathname))
saveRDS(g, file=pathname)

###########################################################################
### This 'tangle' R script was created from an RSP document.
### RSP source document: 'R.devices-overview.tex.rsp'
### Metadata 'title': 'R.devices overview'
### Metadata 'author': 'Henrik Bengtsson'
### Metadata 'engine': 'R.rsp::rsp'
### Metadata 'keywords': 'devices, graphics, plots, figures'
###########################################################################

t0 <- Sys.time()
library("R.devices")
R.rsp <- R.oo::Package("R.rsp")
withCapture <- R.utils::withCapture
hpaste <- R.utils::hpaste

for (type in c("png", "cairo_png", "CairoPNG", "png2")) {
  if (type %in% rownames(devOptions())) {
    devOptions(type, width=840)
    devOptions(type, field="fullname") # Better for LaTeX
  }
}

options(width=85)
options(digits=3)
options(str=strOptions(strict.width="cut"))
R.rsp$version
R.rsp$author
format(as.Date(R.devices$date), format="%B %d, %Y")
envir <- getNamespace("R.devices")
tos <- ls(pattern="^to[A-Z][a-zA-Z]+", envir=envir)
tos <- sprintf("%s()", tos)
hpaste(sprintf("\\code{%s}", tos), maxHead=Inf, lastCollapse=", and ")
toPDF("MyGaussianDensity", aspectRatio=0.6, {
   curve(dnorm, from=-5, to=+5)
  })
devOptions("pdf", reset=TRUE)
withCapture({
str(devOptions("pdf"))
})
withCapture({
devOptions("pdf", width=5, bg="lightblue")
})
withCapture({
devOptions("pdf", reset=TRUE)
})
withCapture({
devOptions()[,c("width", "height", "bg", "fg", "pointsize")]
})
toLatex(sessionInfo())

message("*** devSet() ...")

library("R.devices")

set.seed(0xBEEF)

devList0 <- devList()

# Open several devices in random order
idxs <- sample(2:20, size=5L, replace=FALSE)
for (idx in idxs) {
  idxT <- devSet(idx)
  # Sanity check
  stopifnot(idxT == idx)
}

# Close all devices
devOff(idxs)

# Sanity check
stopifnot(all.equal(devList(), devList0))


keys <- list(
  A = list(a=1, b=2, c=3),
  B = list(a=1, b=2, c=4),
  C = list(a=1, b=2),
  D = c(a=1, b=2, c=3)
)

# Open several devices using key objects
idxs <- NULL
for (key in keys) {
  idx <- devSet(key)
  idxs <- c(idxs, idx)
}

# Close all devices
devOff(idxs)

# Sanity check
stopifnot(all.equal(devList(), devList0))


# Open several devices using labels
labels <- c("A", "B", "C", "D")

for (rev in c(FALSE, TRUE)) {
  for (label in labels) devSet(label)

  if (rev) labels <- rev(labels)
  for (label in labels) {
    devOff(label)
    left <- devList()
    stopifnot(!is.element(label, names(left)))
  }
}

# Sanity check
stopifnot(all.equal(devList(), devList0))


message("*** devSet() ... DONE")

message("*** devListIndexOf() ...")

library("R.devices")

.devListIndexOf <- R.devices:::.devListIndexOf

labels <- list(character(0L), "Device 1", c("Device 1", "Device 1"), c("Device 1", "FooBar"))
for (label in labels) {
  res <- .devListIndexOf(label, error=FALSE)
  print(res)
  stopifnot(is.integer(res))
  stopifnot(is.character(names(res)))
}

message("*** devListIndexOf() ... DONE")

message("*** devEval() ...")

library("R.devices")
library("R.utils")
graphics.off()


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Various types of single and multiple device outputs
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
message("*** devEval() - single and multiple device outputs ...")

types <- list(
  character(0L),
  "{png}",
  "{jpg}",
  "nulldev",
  c("{png}", "{png}", "{jpeg}"),
  "{png},nulldev,pdf"
)

for (type in types) {
  cat("Device types: ", paste(sQuote(type), collapse=", "), "\n", sep="")
  devList0 <- devList()
  res <- devEval(type, name="multi", aspectRatio=2/3, {
    plot(1:10)
  })
  print(res)
  stopifnot(length(res) == length(unlist(strsplit(type, split=","))))
  stopifnot(all.equal(devList(), devList0))
}

# Sanity checks
print(devList())
stopifnot(length(devList()) == 0L)

message("*** devEval() - single and multiple device outputs ... DONE")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# With 'initially' and 'finally' expression
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
message("*** devEval() - initially and finally ...")

devList0 <- devList()
devEval(c("{png}", "{jpg}"), name="count", {
  plot(1:10)
  count <- count + 1L
}, initially = {
  # Emulate an overhead
  cat("Initiate...")
  count <- 0L
  Sys.sleep(1)
  cat("done\n")
}, finally = {
  cat("Number of image files created: ", count, "\n", sep="")
})
stopifnot(all.equal(devList(), devList0))

# Sanity checks
print(devList())
stopifnot(length(devList()) == 0L)

message("*** devEval() - initially and finally ... DONE")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Try several devices until first successful device is found
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
message("*** devEval() - first successful device ...")

types <- list(
  "png|jpg|pdf",                 # PNG, JPG, or PDF
  "dummy|png|jpg|pdf",           # "Non-existing", PNG, JPG, or PDF
  "quartz|x11|windows",          # Any interactive device (depending on OS)
  c("{png}|jpg", "x11|windows"), # PNG or JPG and then x11 or windows
  "eps|postscript|pdf",          # EPS, Postscript or PDF
  "jpeg2|jpeg",                  # JPEG via bitmap() or via jpeg()
  "{png},jpg|x11|windows",       # == c("{png}", "jpg|x11|windows")
  "nulldev|jpeg",                # NULL devices, otherwise jpeg
  "{png}"                        # Any PNG device
)

if (!capabilitiesX11()) {
  message("Skipping test for X11")
  types <- lapply(types, FUN=function(x) gsub("x11|", "", x, fixed=TRUE))
}

devList0 <- devList()

for (type in types) {
  printf("Any of %s\n", paste(sQuote(type), collapse=" + "))

  # Use try-catch in case not supported on some test systems
  tryCatch({
    res <- devEval(type, name="any", aspectRatio=2/3, scale=1.2, {
      plot(100:1)
    })
    printf("Result: %s (%s)\n\n", sQuote(res), attr(res, "type"))

    if (length(devList()) > 0) devOff()
  }, error = function(ex) {
    printf("Failed: %s\n\n", sQuote(ex$message))
  })
} # for (type ...)

# Sanity check
stopifnot(all.equal(devList(), devList0))

message("*** devEval() - first successful device ... DONE")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Plot a parsed expression
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
message("*** devEval() - parsed expressions ...")

expr <- substitute(plot(1:10))
tryCatch({
  res <- devEval("png|jpg|pdf", name="any", width=480L, height=480L, {
    plot(100:1)
  })
  printf("Result: %s (%s)\n\n", sQuote(res), attr(res, "type"))

  if (length(devList()) > 0) devOff()
}, error = function(ex) {
  printf("Failed: %s\n\n", sQuote(ex$message))
})

message("*** devEval() - parsed expressions ... DONE")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Special cases
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
message("*** toDefault(<expr>) ...")

# toX11({ plot(1:10) }) actually results in a call to
# devEval(type="x11", name={ plot(1:10) }); note argument 'name'
# and not 'expr'.  The following tests that devEval() recognizes
# and handles this internally.

## FIXME: The current solution evaluates 'name' internally
## and therefore opens a interactive graphics device.
if (interactive()) {
   res <- toDefault({ plot(1:10) })
   print(res)

   ## FIX ME:
   graphics.off()
}

message("*** toDefault(<expr>) ... DONE")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Device type specified as a device functions
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
message("*** devEval(<fcn>) ...")

types <- list(
  png     = "grDevices::png",
  jpg     = "grDevices::jpeg",
  nulldev = "R.devices::nulldev"
)
types <- types[names(types) %in% rownames(devOptions())]
types <- lapply(types, FUN = function(code) eval(parse(text = code)))

for (name in names(types)) {
  cat("Device types: ", paste(sQuote(name), collapse=", "), "\n", sep="")
  type <- types[[name]]
  str(args(type))
  devList0 <- devList()
  res <- devEval(type, name="multi", tags="function", aspectRatio=2/3, {
    plot(1:10)
  })
  print(res)
  stopifnot(length(res) == length(type))
  stopifnot(all.equal(devList(), devList0))
}

message("*** devEval(<fcn>) ... DONE")


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Special case: Default device
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
message("*** devEval(<default>) ...")

cat("Device types: 'default'\n")
type <- getOption("device")
str(type)
devList0 <- devList()
res <- devEval(type, name="default", aspectRatio=2/3, {
  plot(1:10)
})
print(res)
wasInteractiveOpened <- (length(setdiff(devList(), devList0)) > 0L)
if (wasInteractiveOpened) devOff()

message("*** devEval(<default>) ... DONE")



# Sanity checks
print(devList())
stopifnot(length(devList()) == 0L)

message("*** devEval() ... DONE")

message("*** dataURI ...")

R.devices::devOptions("*", field="dataURI")
uri <- R.devices::toPNG("foo", tags=c("a", "b"), aspectRatio=0.7, {
  plot(1:10)
})
str(uri)


R.devices::devOptions("*", field="pathname")
png <- R.devices::toPNG("foo", tags=c("a", "b"), aspectRatio=0.7, {
  plot(1:10)
}, force=FALSE)
uri2 <- R.devices::asDataURI(png)
str(uri2)
stopifnot(identical(uri2, uri))

message("*** dataURI ... DONE")

message("*** devOptions() ...")

known <- rownames(R.devices::devOptions())

# Without attaching package
opts0 <- R.devices::devOptions()
print(opts0)

if ("eps" %in% known) {
  opts0eps <- R.devices::devOptions("eps")
  str(opts0eps)
}

if ("png" %in% known) {
  opts <- R.devices::devOptions("png")
  print(opts)
}

# With attaching package
library("R.devices")
opts1 <- R.devices::devOptions()
print(opts1)

if ("eps" %in% known) {
  opts1eps <- R.devices::devOptions("eps")
  str(opts1eps)
  stopifnot(identical(opts1eps, opts0eps))
  stopifnot(identical(opts1, opts0))
}

# Options for the PNG device
if ("png" %in% known) {
  opts <- devOptions("png")
  print(opts)
}

# Options for the postscript device
if ("postscript" %in% known) {
  opts <- devOptions("postscript")
  print(opts)

  # Same using alias
  opts2 <- devOptions("ps")
  print(opts2)
  stopifnot(identical(opts2, opts))
}

# Options for all known devices
opts <- devOptions()
print(opts)

# Setting a custom option
if ("png" %in% known) {
  devOptions("png", foo=list(a=1, b=pi))
  str(devOptions("png")$foo)

  # Setting option to NULL, i.e. drop it
  devOptions("png", foo=NULL)
  str(devOptions("png")$foo)
  str(devOptions("png"))

  # Get individual device options
  print(getDevOption("png", "width"))
}

opts1 <- R.devices::devOptions()
print(opts1)

## Assert devOptions(<function>) equals devOptions(<name>)
## Identify all "primitive" functions
fcns <- lapply(R.devices:::devAll(), FUN=`[`, 1L)
for (name in names(fcns)) {
  fcn <- fcns[[name]]
  cat(sprintf("Asserting devOptions('%s') == devOptions(%s) ...\n", name, fcn))
  fcn <- eval(parse(text=fcn))
  optsName <- devOptions(name)
  optsFcn <- devOptions(fcn)
  res <- all.equal(optsFcn, optsName)
  if (!isTRUE(res)) {
    str(list(name=name, fcn=fcn, byName=optsName, byFcn=optsFcn))
    print(res)
    stop(!all.equal(optsFcn, optsName))
  }
}

message("*** devOptions() for each type ...")
types <- rownames(devOptions())
cat("All known types:\n")
print(types)

for (type in types) {
  message(sprintf("*** devOptions('%s') for each type ...", type))
  opts <- devOptions(type)
  str(opts)
  message(sprintf("*** devOptions('%s') for each type ... DONE", type))
}

message("*** devOptions() for each type ... DONE")


message("*** devOptions(drop=FALSE) ...")

if ("png" %in% known) {
  opts <- devOptions("png", drop=TRUE)
  str(opts)
  stopifnot(is.list(opts))
  stopifnot(is.null(dim(opts)))

  opts <- devOptions("png", drop=FALSE)
  str(opts)
  stopifnot(is.list(opts))
  stopifnot(!is.null(dim(opts)))
}

message("*** devOptions(drop=FALSE) ... DONE")

message("*** devOptions(reset=TRUE) ...")

## Reset all
opts <- devOptions()
print(opts)
opts0 <- devOptions(reset=TRUE)
print(opts0)

## Reset one device
if ("png" %in% known) {
  opts <- devOptions("png")
  width <- getDevOption("png", "width")
  devOptions("png", width=2*width)
  stopifnot(getDevOption("png", "width") == 2*width)
  devOptions("png", reset=TRUE)
  stopifnot(getDevOption("png", "width") == width)

  ## Reset "*"
  opts <- devOptions("*")
  path <- getDevOption("*", "path")
  devOptions("*", path="foo")
  stopifnot(getDevOption("*", "path") == "foo")
  devOptions("*", reset=TRUE)
  stopifnot(getDevOption("*", "path") == path)
}


message("*** devOptions(reset=TRUE) ... DONE")


message("*** devOptions() - errors ...")

res <- try(devOptions(type=character(0L), width=32L))
stopifnot(inherits(res, "try-error"))

message("*** devOptions() - errors ... DONE")


message("*** devOptions() - odds'n'ends ...")

devOptions(type=character(0L), reset=FALSE)
devOptions(type=character(0L), reset=TRUE)

message("*** devOptions() - odds'n'ends ... DONE")

message("*** devOptions() ... DONE")

message("*** favicon() ...")

library("R.devices")

message("- toFavicon() ...")

dataURI <- toFavicon({
  plot(NA, xlim=c(0,1), ylim=c(0,6), axes=FALSE, xaxs="i", yaxs="i")
  col <- rev(c("#FF0000", "#FF8000", "#FFFF00", "#008000", "#0000FF", "#A000C0"))
  for (kk in 1:6) rect(0,kk-1,1,kk, col=col[kk], border=NA)
  points(1/2,6/2, pch=21, cex=21, lwd=80, col="#FFFFFF")
})
print(dataURI)

pathname <- toFavicon({
  plot(NA, xlim=c(0,1), ylim=c(0,6), axes=FALSE, xaxs="i", yaxs="i")
  col <- rev(c("#FF0000", "#FF8000", "#FFFF00", "#008000", "#0000FF", "#A000C0"))
  for (kk in 1:6) rect(0,kk-1,1,kk, col=col[kk], border=NA)
  points(1/2,6/2, pch=21, cex=21, lwd=80, col="#FFFFFF")
}, field="pathname")
print(pathname)

message("- toFavicon() ... DONE")


message("- favicon() errors ...")

## Exception handling
res <- try({
  favicon(width=32L, height=64L)
  dev.off()
})
stopifnot(inherits(res, "try-error"))

message("- favicon() errors ... DONE")

message("*** favicon() ... DONE")

message("*** devDump() ...")

library("R.devices")
graphics.off()


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copy content of current screen device
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if (interactive()) {
  # Open device
  plot(1:10)

  devList0 <- devList()
  devEval("png,jpg,pdf", name="count", tags="copy")
  stopifnot(all.equal(devList(), devList0))

  # Sanity checks
  print(devList())
  stopifnot(length(devList()) == 1L)


  # Same using a default name
  devList0 <- devList()
  devEval("png,jpg,pdf")
  stopifnot(all.equal(devList(), devList0))
  stopifnot(length(devList()) == 1L)

  # Close device
  devOff()

  stopifnot(length(devList()) == 0L)


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Copy content of all screen devices
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Open several devices
  idxs <- NULL

  idxs <- c(idxs, devNew())
  plot(1:10)

  idxs <- c(idxs, devNew())
  plot(cos)

  # Automatially "dump" image files of all open devices
  devDump()

  # "Manual" saving of defined
  which <- devList()
  print(which)

  # Save all
  devEval("png,pdf", which=which)

  # Close all opened devices
  devOff(idxs)
} # if (interactive())


message("*** devDump() ... DONE")

message("*** devIsOpen() ...")

library("R.devices")

for (which in list(c(), 1L, 1:5)) {
  print(which)
  res <- devIsOpen(which)
  print(res)
  stopifnot(is.logical(res))
  stopifnot(is.character(names(res)))
  stopifnot(length(res) == length(which))
}

message("*** devIsOpen() ... DONE")

getPathname <- R.devices:::getPathname
getPath <- R.devices:::getPath
view <- R.devices:::view
getData <- R.devices:::getData

message("*** DevEvalFileProduct ...")

# Return the DevEvalProduct object by default
R.devices::devOptions("*", field="*")

message("*** DevEvalFileProduct - image file ...")

p <- R.devices::toEPS("foo", tags=c("a", "b"), aspectRatio=0.7, {
  plot(1:10)
})
print(p)
str(p)

library("R.devices")

fields <- c("name", "fullname", "filename", "pathname", "path", "mime", "dataURI")
for (ff in fields) {
  cat(sprintf("%s: %s\n", ff, substring(p[[ff]], 1, 50)))
}

pathnameA <- getPathname(p, relative=FALSE)
cat(sprintf("Absolute pathname: %s\n", pathnameA))
path <- getPath(p, relative=FALSE)
cat(sprintf("Path: %s\n", path))
pathnameR <- getPathname(p, relative=TRUE)
cat(sprintf("Relative pathname: %s\n", pathnameR))

data <- getData(p, mode="character")
str(data)

data <- getData(p, mode="raw")
str(data)


## Call view() but use void browser
view(p, browser="false")

message("*** DevEvalFileProduct - image file ... DONE")

message("*** DevEvalFileProduct - missing file ...")

# An empty file product
na <- DevEvalFileProduct()
print(na)
cat(sprintf("Pathname: %s\n", getPathname(na)))

message("*** DevEvalFileProduct - missing file ... DONE")

message("*** DevEvalFileProduct ... DONE")

message("*** devTypeName() ...")

library("R.devices")
printf <- R.utils::printf

.devTypeName <- R.devices:::.devTypeName

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# By name
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
types <- list(empty=character(0L), png="png", jpg="jpg", mixed=c("png", "png", "jpeg"))
for (name in names(types)) {
  type <- types[[name]]
  printf("%s: .devTypeName(%s): ", name, deparse(type))
  res <- .devTypeName(type)
  printf("'%s'\n", deparse(res))
  stopifnot(is.character(res))
  stopifnot(is.character(names(res)))
}


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# By function (returns the same function)
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
types <- list(png=grDevices::png, jpg=grDevices::jpeg)
for (name in names(types)) {
  type <- types[[name]]
  printf("%s: .devTypeName(%s): ", name, deparse(args(type)))
  res <- .devTypeName(type)
  printf("'%s'\n", deparse(args(res)))
  stopifnot(is.function(res))
  stopifnot(identical(res, type))
}


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Special cases
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
## Special case: Default device
type <- getOption("device")

message("*** devTypeName() ... DONE")

library("R.devices")
devEqualTypes <- R.devices:::.devEqualTypes
png <- grDevices::png
postscript <- grDevices::postscript

message("*** devEqualTypes() ...")

message("*** devEqualTypes('png', 'png') ...")
res <- devEqualTypes("png", "png")
stopifnot(res)

message("*** devEqualTypes('png', png) ...")
res <- devEqualTypes("png", png)
stopifnot(res)

message("*** devEqualTypes(png, 'png') ...")
res <- devEqualTypes(png, "png")
stopifnot(res)

message("*** devEqualTypes(foo, png) ...")
foo <- png
res <- devEqualTypes(foo, png)
stopifnot(res)

message("*** devEqualTypes('png', postscript) ...")
res <- devEqualTypes("png", postscript)
stopifnot(!res)

message("*** devEqualTypes(postscript, 'png') ...")
res <- devEqualTypes(postscript, "png")
stopifnot(!res)

message("*** devEqualTypes('non-existing', png) ...")
res <- devEqualTypes("non-existing", png)
stopifnot(!res)

message("*** devEqualTypes(png, 'non-existing') ...")
res <- devEqualTypes(png, "non-existing")
stopifnot(!res)

message("*** devEqualTypes(nulldev, 'nulldev') ...")
res <- devEqualTypes(nulldev, "nulldev")
stopifnot(res)

message("*** devEqualTypes() ... DONE")

message("*** capabilitiesX11() ...")

res <- R.devices::capabilitiesX11()
print(res)

res2 <- R.devices::capabilitiesX11()
stopifnot(identical(res2, res))

res3 <- R.devices::capabilitiesX11(reset=TRUE)
print(res3)

message("*** capabilitiesX11() ... DONE")

library("R.devices")

message("*** capturePlot() ...")

cat("Default graphics device:\n")
str(getOption("device"))

message("*** capturePlot() - as.architecture() ...")

path <- system.file("exdata", package="R.devices")
pattern <- "^capturePlot,.*[.]rds$"
pathnames <- dir(path=path, pattern=pattern, full.names=TRUE)

for (kk in seq_along(pathnames)) {
  pathname <- pathnames[kk]
  message(sprintf("- File #%d ('%s') ...", kk, pathname))
  
  g <- readRDS(pathname)
  arch <- architecture(g)
  str(arch)

  ## Currently, as.architecture() cannot coerce endianess
  if (arch$endian == .Platform$endian) {
    ## Currently, as.architecture() cannot change pointer size
#   if (arch$ptrsize == .Machine$sizeof.pointer) {
      g8_1 <- as.architecture(g, ptrsize=8L)
      arch8_1 <- architecture(g8_1)
      str(arch8_1)
    
      g8_1b <- as.architecture(g8_1, ptrsize=8L)
      arch8_1b <- architecture(g8_1b)
      str(arch8_1b)
      stopifnot(identical(arch8_1b, arch8_1))
      stopifnot(identical(g8_1b, g8_1))
    
      g4_1 <- as.architecture(g8_1, ptrsize=4L)
      arch4_1 <- architecture(g4_1)
      str(arch4_1)
    
      g8_2 <- as.architecture(g4_1, ptrsize=8L)
      arch8_2 <- architecture(g8_2)
      str(arch8_2)
      stopifnot(identical(g8_2, g8_1))
    
      g_2 <- as.architecture(g)
      arch_2 <- architecture(g_2)
      str(arch_2)
    
      if (getRversion() >= "3.3.0") {
        try(replayPlot(g_2))
      }
#   } ## if (arch$ptrsize == .Machine$sizeof.pointer)
  } ## if (arch$endian == .Platform$endian)
} ## for (kk ...)

message("*** capturePlot() - as.architecture() ... DONE")

message("*** capturePlot() - capture and replay ...")

if (getRversion() >= "3.3.0") {
  message("- capture")
  g <- capturePlot({
    plot(1:10)
  })

  message("- system information")
  system <- attr(g, "system")
  print(system)

  message("- saving")
  ## Record for troubleshooting
  tags <- sprintf("%s=%s", names(system), system)
  pathname <- sprintf("capturePlot,%s.rds", paste(tags, collapse=","))
  saveRDS(g, file=pathname)

  message("- architecture")
  print(architecture(g))

  message("- replay")
  ## Replay
  replayPlot(g)

  message("- display")
  ## Display
  print(g)

  message("- toDefault()")
  ## Display with a 2/3 aspect ratio
  toDefault(aspectRatio=2/3, print(g))

  message("- devEval()")
  ## Redraw to many output formats
  devEval(c("{png}", "{eps}", "{pdf}"), aspectRatio=2/3, print(g))

} ## if (getRversion() >= "3.3.0")

message("*** capturePlot() - capture and replay ... DONE")

message("*** capturePlot() ... DONE")

message("*** devList() ...")

library("R.devices")

res <- devList()
print(res)
stopifnot(is.integer(res))
stopifnot(is.character(names(res)))

res <- devList(dropNull=FALSE)
print(res)
stopifnot(is.integer(res))
stopifnot(is.character(names(res)))

res <- devList(interactiveOnly=TRUE)
print(res)
stopifnot(is.integer(res))
stopifnot(is.character(names(res)))

# - - - - - - - - - - - - - - - - - - - - - - - - -
# Labels
# - - - - - - - - - - - - - - - - - - - - - - - - -
devSetLabel(which=1L, label="foo")
label <- devGetLabel(1L)
print(label)
stopifnot(label == "foo")
label <- devGetLabel("foo")
print(label)
stopifnot(label == "foo")

devSetLabel(which=1L, label="bar")
label <- devGetLabel(1L)
print(label)
stopifnot(label == "bar")
label <- devGetLabel("bar")
print(label)
stopifnot(label == "bar")

devSetLabel(which="bar", label="foo")
label <- devGetLabel(1L)
print(label)
stopifnot(label == "foo")
label <- devGetLabel("foo")
print(label)
stopifnot(label == "foo")


res <- try(devGetLabel(which=10L))
stopifnot(inherits(res, "try-error"))

message("*** devList() ... DONE")

message("*** devIsInteractive() ...")

library("R.devices")

types <- list(c(), "png", "x11", c("png", "jpeg", "png", "x11"), x11, png)
for (types in types) {
  print(types)
  res <- devIsInteractive(types)
  print(res)
  stopifnot(is.logical(res))
  stopifnot(is.character(names(res)))
  stopifnot(length(res) == length(types))
  if (is.character(types)) stopifnot(all(names(res) == types))
}

message("*** devIsInteractive() ... DONE")

library("R.devices")

message("*** toNnn() ...")

## Get all toNnn() functions
envir <- getNamespace("R.devices")
names <- ls(pattern="^to", envir=envir)
fcns <- list()
for (name in names) {
  if (exists(name, envir=envir, mode="function")) {
    fcns[[name]] <- get(name, envir=envir, mode="function")
  }
}

for (name in names(fcns)) {
  message(sprintf("*** %s() ...", name))
  toFcn <- fcns[[name]]
  tryCatch({
    toFcn(name=name, { plot(1:10) })
  }, error = function(ex) {
    print(ex)
  })
  message(sprintf("*** %s() ... DONE", name))
}

## FIXME:
graphics.off()

message("*** toNnn() ... DONE")

view <- R.devices:::view

message("*** DevEvalProduct ...")

message("*** DevEvalProduct - subsetting ...")

p <- R.devices:::DevEvalProduct("foo", tags=c("a", "b"))
for (ff in c("fullname", "name", "tags")) {
  cat(sprintf("%s: %s\n", ff, substring(p[[ff]], 1, 50)))
}

library("R.devices")

p <- DevEvalProduct("foo", tags=c("a", "b"))
for (ff in c("fullname", "name", "tags")) {
  cat(sprintf("%s: %s\n", ff, substring(p[[ff]], 1, 50)))
}

p <- DevEvalProduct("foo", tags=c("a", "b"))
valueA <- p[["name"]]
valueB <- p$name
stopifnot(identical(valueB, valueA))
value <- p[["non-existing-field"]]
stopifnot(is.null(value))

message("*** DevEvalProduct - subsetting ... DONE")


message("*** DevEvalProduct - view() ...")
view(p)
!p
message("*** DevEvalProduct - view() ... DONE")

message("*** DevEvalProduct ... DONE")

message("*** devNew() ...")

library("R.devices")


message("*** devNew(which=which) ...")

idx <- devNew()
plot(1:10)
idx2 <- devNew(which=idx)
plot(1:10)
devOff(idx2)
if (idx2 != idx) devOff(idx)

message("*** devNew(which=which) ... DONE")


message("*** devNew(aspectRatio) ...")
## Height is inferred
devNew(width=10, aspectRatio=1.5)
plot(1:10)
devOff()

## Width is inferred
devNew(height=10, aspectRatio=1.5)
plot(1:10)
devOff()
message("*** devNew(aspectRatio) ... DONE")


message("*** devNew(scale) ...")

devNew(width=10, scale=1.5)
plot(1:10)
devOff()

devNew(height=10, scale=1.5)
plot(1:10)
devOff()

## Force 'width' from device options
devNew(aspectRatio=NULL, scale=1.5)
plot(1:10)
devOff()

message("*** devNew(scale) ... DONE")


message("*** devNew() - warnings ...")

ok <- tryCatch({
  devNew(width=10, height=10, aspectRatio=1.5)
  FALSE
}, warning = function(warn) {
  print(warn)
  TRUE
})
stopifnot(ok)

res <- try(devNew(par=list(1)))
stopifnot(inherits(res, "try-error"))

message("*** devNew() - warnings ... DONE")


message("*** devNew() - errors ...")

res <- try(devNew(par=c(pch=1)))
stopifnot(inherits(res, "try-error"))

res <- try(devNew(par=list(1)))
stopifnot(inherits(res, "try-error"))

devNew(label="foo")
plot(1:10)
res <- try(devNew(label="foo"))
stopifnot(inherits(res, "try-error"))
devOff()

message("*** devNew() - errors ... DONE")

message("*** devOff() ... ")

## Open and close device
idx0 <- devNew()
idx1 <- devOff(idx0)
str(idx1)

## Close same device again (should silently return)
idx2 <- devOff(idx0)
str(idx2)

stopifnot(identical(idx2, idx1))

## Close many devices
idx3 <- devOff(2:5)
str(idx3)

message("*** devOff() ... DONE")


message("*** devDone() ... ")

## Open and close device
idx0 <- devNew()
idx1 <- devDone(idx0)
str(idx1)

## Close same device again (should silently return)
idx2 <- devDone(idx0)
str(idx2)

stopifnot(identical(idx2, idx1))

## Close many devices
idx3 <- devDone(1:5)
str(idx3)

message("*** devDone() ... DONE")

message("*** devNew() ... DONE")

message("*** withPar() ...")

library("R.devices")

x <- 1
message("x=", x)

withPar({
  layout(1:4)

  x <- c(x, 2)
  message("x=", x)
  stopifnot(all(x == 1:2))

  withPar({
    plot(1:10)
    plot(10:1)
    x <- c(x, 3)
    message("x=", x)
  }, pch=4)
  message("x=", x)
  stopifnot(all(x == 1:3))

  withPar({
    plot(1:10)
    plot(10:1)
    x <- c(x, 4)
    message("x=", x)
  }, pch=0, bg="yellow")
  message("x=", x)
  stopifnot(all(x == 1:4))

  x <- c(x, 5)
  message("x=", x)
  stopifnot(all(x == 1:5))
}, mar=c(2,2,1,1))

message("x=", x)
stopifnot(all(x == 1:5))


# Graphical parameters set "manually" are also reset
opar <- par()
withPar({
  par(pch=4L, lwd=3)
  plot(1:10)
})
stopifnot(all.equal(par(), opar))

message("*** withPar() ... DONE")

message("*** devEval() - errors ...")

library("R.devices")
hpaste <- R.utils::hpaste
graphics.off()


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Error handling: Incomplete image file is removed
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
path <- getDevOption("pdf", "path")
files0 <- dir(path=path)

tryCatch({
  res <- devEval(type="pdf", name="error", {
    plot(1:10)
    v <- log("0")
    abline(v=v)
  }, onIncomplete="remove")
  print(res)
}, error = function(ex) {
  message("An error occurred while plotting: ", ex$message)
})

## Assert that any image created was removed
files <- dir(path=path)
new <- setdiff(files, files0)
if (length(new) > 0L) {
  stop("Failed to remove incomplete image file: ", hpaste(sQuote(new)))
}


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Error handling: Incomplete image file is renamed
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
path <- getDevOption("pdf", "path")

## Try many times to test unique renaming of files.
for (kk in 1:5) {
  files0 <- dir(path=path)
  tryCatch({
    res <- devEval(type="pdf", name="error", {
      plot(1:10)
      v <- log("0")
      abline(v=v)
    }, onIncomplete="rename")
    print(res)
  }, error = function(ex) {
    message("An error occurred while plotting: ", ex$message)
  })

  ## Assert that any image created was removed
  files <- dir(path=path)
  new <- setdiff(files, files0)
  if (length(new) != 1L) {
    stop("Failed to rename incomplete image file")
  }
  cat("Incomplete image file: ", new, "\n", sep="")
}

# Sanity checks
print(devList())
stopifnot(length(devList()) == 0L)

message("*** devEval() - errors ... DONE")

#' Get the architecture of an object or coerce it into another
#'
#' @param x The object to be coerced.
#' @param ... (optional) Additional arguments passed to the underlying method.
#'
#' @return
#' \code{architecture()} returns a named list with
#'     character element \code{ostype} and \code{arch},
#'     integer element \code{ptrsize}, and character element \code{endian}.
#'     These elements take a missing values if they could not be inferred.
#'
#' @aliases as.architecture
#' @export
architecture <- function(x, ...) {
  UseMethod("architecture")
}

#' @param ostype A character string, e.g. \code{"unix"} and \code{"windows"}.
#' @param arch A character string, e.g. \code{"i386"}, \code{"i686"} and \code{"x86_64"}.
#' @param ptrsize The target pointer size - either \code{4L} or \code{8L}.
#' @param endian The target endianess - either \code{"little"} or \code{"big"}.
#'
#' @return
#' \code{as.architecture()} returns a coerced version of \code{x}.
#'     If no coercion was needed, then \code{x} itself is returned.
#'
#' @rdname architecture
#' @export
as.architecture <- function(x, ostype=.Platform$OS.type, arch=R.version$arch, ptrsize=.Machine$sizeof.pointer, endian=.Platform$endian, ...) {
  UseMethod("as.architecture")
}


#' @export
setMethodS3("architecture", "RecordedPlot", function(x, ...) {
  system <- attr(x, "system")
  if (is.null(system)) return(NextMethod())

  ostype <- system$ostype
  if (is.null(ostype)) ostype <- NA_character_

  arch <- system$arch
  if (is.null(arch)) arch <- NA_character_
  
  ptrsize <- system$ptrsize
  if (is.null(ptrsize)) {
    ## Default pointer size is 8 bytes (64-bit)
    gpar_raw <- gpar(x)
    n <- length(gpar_raw)
    known_sizes <- c("32 bit"=35956L, "64 bit"=35992L)
    if (n == known_sizes["32 bit"]) {
      ptrsize <- 4L
    } else if (n == known_sizes["64 bit"]) {
      ptrsize <- 8L
    } else {
      ptrsize <- NA_integer_
    }
  }
  
  endian <- system$endian
  if (is.null(ptrsize)) endian <- NA_character_

  list(ostype=ostype, arch=arch, ptrsize=ptrsize, endian=endian)
}) ## architecture() for RecordedPlot


#' @export
setMethodS3("architecture", "recordedplot", function(x, ...) {
  ## OS type is unknown by default
  ostype <- NA_character_
  
  ## Architecture label is unknown by default
  arch <- NA_character_

  ## Guess pointer size for size of 'gpar' element
  ## NOTE: This is not always a correct guess, but
  ##       it's better than nothing. /HB 2016-09-18
  gpar_raw <- gpar(x)
  n <- length(gpar_raw)
  known_sizes <- c("32 bit"=35956L, "64 bit"=35992L)
  if (n == known_sizes["32 bit"]) {
    ptrsize <- 4L
  } else if (n == known_sizes["64 bit"]) {
    ptrsize <- 8L
  } else {
    ptrsize <- NA_integer_
  }

  ## Endian is unknown by default
  endian <- NA_character_

  list(ostype=ostype, arch=arch, ptrsize=ptrsize, endian=endian)
}) ## architecture() for recordedplot



#' @export
setMethodS3("as.architecture", "recordedplot", function(x, ostype=.Platform$OS.type, arch=R.version$arch, ptrsize=.Machine$sizeof.pointer, endian=.Platform$endian, ...) {
  stop_if_not(is.character(arch), length(arch) == 1)
  stop_if_not(ptrsize %in% c(4L, 8L))
  stop_if_not(is.character(endian), length(endian) == 1, (is.na(endian) || endian %in% c("little", "big")))

  ## Default pointer size is 8 bytes (64-bit)
  x_arch <- architecture(x)

  ## Nothing to do?
  if (!is.na(arch) && !is.na(x_arch$arch) && arch == x_arch$arch && ptrsize == x_arch$ptrsize) return(x)

  ## SPECIAL: Source and target architectures are known
  ## to be compatible even though their ptrsizes differ
  if (all(c(arch, x_arch$arch) %in% c("i386", "x86_64"))) {
    # i386 (e.g. 32-bit Windows) <-> x86_64 (64-bit Linux)
    return(x)
  }

  ## Endianess?
  if (is.na(endian) || is.na(x_arch$endian) || endian != x_arch$endian) {
    stop(sprintf("NON-IMPLEMENTED FEATURE: Don't know how to coerce from %s to %s endianess", sQuote(x_arch$endian), sQuote(endian)))
  }

  ## Pointer size, i.e. 32-bit or 64-bit address space?
  known_sizes <- c("32 bit"=35956L, "64 bit"=35992L)
  if (is.na(x_arch$ptrsize)) {
     stop(sprintf("Failed to infer architecture.  The size of the %s structure is not among the known ones (%s): %d bytes", sQuote("gpar"), paste(sprintf("%s: %s bytes", names(known_sizes), known_sizes), collapse=", "), length(gpar(x))))
  }

  ## Nothing to do?
  if (ptrsize == x_arch$ptrsize && endian == x_arch$endian) return(x)

  ## Coerce 'gpar' structure
  pad64pos <- c(cex=29, crt=53, lwd=325, ps=389, srt=405,
                heights=597, plt=35485, mar=35549, mex=35621)
  gpar <- gpar(x)
  pkgName <- attr(gpar, "pkgName")

  if (x_arch$ptrsize == 8L && ptrsize == 4L) {
    ## 64-bit -> 32-bit
    drop <- rep(pad64pos, each=4L) + 0:3
    gpar <- gpar[-drop]
  } else if (x_arch$ptrsize == 4L && ptrsize == 8L) {
    ## 32-bit -> 64-bit
    for (pos in pad64pos) gpar <- append(gpar, raw(4L), after=pos-1L)
  }

  attr(gpar, "pkgName") <- pkgName

  ## Keep the result only if padded to a known length
  if (length(gpar) %in% known_sizes) gpar(x) <- gpar

  x
}) ## as.architecture() for recordedplot


## Internal gpar() and gpar<-() functions for recordedplot
gpar <- function(x) {
  stop_if_not(inherits(x, "recordedplot"))
  idx <- which(sapply(x, FUN=function(x) identical(attr(x, "pkgName"), "graphics")))
  stop_if_not(length(idx) > 0)
  raw <- x[[idx]]
  stop_if_not(is.raw(raw))
  raw
} ## gpar()


`gpar<-` <- function(x, value) {
  stop_if_not(is.raw(value))
  stop_if_not(inherits(x, "recordedplot"))
  idx <- which(sapply(x, FUN=function(x) identical(attr(x, "pkgName"), "graphics")))
  stop_if_not(length(idx) > 0)
  x[[idx]] <- value
  invisible(x)
} ## gpar<-()

#' @export
setMethodS3("as.architecture", "RecordedPlot", function(x, ...) {
  y <- NextMethod()
  system <- attr(x, "system")
  if (is.null(system)) return(y)

  ## Update 'system' attribute
  attr(y, "system") <- NULL
  arch <- architecture(y)
  if (is.na(arch$ostype)) arch$ostype <- system$ostype
  if (is.na(arch$ptrsize)) arch$ptrsize <- system$ptrsize
  if (is.na(arch$endian)) arch$endian <- system$endian
  attr(y, "system") <- arch
  
  y
})

#########################################################################/**
# @RdocFunction favicon
#
# @title "Favicon graphics device"
#
# \description{
#  Device driver for PNG favicons (Favorite icon) also known as
#  shortcut icon, Web site icon, tab icon or bookmark icon.
#  This driver is the same as the png driver where some arguments
#  have different default values.
# }
#
# @synopsis
#
# \arguments{
#   \item{file}{Default file name (pattern).}
#   \item{width, height}{The width and height of the figure.}
#   \item{par}{A named @list of graphical parameters to use.}
#   \item{...}{Other arguments accepted by \code{png()}.}
# }
#
# \value{
#   A plot device is opened; nothing is returned.
# }
#
# \examples{\dontrun{
#   favicon(width=32L)
#
#   # is identical to
#
#   suppressWarnings({
#     png("favicon.png", width=32L, height=32L, bg="transparent",
#                        par=list(mar=c(0,0,0,0)))
#   })
# }}
#
# \keyword{device}
#
# \seealso{
#   Internally, @see "grDevices::png" is used.
#   %% Add HTML Favicon script
#   %% toFavicon(plot(1, col="blue", bg="yellow",
#   %%           pch=21, cex=4, lwd=4, axes=FALSE))
#   \if{html}{\out{
#    <script>
#     var link = document.createElement('link');
#     link.rel = 'icon';
#     link.href = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAMAAABEpIrGAAAADFBMVEX9/v0AAP/9/v3//wBEQjoBAAAABHRSTlP//wD//gy7CwAAAGJJREFUOI3N0rESwCAIA9Ag///PXdoiBk0HhmbNO49DMETQCexNCSyFgdlGoO5DYOr9ThLgPosA7osIQP0sHuDOog8UI/ALa988wzdwXJRctf4s+d36YPTJ6aMd8ux3+QO4ABTtB85yDAh9AAAAAElFTkSuQmCC";
#     document.getElementsByTagName('head')[0].appendChild(link);
#    </script>
#   }}
# }
#
# @author
#
# @keyword device
# @keyword internal
#*/#########################################################################
favicon <- function(filename="favicon.png", width=32L, height=width, bg="transparent", par=list(mar=c(0,0,0,0)), ...) {
  # Argument 'width' and 'height':
  if (height != width) {
    throw("The width and the height must be the same for a favicon: ",
          width, " != ", height)
  }

  # Argument 'par':
  if (!is.null(par)) {
    if (!is.list(par) || is.null(names(par))) {
      throw("Argument 'par' has to be a named list: ", mode(par))
    }
  }

  # Create PNG file
  # png() will generate a warning that "width=.. , height=.., are unlikely
  # values in pixels" if width < 20.
  suppressWarnings({
    png(filename=filename, width=width, height=height, bg=bg, ...)
  })

  # Set graphical parameters
  par(par)
} # favicon()

###########################################################################/**
# @RdocDocumentation toNNN
# @alias toBMP
# @alias toEMF
# @alias toEPS
# @alias toPDF
# %% @alias toCairoPNG
# @alias toPNG
# @alias toSVG
# @alias toTIFF
# @alias toWMF 
# @alias toFavicon
# @alias toNullDev
# @alias asDataURI
#
# % Interactive/screen devices
# @alias toDefault
# @alias toQuartz
# @alias toWindows
# @alias toX11
# @alias toCairoWin
# @alias toCairoX11
# %% @alias toJavaGD
#@alias toRStudioGD
#
# @title "Methods for creating image files of a specific format"
#
# \description{
#  @get "title".
# }
#
# \usage{
#   toBMP(name, ...)
#   toPDF(name, ...)
#   toPNG(name, ...)
#   toSVG(name, ...)
#   toTIFF(name, ...)
#   toEMF(name, ..., ext="emf")
#   toWMF(name, ..., ext="wmf")
#
#   toFavicon(..., name="favicon", ext="png",
#             field=getDevOption("favicon", "field", default="htmlscript"))
#
#   % Interactive/screen devices
#   toDefault(name, ...)
#   toQuartz(name, ...)
#   toX11(name, ...)
#   toWindows(name, ...)
#
#   % Interactive/screen devices from other packages
#   toCairoWin(name, ...)
#   toCairoX11(name, ...)
#   %% toJavaGD(name, ...)
#
#   % Interactive/screen device for RStudio
#   toRStudioGD(name, ..., .allowUnknownArgs = TRUE)
# }
#
# \arguments{
#   \item{name}{A @character string specifying the name of the image file.}
#   \item{..., .allowUnknownArgs}{Additional arguments passed to @see "devEval", e.g.
#      \code{tags} and \code{aspectRatio}.}
#   \item{ext, field}{Passed to @see "devEval".}
# }
#
# \value{
#   Returns by default the @see "DevEvalProduct".
#   For \code{toFavicon()} the default return value is a @character string.
# }
#
# \section{Windows Metafile Format}{
#   Both \code{toEMF()} and \code{toWMF()} use the exact same graphics
#   device (\code{win.metafile()}) and settings.  They only differ by
#   filename extension.  The \code{win.metafile()} device function exists
#   on Windows only; see the \pkg{grDevices} package for more details.
# }
#
# @author
#
# \seealso{
#   These functions are wrappers for @see "devEval".
#   See @see "devOptions" to change the default dimensions for
#   a specific device type.
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
toBMP <- function(name, ...) {
  devEval(type="{bmp}", name=name, ...);
}

toEMF <- function(name, ..., ext="emf") {
  devEval(type="{win.metafile}", name=name, ..., ext=ext);
}

toWMF <- function(name, ..., ext="wmf") {
  devEval(type="{win.metafile}", name=name, ..., ext=ext);
}

toEPS <- function(name, ...) {
  devEval(type="{eps}", name=name, ...);
}

toPDF <- function(name, ...) {
  devEval(type="{pdf}", name=name, ...);
}

toPNG <- function(name, ...) {
  devEval(type="{png}", name=name, ...);
}

toSVG <- function(name, ...) {
  devEval(type="{svg}", name=name, ...);
}

toTIFF <- function(name, ...) {
  devEval(type="{tiff}", name=name, ...);
}

toFavicon <- function(..., name="favicon", ext="png", field=getDevOption("favicon", "field", default="htmlscript")) {
  # Output as "<script>...</script>"?
  if (identical(field, "htmlscript")) {
    uri <- devEval(type="{favicon}", name=name, ext=ext, field="dataURI", ...);
    script <- c(
      "<script>",
      " var link = document.createElement('link');",
      " link.rel = 'icon';",
      sprintf(' link.href = "%s"', uri),
      " document.getElementsByTagName('head')[0].appendChild(link);",
      "</script>",
      ""
    )
    paste(script, collapse="\n")
  } else {
    devEval(type="favicon", name=name, ext=ext, field=field, ...);
  }
} # toFavicon()

toNullDev <- function(name, ...) {
  devEval(type = "{nulldev}", name = name, ...)
}

asDataURI <- function(..., mime=NULL) {
  img <- DevEvalFileProduct(...);
  if (is.null(mime)) {
    mime <- getMimeType(img);
  }
  getDataURI(img, mime=mime);
}


toDefault <- function(name, ...) {
  devEval(type=getOption("device"), name=name, ...);
}

toQuartz <- function(name, ...) {
  devEval(type="{quartz}", name=name, ...);
}

toWindows <- function(name, ...) {
  devEval(type="{windows}", name=name, ...);
}

toX11 <- function(name, ...) {
  devEval(type="{x11}", name=name, ...);
}

toCairoWin <- function(name, ...) {
  devEval(type="{CairoWin}", name=name, ...);
}

toCairoX11 <- function(name, ...) {
  devEval(type="{CairoX11}", name=name, ...);
}

toRStudioGD <- function(name, ..., .allowUnknownArgs = TRUE) {
  ## WORKAROUND: Calling dev.new() in RStudio will open a RStudioGD device
  ## unless it is already opened, which in case it will open new graphics
  ## devices as when done outside of RStudio. In other words, there can only
  ## be one RStudioGD device open at any time.
  ## Because of the above, call to toRStudioGD() should set current device to
  ## the RStudioGD one if already open, otherwise create a new one. To open
  ## an existing device, all we need to do is set 'name' to its label.
  ## PS. We cannot just close it and have it reopened, because then we'll
  ##     loose the plot history in RStudio itself.

  ## (a) RStudioGD is not opened
  idx <- which("RStudioGD" == names(dev.list())) + 1L
  if (length(idx) == 0) {
    return(devEval(type = "{RStudioGD}", name = name, ...,
                   .allowUnknownArgs = .allowUnknownArgs))
  }

  ## (b) RStudioGD _is_ opened
  
  ## Existing label of the RStudioGD device
  label <- devGetLabel(idx)

  ## Was the expression passed implicitly via 'name' instead?  If so, then we
  ## need to make sure to call devEval() with 'name' set to the label of the
  ## current RStudioGD device.
  expr <- substitute(name)
  class <- eval(expression(class(expr)))
  is_expression <- is.element(class, c("call", "{", "("))
  if (isTRUE(is_expression)) {
    args <- list(type = "RStudioGD", name = label, expr = expr, ...,
                 .allowUnknownArgs = .allowUnknownArgs)
    return(do.call(devEval, args = args, envir = parent.frame()))
  }

  ## Otherwise, just call it as is
  res <- devEval(type = "RStudioGD", name = label, ...,
                 .allowUnknownArgs = .allowUnknownArgs)

  ## Should the label of the RStudioGD device be updated  
  if (name != label) devSetLabel(idx, label = name)
  
  res
}

## toJavaGD <- function(name, ...) {
##   devEval(type="{JavaGD}", name=name, ...);
## }

## covr: skip=all

# Allows conflicts. For more information, see library() and
# conflicts() in [R] base.
.conflicts.OK <- TRUE

.onLoad <- function(libname, pkgname) {
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Register vignette engines
  # NOTE: Are we doing this here because of backward compatibility
  #       issues, e.g. early version of R 3.0.x? /HB 2013-09-28
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  try({
    vignetteEngine <- get("vignetteEngine", envir=asNamespace("tools"))
    # Add vignette engine, iff missing.  But why?!? /HB 2013-09-28
    if (is.element("R.rsp::rsp", names(tools::vignetteEngine()))) {
      vignetteEngine("rsp", package=pkgname, pattern="[.][^.]*[.]rsp$",
                      weave=R.rsp::rspWeave, tangle=R.rsp::rspTangle)
    }
  }, silent=TRUE)
  ns <- getNamespace(pkgname)
  pkg <- Package(pkgname)
  assign(pkgname, pkg, envir=ns)

  # Set default global options when package is loaded
  devOptions("*", reset=TRUE)
}


.onAttach <- function(libname, pkgname) {
  startupMessage(get(pkgname, envir=getNamespace(pkgname)))
}

###########################################################################/**
# @RdocFunction devEval
# @alias devDump
# @alias suppressGraphics
#
# @title "Opens a new graphics device, evaluate (graphing) code, and closes device"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{type}{Specifies the type of graphics device to be used.
#    The device is created and opened using @see "devNew".
#    Multiple types may be specified.}
#   \item{expr}{The @expression of graphing commands to be evaluated.}
#   \item{initially, finally}{Optional @expression:s to be evaluated
#    before and after \code{expr}. If \code{type} specifies multiple
#    devices, these optional @expression:s are only evaluated ones.}
#   \item{envir}{The @environment where \code{expr} should be evaluated.}
#   \item{name, tags, sep}{The fullname name of the image is specified
#     as the name with optional \code{sep}-separated tags appended.}
#   \item{ext}{The filename extension of the image file generated, if any.
#    By default, it is inferred from argument \code{type}.}
#   \item{...}{Additional arguments passed to @see "devNew".}
#   \item{filename}{The filename of the image saved, if any.
#    By default, it is composed of arguments \code{name}, \code{tags},
#    \code{sep}, and \code{ext}.  See also below.}
#   \item{path}{The directory where then image should be saved, if any.}
#   \item{field}{An optional @character string specifying a specific
#     field of the named result @list to be returned.}
#   \item{onIncomplete}{A @character string specifying what to do with
#     an image file that was incompletely generated due to an interrupt
#     or an error.}
#   \item{force}{If @TRUE, and the image file already exists, then it is
#     overwritten, otherwise not.}
#   \item{which}{A @vector of devices to be copied.  Only applied if
#     argument \code{expr} is missing.}
#   \item{.exprAsIs, .allowUnknownArgs}{(Internal use only).}
# }
#
# \value{
#   Returns a @see "DevEvalFileProduct" if the device generated an
#   image file, otherwise an @see "DevEvalProduct".
#   If argument \code{field} is given, then the field of the
#   @see "DevEvalProduct" is returned instead.
#   \emph{Note that the default return value may be changed in future releases.}
# }
#
# \section{Generated image file}{
#   If created, the generated image file is saved in the directory
#   specified by argument \code{path} with a filename consisting of
#   the \code{name} followed by optional comma-separated \code{tags}
#   and a filename extension given by argument \code{ext}.
#
#   By default, the image file is only created if the \code{expr}
#   is evaluated completely.  If it is, for instance, interrupted
#   by the user or due to an error, then any incomplete/blank image
#   file that was created will be removed.  This behavior can be
#   turned of using argument \code{onIncomplete}.
# }
#
# @examples "../incl/devEval.Rex"
#
# @author
#
# \seealso{
#   To change default device parameters such as the width or the height,
#   @see "devOptions".
#   @see "devNew".
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devEval <- function(type=getOption("device"), expr, initially=NULL, finally=NULL, envir=parent.frame(), name=NULL, tags=NULL, sep=getDevOption(type, "sep", default=","), ..., ext=NULL, filename=NULL, path=getDevOption(type, "path", default="figures"), field=getDevOption(type, name="field"), onIncomplete=c("remove", "rename", "keep"), force=getDevOption(type, "force", default=TRUE), which=dev.cur(), .exprAsIs=FALSE, .allowUnknownArgs=FALSE) {
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Vectorized version
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Nothing to do?
  if (length(type) == 0L) {
    return(structure(character(0L), class=class(DevEvalProduct())))
  }

  # Parse 'type' in case multiple types is specified in one string
  if (is.character(type)) {
    type <- unlist(strsplit(type, split=","), use.names=FALSE)
    type <- trim(type)
  }

  # Argument 'expr':
  hasExpr <- !missing(expr)

  # Argument '.allowUnknownArgs':
  ## If TRUE, it tells devNew() that it's ok to silently ignore
  ## arguments not recognized by certain device functions.
  ## Used when multiple device types are used in one devEval() call.
  .allowUnknownArgs <- as.logical(.allowUnknownArgs)


  ## Expand device type names by regexp matching, iff any
  type <- .devTypeName(type, pattern=TRUE)


  ## SPECIAL CASE: Handle calls like toNnn({ expr }) where the actual plot
  ## expression is actually passed via the first argument which is 'name'
  ## and not via 'expr', e.g. toX11({ plot(1:3) }).
  if (!hasExpr) {
    # Was the expression passed implicitly via 'name' instead?
    # In order to infer this, we have to inspect 'name' in the
    # parent frame:
    nameClass <- eval(expression(class(substitute(name))), envir = parent.frame(), enclos = baseenv())
    isExpr <- is.element(nameClass, c("call", "{", "("))
    if (isTRUE(isExpr)) {
      # This avoid the plot expression to be evaluated here.
      delayedAssign("expr", name)
      hasExpr <- TRUE
      nameOrg <- NULL
    } else {
      nameOrg <- name
    }
  } else {
    nameOrg <- name
  }


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # (a) No plot code expression?  Then copy multiple input devices...
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (!hasExpr && length(which) > 1L) {
    res <- list()
    for (kk in seq_along(which)) {
      idx <- which[kk]
      devSet(idx)
      res[[kk]] <- devEval(type=type, initially=NULL, finally=NULL, envir=envir, name=nameOrg, tags=tags, sep=sep, ..., ext=ext, filename=filename, path=path, field=field, onIncomplete=onIncomplete, force=force, which=idx, .allowUnknownArgs=TRUE)
    } # for (idx ...)
    names(res) <- names(which)
    return(res)
  } # if (length(which) > 1L)


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # (b) Plot code expression, but with multiple output types/devices?
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (length(type) > 1L) {
    types <- type

    # Evaluate 'initially' only once
    eval(initially, envir = envir, enclos = baseenv())

    if (hasExpr) {
      # Expression must be substitute():d to avoid the being evaluated here
      if (!.exprAsIs) expr <- substitute(expr)

      # Evaluate 'expr' once per graphics device
      res <- lapply(types, FUN=function(type) {
        devEval(type=type, expr=expr, initially=NULL, finally=NULL, envir=envir, name=nameOrg, tags=tags, sep=sep, ..., ext=ext, filename=filename, path=path, field=field, onIncomplete=onIncomplete, force=force, .exprAsIs=TRUE, .allowUnknownArgs=TRUE)
      })
    } else {
      # Evaluate 'expr' once per output device
      res <- lapply(types, FUN=function(type) {
        devEval(type=type,            initially=NULL, finally=NULL, envir=envir, name=nameOrg, tags=tags, sep=sep, ..., ext=ext, filename=filename, path=path, field=field, onIncomplete=onIncomplete, force=force, which=which, .allowUnknownArgs=TRUE)
      })
    } # if (hasExpr)
    names(res) <- types

    # Evaluate 'finally' only once
    eval(finally, envir = envir, enclos = baseenv())

    return(res)
  } # if (length(type) > 1L)

  # Sanity check
  stop_if_not(length(type) == 1L)


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # (c) Use first successful device type among several options?
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  types <- type
  if (is.character(types)) {
    types <- unlist(strsplit(types, split="|", fixed=TRUE))
    types <- trim(types)
    types <- unique(types)
  }

  if (length(types) > 1L) {
    res <- NULL
    errors <- list(); # Record any errors

    # Evaluate 'initially' only once
    eval(initially, envir = envir, enclos = baseenv())

    if (hasExpr) {
      # Expression must be substitute():d to avoid the being evaluated here
      if (!.exprAsIs) expr <- substitute(expr)

      # Evaluate 'expr' once per graphics device
      for (type in types) {
        tryCatch({
          res <- devEval(type=type, expr=expr, initially=NULL, finally=NULL, envir=envir, name=nameOrg, tags=tags, sep=sep, ..., ext=ext, filename=filename, path=path, field=field, onIncomplete=onIncomplete, force=force, .exprAsIs=TRUE, .allowUnknownArgs=TRUE)
          attr(res, "type") <- type
          break
        }, error = function(ex) {
          errors <<- c(errors, list(ex))
        })
      } # for (type ...)
    } else {
      # Evaluate 'expr' once per output device
      for (type in types) {
        tryCatch({
          res <- devEval(type=type,            initially=NULL, finally=NULL, envir=envir, name=nameOrg, tags=tags, sep=sep, ..., ext=ext, filename=filename, path=path, field=field, onIncomplete=onIncomplete, force=force, which=which, .allowUnknownArgs=TRUE)
          attr(res, "type") <- type
          break
        }, error = function(ex) {
          errors <<- c(errors, list(ex))
        })
      } # for (type ...)
    } # if (hasExpr)

    # Evaluate 'finally' only once
    eval(finally, envir = envir, enclos = baseenv())

    # Successfully created output?
    if (!is.null(res)) {
      return(res)
    }

    msg <- sprintf("None of the specified device types (%s) generated output", paste(sQuote(types), collapse=", "))
    if (length(errors) > 0L) {
      reasons <- sapply(errors, FUN=function(ex) ex$message)
      reasons <- unique(reasons)
      reasons <- sprintf("(%d) %s", seq_along(reasons), reasons)
      reasons <- paste(reasons, collapse="; ")
      msg <- sprintf("%s. The reasons reported were: %s", msg, reasons)
    }
    throw(msg)
  } # if (length(type) > 1L)


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # (d) Plot code expression and a single output type/device
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Sanity check
  stop_if_not(length(type) == 1L)

  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Validate arguments
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Argument 'type':
  # Sanity check
  if (length(type) != 1L) {
    throw("Argument 'type' must be a single object: ", length(type))
  }
  if (is.function(type)) {
  } else {
    type <- as.character(type)
    type <- .devTypeName(type)
  }

  # An interactive/non-file device?
  isInteractive <- devIsInteractive(type)

  # Argument 'name', 'tags' and 'sep':
  # Default 'name' value
  if (is.null(nameOrg) && !isInteractive) {
    if (hasExpr) {
      nameOrg <- "Rplot"; # Backward compatible
    } else {
      nameOrg <- devGetLabel(which)
    }
  }

  # Construct the full name
  fullname <- paste(c(nameOrg, tags), collapse=sep)
  fullname <- unlist(strsplit(fullname, split=sep, fixed=TRUE))
  fullname <- sub("^[\t\n\f\r ]*", "", fullname); # trim tags
  fullname <- sub("[\t\n\f\r ]*$", "", fullname); #
  fullname <- fullname[nchar(fullname) > 0L];     # drop empty tags
  fullname <- paste(fullname, collapse=sep)

  if (!isInteractive) {
    # Argument 'ext':
    if (is.null(ext)) {
      # Infer filename extension from the type
      if (is.character(type)) {
        ext <- .devTypeExt(type)
      } else {
        ext <- .devTypeExt(substitute(type))
      }
    }

    # Argument 'filename' & 'path':
    if (ext == "nulldev") {
      pathname <- nullfile()
    } else {
      if (is.null(filename)) {
        filename <- sprintf("%s.%s", fullname, ext)
      }
      pathname <- Arguments$getWritablePathname(filename, path=path)
    }
  }


  # Argument 'field':
  if (!is.null(field)) {
    field <- Arguments$getCharacter(field)
  }

  # Argument 'onIncomplete':
  onIncomplete <- match.arg(onIncomplete)

  # Argument 'force':
  force <- Arguments$getLogical(force)

  devIdx <- NULL
  if (!isInteractive) {
    # Make sure the currently open device, iff any, is still the active
    # one when returning from this function.
    devListEntry <- devList()
    devCur <- dev.cur()
    on.exit({
      if (devCur != 1L) devSet(devCur)

      # Assert that no temporarily opened devices are left behind,
      devListExit <- setdiff(devList(), devIdx)
      devListDiff <- setdiff(devListExit, devListEntry)
      if (length(devListDiff) > 0L) {
        types <- unlist(.devList())
        types <- types[devListDiff]
        throw("Detected new graphics devices that was opened but not closed while executing devEval(): ", paste(sprintf("%s (%s)", sQuote(devListDiff), types), collapse=", "))
      }
    }, add=TRUE)
  }

  # Result object
  if (isInteractive) {
    res <- DevEvalProduct(fullname, type=type)
  } else {
    res <- DevEvalFileProduct(pathname, type=type)
  }

  done <- FALSE
  if (force || isInteractive || !isFile(pathname)) {
    # Open device
    if (isInteractive) {
      devIdx <- devNew(type, which=fullname, ..., .allowUnknownArgs=.allowUnknownArgs)
    } else {
      devIdx <- devNew(type, pathname, ..., .allowUnknownArgs=.allowUnknownArgs)

      on.exit({
        # Make sure to close the device (the same that was opened),
        # unless it's an interactive device.
        if (!is.null(devIdx)) {
          devDone(devIdx)
        }

        # Archive file?
        if (isPackageLoaded("R.archive")) {
          # To please R CMD check
          getArchiveOption <- archiveFile <- NULL
          if (getArchiveOption("devEval", FALSE)) archiveFile(pathname)
        }

        if (!done && isFile(pathname)) {
          if (onIncomplete == "remove") {
            # Remove incomplete image file
            file.remove(pathname)
          } else if (onIncomplete == "rename") {
            # Rename incomplete image file to denote that
            # it is incomplete.
            pattern <- "^(.*)[.]([^.]*)$"
            if (regexpr(pattern, pathname) != -1L) {
              fullname <- gsub(pattern, "\\1", pathname)
              ext <- gsub(pattern, "\\2", pathname)
              fmtstr <- sprintf("%s,INCOMPLETE_%%03d.%s", fullname, ext)
            } else {
              fmtstr <- sprintf("%s,INCOMPLETE_%%03d", pathname)
            }

            # Try to rename
            maxTries <- 999L
            for (kk in seq_len(maxTries)) {
              pathnameN <- sprintf(fmtstr, kk)
              if (isFile(pathnameN)) next
              resT <- file.rename(pathname, pathnameN)
              # Done?
              if (resT && isFile(pathnameN)) {
                pathname <- pathnameN
                break
              }
            } # for (kk ...)

            # Failed to rename?
            if (!isFile(pathnameN) && isFile(pathname)) {
              throw(sprintf("Failed to rename incomplete image file (after %d tries): %s", pathname, kk))
            }
          } # if (onIncomplete == ...)
        } # if (!done && isFile(...))
      }, add=TRUE); # on.exit()
    } # if (isInteractive)

    # Evaluate 'initially', 'expr' and 'finally' (in that order)
    eval(initially, envir = envir, enclos = baseenv())
    if (hasExpr) {
      eval(expr, envir = envir, enclos = baseenv())
    } else {
      # Copy the device specified by argument 'which'.
      # Assert that device is interactive
      devList <- devList(interactiveOnly=TRUE)
      if (is.numeric(which)) {
        ok <- is.element(which, devList)
      } else {
        ok <- is.element(which, names(devList))
      }
      if (!ok) {
        types <- unlist(.devList())
        type <- types[which]
        throw(sprintf("Cannot copy a %s device - only interactive/screen devices are supported: %s", sQuote(type), sQuote(which)))
      }

      devSet(which)
      dev.copy(which=devIdx)
    }
    eval(finally, envir = envir, enclos = baseenv())

    done <- TRUE
  }

  # Close it here to make sure the image file is created.
  # This is needed if field="dataURI" (which triggers a
  # reading the image file).
  if (done) {
    devDone(devIdx)
    devIdx <- NULL
  }

  # Subset?
  if (!isInteractive && !is.null(field)) {
    res <- res[[field]]
  }

  res
} # devEval()


devDump <- function(type=c("png", "pdf"), ..., path=NULL, envir=parent.frame(), field=NULL, which=devList(interactiveOnly=TRUE)) {
  if (is.null(path)) {
    # Timestamp, e.g. 2011-03-10_041359.032
    timestamp <- format(Sys.time(), "%Y-%m-%d_%H%M%OS3", tz="")
    path <- getDevOption("*", "path", default="figures")
    path <- file.path(path, timestamp)
  }

  devEval(type=type, ..., path=path, envir=envir, field=field, which=which)
} # devDump()

###########################################################################/**
# @RdocFunction withPar
#
# @title "Evaluate an R expression with graphical parameters set temporarily"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{expr}{The R expression to be evaluated.}
#   \item{...}{Named options to be used.}
#   \item{args}{(optional) Additional named options specified as a named @list.}
#   \item{envir}{The @environment in which the expression should be evaluated.}
# }
#
# \value{
#  Returns the results of the expression evaluated.
# }
#
# \details{
#   Upon exit (also on errors), this function will reset \emph{all}
#   (modifiable) graphical parameters to the state of options available
#   upon entry.  This means any parameters \emph{modified} from evaluating
#   \code{expr} will also be undone upon exit.
# }
#
# @author
#
# @examples "../incl/withPar.Rex"
#
# \seealso{
#   Internally, @see "base::eval" is used to evaluate the expression,
#   and @see "graphics::par" to set graphical parameters.
# }
#
# @keyword IO
# @keyword programming
#*/###########################################################################
withPar <- function(expr, ..., args=list(), envir=parent.frame()) {
  # Argument '.expr':
  expr <- substitute(expr)

  # Argument 'args':
  if (!is.list(args)) {
    throw("Argument 'args' is not a list: ", class(args)[1L])
  }

  # Argument 'envir':
  if (!is.environment(envir)) {
    throw("Argument 'envir' is not a list: ", class(envir)[1L])
  }

  # All arguments specified
  new <- c(list(...), args)

  # Set parameters temporarily (restore *all* upon exit)
  prev <- par(no.readonly=TRUE)
  on.exit(par(prev))
  if (length(new) > 0L) par(new)

  eval(expr, envir = envir, enclos = baseenv())
} # withPar()

nullfile <- function() {
  switch(.Platform$OS.type,
    windows = "NUL",
    "/dev/null" #nolint
  )
}

is_nullfile <- function(file) {
  file <- normalizePath(file, mustWork = FALSE)
  file <- tolower(file)

  if (.Platform$OS.type == "windows") {
    # Any filename of NUL, NUL.<any extension>, NUL:<any path>
    file <- gsub("[.:].*", "", basename(file))
    (file == "nul")
  } else {
    (file == "/dev/null") #nolint
  }
}

stop_if_not <- function(...) {
  res <- list(...)
  n <- length(res)
  if (n == 0L) return()

  for (ii in 1L:n) {
    res_ii <- .subset2(res, ii)
    if (length(res_ii) != 1L || is.na(res_ii) || !res_ii) {
        mc <- match.call()
        call <- deparse(mc[[ii + 1]], width.cutoff = 60L)
        if (length(call) > 1L) call <- paste(call[1L], "...")
        stop(sQuote(call), " is not TRUE", call. = FALSE, domain = NA)
    }
  }
}

###########################################################################/**
# @RdocFunction png2
#
# @title "A PNG device for Bitmap Files via GhostScript"
#
# \description{
#   @get "title".
# }
#
# @synopsis
#
# \arguments{
#  \item{filename}{The name of the file to be produced.}
#  \item{width, height}{The width and height (in pixels) of the result image.}
#  \item{res}{The resolution of the image.}
#  \item{type}{The output type. See @see "grDevices::dev2bitmap" for details.}
#  \item{...}{Additional arguments passed to \code{bitmap()}.}
# }
#
# \value{
#   Returns nothing.
# }
#
# @author
#
# @keyword device
# @keyword internal
#*/###########################################################################
png2 <- function(filename, width=480, height=480, res=144, type="png256", ...) {
  bitmap(file=filename, type=type, width=width/res, height=height/res, res=res, ...)
}

###########################################################################/**
# @RdocFunction devNew
#
# @title "Opens a new device"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{type}{A @character string specifying the type of device to be
#     opened. This string should match the name of an existing device
#     @function.}
#   \item{...}{Additional arguments passed to the device @function, e.g.
#     \code{width} and \code{height}.  If not given, the are inferred
#     from @see "devOptions".}
#   \item{scale}{A @numeric scalar factor specifying how much the
#     width and the height should be rescaled.}
#   \item{aspectRatio}{A @numeric ratio specifying the aspect ratio
#     of the image.  See below.}
#   \item{par}{An optional named @list of graphical settings applied,
#     that is, passed to @see "graphics::par", immediately after
#     opening the device.}
#   \item{label}{An optional @character string specifying the label of the
#     opened device.}
# }
#
# \value{
#   Returns the device index of the opened device.
# }
#
# \section{Width and heights}{
#   The default width and height of the generated image is specific to
#   the type of device used.  There is not straightforward programmatic
#   way to infer these defaults; here we use @see "devOptions", which
#   in most cases returns the correct defaults.
# }
#
# \section{Aspect ratio}{
#   The aspect ratio of an image is the height relative to the width.
#   If argument \code{height} is not given (or @NULL), it is
#   calculated as \code{aspectRatio*width} as long as they are given.
#   Likewise, if argument \code{width} is not given (or @NULL), it is
#   calculated as \code{width/aspectRatio} as long as they are given.
#   If neither \code{width} nor \code{height} is given, then \code{width}
#   defaults to \code{devOptions(type)$width}.
#   If both \code{width} and \code{height} are given, then
#   \code{aspectRatio} is ignored.
# }
#
# @author
#
# \seealso{
#   @see "devDone" and @see "devOff".
#   For simplified generation of image files, see @see "devEval".
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devNew <- function(type=getOption("device"), ..., scale=1, aspectRatio=1, par=NULL, label=NULL) {
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Local functions
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  cleanLength <- function(x, ...) {
    name <- substitute(x)
    if (is.null(x) || !is.numeric(x) || !is.finite(x)) {
      warning("Ignoring non-finite '", name, "' value: ", x)
      x <- NULL
    }
    x
  } # cleanLength()

  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Validate arguments
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Argument 'type':
  if (length(type) != 1L) {
    throw("Argument 'type' must be a single object: ", length(type))
  }
  if (is.function(type)) {
  } else {
    type <- as.character(type)
    type <- .devTypeName(type)
  }

  # Argument 'scale':
  if (!is.null(scale)) {
    scale <- Arguments$getDouble(scale, range=c(0,Inf))
  }

  # Argument 'aspectRatio':
  if (!is.null(aspectRatio)) {
    aspectRatio <- Arguments$getDouble(aspectRatio, range=c(0,Inf))
  }

  # Argument 'par':
  if (!is.null(par)) {
    if (!is.list(par) || is.null(names(par))) {
      throw("Argument 'par' has to be a named list: ", mode(par))
    }
  }

  # Argument 'label':
  if (!is.null(label)) {
    if (any(label == names(devList())))
      throw("Cannot open device. Label is already used: ", label)
  }


  # Arguments to be passed to the device function
  args <- list(...)

  ## Secret argument from devEval(), which is intended to be used when
  ## multiple devices are used and not all accepts the same arguments.
  .allowUnknownArgs <- args$.allowUnknownArgs
  if (is.null(.allowUnknownArgs)) .allowUnknownArgs <- FALSE
  args$.allowUnknownArgs <- NULL

  # Drop 'width' and 'height', iff NULL (=treat as non-specified/missing)
  args$width <- args$width
  args$height <- args$height


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Update the 'height' by argument 'aspectRatio'?
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (!is.null(aspectRatio)) {
    width <- args$width
    height <- args$height

    # Were both 'width' and 'height' explicitly specified?
    if (!is.null(width) && !is.null(height)) {
      if (aspectRatio != 1) {
        warning("Argument 'aspectRatio' was ignored because both 'width' and 'height' were given: ", aspectRatio)
      }
    } else {
      # None of 'width' and 'height' was specified?
      if (is.null(width) && is.null(height)) {
        # (a) Infer 'width' from devOptions()...
        width <- devOptions(type)$width
        width <- cleanLength(width)
        if (!is.null(width)) {
          args$width <- width
          args$height <- aspectRatio * width
        } else {
	  typeT <- if (is.character(type)) dQuote(type) else "<function>"
          warning("Argument 'aspectRatio' was ignored because none of 'width' and 'height' were given and 'width' could not be inferred from devOptions(", typeT, "): ", aspectRatio)
        }
      } else if (!is.null(width)) {
        # Argument 'width' was specified but not 'height'
        args$height <- aspectRatio * width
      } else if (!is.null(height)) {
        # Argument 'height' was specified but not 'width'
        args$width <- height / aspectRatio
      }
    }
  } # if (!is.null(aspectRatio))


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Rescale 'width' & 'height' by argument 'scale'?
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (!is.null(scale) && scale != 1.0) {
    width <- args$width

    # Infer 'width' from the settings
    if (is.null(width)) {
      width <- devOptions(type)$width
      width <- cleanLength(width)
    }

    # Possible to rescale?
    if (is.null(width)) {
      warning("Argument 'scale' was ignored because it was not possible to infer 'width': ", scale)
    } else {
      # Infer 'height'...
      if (!is.null(aspectRatio)) {
        # ...from aspect ratio
        height <- aspectRatio * width
      } else {
        # ...from settings
        height <- args$height
        if (is.null(height)) {
          height <- devOptions(type)$height
          height <- cleanLength(height)
        }
        if (is.null(height)) {
          warning("Argument 'scale' was ignored because it was not possible to infer 'height': ", scale)
        }
      }
    }

    # So finally, possible to rescale?
    if (!is.null(width) && !is.null(height)) {
      args$width <- scale * width
      args$height <- scale * height
    }
  }


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Exclude 'file' and 'filename' arguments?
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  isInteractive <- devIsInteractive(type)
  if (isInteractive) {
    keep <- !is.element(names(args), c("file", "filename"))
    args <- args[keep]
  }


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Open an existing device?
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  which <- args[["which"]]
  if (is.null(which) || (nchar(which) == 0L) || !isInteractive) {
    # Default is to open a new one
    devIdx <- NA_integer_
  } else {
    # ...otherwise, is requested device already opened?
    devList <- devList(dropNull=FALSE)
    labels <- names(devList)

    # Default is to open a new one
    devIdx <- NA_integer_
    if (is.character(which)) {
      # An existing device by its label?
      devIdx <- match(which, table=labels)
      names(devIdx) <- labels[devIdx]
      if (is.null(label)) label <- which
    } else if (is.numeric(which)) {
      # An existing device by its index?
      if (which <= length(devList)) {
        devIdx <- which
        names(devIdx) <- labels[devIdx]
      }
    }
  }
  # Drop 'which' argument, if specified
  keep <- !is.element(names(args), "which")
  args <- args[keep]


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Open (new or existing) device
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # New or existing?
  if (is.na(devIdx)) {
    # (a) New, i.e. call the device function
    devList0 <- devList()

    typeT <- type
    if (.allowUnknownArgs) {
      # (b) Temporary append '...' to the device function to make it
      #     allow for more "unknown" arguments
      if (!is.function(type)) {
        typeT <- get(type, mode="function", inherits=TRUE)
      }
      typeT <- appendVarArgs(typeT)
    }

    if (getOption("R.devices::devNew/debug", FALSE)) {
      call <- list("devNew", typeT=typeT, args=args)
      R.utils::mstr(call)
    }

    do.call(typeT, args=args)

    # Make sure a new device was indeed opened.  This can happen
    # for graphics devices that does not throw an error, but only
    # a warning, e.g. quartz().
    opened <- setdiff(devList(), devList0)
    if (length(opened) == 0L) {
      throw("Failed to open graphics device: ", type)
    }

    # Retrieve the index of the recently opened device
    devIdx <- dev.cur()
  } else {
    # (b) Existing one, i.e. set focus.
    dev <- devSet(devIdx)

    # Assert that the existing device is of the requested type
    if (!.devEqualTypes(type, other=names(dev), args=args)) {
      if (is.function(type)) type <- "<a function>"
      throw(sprintf("Detected an existing devices with the requested label (which='%s'), but its device type is different from the requested type: '%s' != '%s'", names(devIdx), type, names(dev)))
    }
    # Make sure not to reset the device label below
    if (is.null(label)) label <- names(devIdx)
  }

  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Set the label of the recently opened device
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  devSetLabel(which=devIdx, label=label)

  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Default and user-specific parameters
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  parT <- getDevOption(type=type, name="par", old="devNew/args/par")
  # Append
  parT <- c(parT, par)
  if (length(parT) > 0L) {
    par(parT)
  }

  invisible(devIdx)
} # devNew()

###########################################################################/**
# @RdocClass DevEvalProduct
#
# @title "The DevEvalProduct class"
#
# \description{
#  @classhierarchy
#
#  A DevEvalProduct represents a handle to the "product" returned by
#  @see "devEval".
# }
#
# @synopsis
#
# \arguments{
#   \item{name, tags}{The name and optional tags of the product.}
#   \item{type}{The device type.}
#   \item{...}{Not used.}
# }
#
# \section{Fields}{
#  The following (virtual; calculate on-the-fly) fields are available:
#  \itemize{
#   \item \code{fullname}: the fullname of an image, e.g. 'foo,a,b'
#   \item \code{name}: the part of the fullname before the first comma, e.g. 'foo'
#   \item \code{tags}: the part of the fullname after the first comma, e.g. 'a,b'
#  }
# }
#
# @author
#
# @keyword internal
#*/###########################################################################
setConstructorS3("DevEvalProduct", function(name=NULL, tags=NULL, type=NULL, ...) {
  if (!is.null(name)) {
    name <- Arguments$getCharacter(name)
    tags <- Arguments$getCharacters(tags)
    fullname <- paste(c(name, tags), collapse=",")
  } else {
    fullname <- NA_character_
  }

  extend(BasicObject(fullname), "DevEvalProduct",
    type = type
  )
})

###########################################################################/**
# @RdocMethod as.character
# @alias as.character.DevEvalFileProduct
# @alias as.character
#
# @title "Gets a character representation of the product"
#
# \description{
#   @get "title".
# }
#
# @synopsis
#
# \arguments{
#  \item{...}{Not used.}
# }
#
# \value{
#   Returns a @character string.
# }
#
# @author
#
# \seealso{
#   @seeclass
# }
#*/###########################################################################
setMethodS3("as.character", "DevEvalProduct", function(x, ...) {
  getFullname(x, ...)
}, private=TRUE)


setMethodS3("view", "DevEvalProduct", function(object, ...) {})

setMethodS3("!", "DevEvalProduct", function(x) {
  view(x)
}, appendVarArgs=FALSE)

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# BEGIN: PATCH until `[[.BasicObject` finds methods in namespaces
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethodS3("[[", "DevEvalProduct", function(x, name, ...) {
  if (name == "*") return(x)
  fcn <- sprintf("get%s", capitalize(name))
  expr <- substitute(fcn(x), list(fcn=as.name(fcn)))
  res <- tryCatch({
    eval(expr, ..., enclos = baseenv())
  }, error = function(ex) {
    ex
  })
  if (inherits(res, "Exception")) throw(res)
  if (!inherits(res, "simpleError")) return(res)
  NextMethod()
}, private=TRUE)


setMethodS3("$", "DevEvalProduct", function(x, name) {
  name <- as.character(name)
  x[[name]]
}, private=TRUE)
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# END: PATCH until `[[.BasicObject` finds methods in namespaces
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


###########################################################################/**
# @RdocMethod getFullname
# @alias getFullname.DevEvalFileProduct
# @aliasmethod getName
# @aliasmethod getTags
# @alias getFullname
# @alias getName
# @alias getTags
#
# @title "Gets the full name, name and tags"
#
# \description{
#   @get "title" consisting of a name and tags.
# }
#
# \usage{
# @usage "getFullname,DevEvalProduct"
# @usage "getFullname,DevEvalFileProduct"
# @usage "getName,DevEvalProduct"
# @usage "getTags,DevEvalProduct"
# }
#
# \arguments{
#  \item{...}{Not used.}
# }
#
# \value{
#   Returns a @character or a @character @vector.
# }
#
# @author
#
# \seealso{
#   @seeclass
# }
#*/###########################################################################
setMethodS3("getFullname", "DevEvalProduct", function(this, ...) {
  as.character(unclass(this), ...)
})


setMethodS3("getName", "DevEvalProduct", function(this, ...) {
  fullname <- getFullname(this, ...)
  parts <- unlist(strsplit(fullname, split=","), use.names=FALSE)
  name <- parts[1L]
  name
})

setMethodS3("getTags", "DevEvalProduct", function(this, collapse=",", ...) {
  fullname <- getFullname(this, ...)
  parts <- unlist(strsplit(fullname, split=","), use.names=FALSE)
  tags <- parts[-1L]
  if (!is.null(collapse)) {
    tags <- paste(tags, collapse=collapse)
  }
  tags
})

#########################################################################/**
# @RdocMethod getType
# @alias getType
#
# @title "Gets the type"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{...}{Not used.}
# }
#
# \value{
#  Returns a @character string or @NULL.
# }
#
# @author
#
# \seealso{
#   @seeclass
# }
#*/#########################################################################
setMethodS3("getType", "DevEvalProduct", function(this, ...) {
  attr(this, "type")
})




###########################################################################/**
# @RdocClass DevEvalFileProduct
#
# @title "The DevEvalFileProduct class"
#
# \description{
#  @classhierarchy
#
#  A DevEvalFileProduct is a @see "DevEvalProduct" referring to a image
#  file created by @see "devEval".
# }
#
# @synopsis
#
# \arguments{
#   \item{filename, path}{The filename and the optional path of the image
#    file product.}
#   \item{...}{Additional arguments passed to @see "DevEvalProduct".}
# }
#
# \section{Fields}{
#  The following (virtual; calculate on-the-fly) fields are available:
#  \itemize{
#   \item \code{pathname}: the (relative) pathname of the image file, e.g. 'figures/foo,a,b.png'.  This can be used to \emph{link to} image files in for instance HTML and Markdown documents.
#   \item \code{path}, the (relative) path to the image file, e.g. 'figures/'
#   \item \code{filename}: the filename ("basename") of the image file, e.g. 'foo,a,b.png'
#   \item \code{fullname}: the fullname (the filename without the filename extension), e.g. 'foo,a,b'.  It is recommended to use this when including images in LaTeX documents that use the \code{\\usepackage{graphicx}} package.
#   \item \code{name}: the part of the fullname before the first comma, e.g. 'foo'
#   \item \code{tags}: the part of the fullname after the first comma, e.g. 'a,b'
#   \item \code{dataURI}: the Base64-encoded Data URI representation of the image, e.g. 'data:image/png;base64,iVBORw0KGgoAAAA...'.  This can be used to \emph{include} ("inlining") image files in for instance self-contained HTML and Markdown documents.
#   \item \code{data}: the character content of the image file.  This can be used to \emph{include} ("inlining") WebGL HTML-based image files in for instance self-contained HTML and Markdown documents.
#  }
# }
#
# \seealso{
#   In order to retrieve the Data URI, the \pkg{base64enc} package
#   must be installed.
# }
#
# @author
#
# @keyword internal
#*/###########################################################################
setConstructorS3("DevEvalFileProduct", function(filename=NULL, path=NULL, ...) {
  if (!is.null(filename)) {
    pathname <- Arguments$getReadablePathname(filename, path=path, mustExist=FALSE)
    path <- dirname(pathname)
    filename <- basename(pathname)
  } else {
    pathname <- NA_character_
  }

  this <- extend(DevEvalProduct(pathname, ...), "DevEvalFileProduct")

  # Infer 'type' from pathname?
  type <- getType(this)
  if (is.null(type) && !is.na(pathname)) {
    ext <- getExtension(this)
    type <- .devTypeName(ext)
    attr(this, "type") <- type
  }

  this
})

setMethodS3("as.character", "DevEvalFileProduct", function(x, ...) {
  getPathname(x, ...)
}, private=TRUE)


setMethodS3("getFullname", "DevEvalFileProduct", function(this, ...) {
  filename <- getFilename(this, ...)
  gsub("[.]([^.]*)$", "", filename)
})


###########################################################################/**
# @RdocMethod getPathname
# @aliasmethod getFilename
# @aliasmethod getPath
# @aliasmethod getExtension
# @alias getPathname
# @alias getFilename
# @alias getPath
# @alias getExtension
#
# @title "Gets the (relative) pathname, filename and path"
#
# \description{
#   @get "title".
# }
#
# \usage{
# @usage "getPathname,DevEvalFileProduct"
# @usage "getFilename,DevEvalFileProduct"
# @usage "getPath,DevEvalFileProduct"
# @usage "getExtension,DevEvalFileProduct"
# }
#
# \arguments{
#  \item{...}{Not used.}
# }
#
# \value{
#   Returns a @character string.
# }
#
# @author
#
# \seealso{
#   @seeclass
# }
#*/###########################################################################
setMethodS3("getPathname", "DevEvalFileProduct", function(this, relative=TRUE, ...) {
  pathname <- as.character(unclass(this), ...)
  if (is_nullfile(pathname)) return(pathname)
  if (relative) {
    pathname <- getRelativePath(pathname)
  } else {
    pathname <- getAbsolutePath(pathname)
  }
  pathname
})

setMethodS3("getFilename", "DevEvalFileProduct", function(this, ...) {
  basename(getPathname(this, ...))
})

setMethodS3("getPath", "DevEvalFileProduct", function(this, ...) {
  dirname(getPathname(this, ...))
})

setMethodS3("getExtension", "DevEvalFileProduct", function(this, ...) {
  filename <- getFilename(this, ...)
  gsub(".*[.]([^.]*)$", "\\1", filename)
}, private=TRUE)

setMethodS3("view", "DevEvalFileProduct", function(object, ...) {
  pathname <- object

  # WORKAROUND: browseURL('foo/bar.html', browser=NULL), which in turn
  # calls shell.exec('foo/bar.html'), does not work on Windows, because
  # the OS expects backslashes.  [Should shell.exec() convert to
  # backslashes?]  By temporarily setting the working directory to that
  # of the file, this works around this issue.
  # Borrowed from R.rsp. /HB 2014-09-19
  if (isFile(pathname)) {
    path <- dirname(pathname)
    pathname <- basename(pathname)
    opwd <- getwd()
    on.exit(setwd(opwd))
    setwd(path)
  }
  browseURL(pathname, ...)

  invisible(object)
})



#########################################################################/**
# @RdocMethod getMimeType
# @aliasmethod getMime
# @alias getMimeType
# @alias getMime
#
# @title "Gets the MIME type"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{default}{The value returned, if the MIME type could not be inferred.}
#   \item{...}{Not used.}
# }
#
# \value{
#  Returns a @character string.
# }
#
# @author
#
# \seealso{
#   @seeclass
# }
#*/#########################################################################
setMethodS3("getMimeType", "DevEvalFileProduct", function(this, default="", ...) {
  ext <- getExtension(this, ...)
  ext <- .devTypeName(ext)
  mimeTypes <- c(
    bmp="image/bmp",
    gif="image/gif",
    jpg="image/jpeg",
    pdf="application/pdf",
    eps="application/postscript",
    postscript="application/postscript",
    png="image/png",
    svg="image/svg+xml",
    tiff="image/tiff"
  )
  mime <- mimeTypes[ext]
  if (is.na(mime)) mime <- default
  mime
})

setMethodS3("getMime", "DevEvalFileProduct", function(this, ...) {
  getMimeType(this, ...)
}, private=TRUE)



#########################################################################/**
# @RdocMethod getDataURI
# @alias getDataURI
# @aliasmethod getData
# @alias getData
#
# @title "Gets content as a Base64-encoded data URI"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{mime}{The MIME type to be embedded in the data URI.}
#   \item{...}{Not used.}
# }
#
# \value{
#  Returns a @character string.
# }
#
# @author
#
# \seealso{
#   @seeclass
# }
#*/#########################################################################
setMethodS3("getDataURI", "DevEvalFileProduct", function(this, mime=getMimeType(this), ...) {
  dataURI(file=getPathname(this), mime=mime, encoding="base64")
})

setMethodS3("getData", "DevEvalFileProduct", function(this, mode=c("character", "raw"), ...) {
  # Argument 'mode':
  mode <- match.arg(mode)

  pathname <- this
  size <- file.info(pathname)$size
  if (mode == "character") {
    res <- readChar(con=pathname, nchars=size)
  } else if (mode == "raw") {
    res <- readBin(con=pathname, what=raw(), n=size)
  }
  res
}) # getData()

###########################################################################/**
# @RdocFunction capturePlot
#
# @title "Captures a plot such that it can be redrawn later/elsewhere"
#
# \description{
#   @get "title".
#
#   \emph{This feature is only supported in R (>= 3.3.0).}
# }
#
# @synopsis
#
# \arguments{
#   \item{expr}{The @expression of graphing commands to be evaluated.}
#   \item{envir}{The @environment where \code{expr} should be evaluated.}
#   \item{type}{The type of graphics device used in the background.
#    The choice should not matter since the result should be identical
#    regardless.  All graphics is captured but any output is also voided
#    by sending the output to a "null" file.}
#  \item{...}{Additional arguments passed to the graphics device.}
# }
#
# \value{
#   Returns a \code{recordedplot} object, which can be
#   \code{\link[grDevices]{replayPlot}()}:ed.  If replayed in an
#   interactive session, the plot is displayed in a new window.
#   For conveniency, the object is also replayed when \code{print()}:ed.
# }
#
# \details{
#   Note that plot dimensions/aspect ratios are not recorded.  This
#   means that one does not have to worry about those when recording
#   the plot.  Instead, they are specified when setting up the graphics
#   device(s) in which the recorded plot is replayed (see example).
# }
#
# \section{Replaying / replotting on a different architecture}{
#  In order to replay a \code{recordedplot} object, it has to be replayed
#  on an architecture that is compatible with the one who created the
#  object.
#  If this is not the case, then \code{\link[grDevices]{replayPlot}()}
#  will generate an \emph{Incompatible graphics state} error.
#  The \code{\link{as.architecture}()} function of this package tries
#  to coerce between different architectures, such that one can replay
#  across architectures using \code{replayPlot(as.architectures(g))}.
#  For convenience, the recorded plot returned by \code{capturePlot()}
#  is automatically coerced when \code{print()}:ed.
# }
#
# @examples "../incl/capturePlot.Rex"
#
# @author
#
# \seealso{
#   Internally \code{\link[grDevices]{recordPlot}()} is used.
# }
#
# \references{
#  [1] Paul Murrell et al.,
#      \emph{Recording and Replaying the Graphics Engine Display List},
#      December 2015.
#      \url{https://www.stat.auckland.ac.nz/~paul/Reports/DisplayList/dl-record.html}\cr
# }
#
# @keyword device
#*/###########################################################################
capturePlot <- function(expr, envir=parent.frame(), type=nulldev, ...) {
  if (getRversion() < "3.3.0") {
    throw("Insufficient R version. R.devices::capturePlot() requires R (>= 3.3.0): ", getRversion())
  }

  expr <- substitute(expr)

  ## Plot to /dev/null file (or NUL on Windows)
  type(nullfile(), ...)
  on.exit(dev.off())

  dev.control("enable")
  eval(expr, envir = envir, enclos = baseenv())
  g <- recordPlot()

  ## Record details of machine's architecture (helps troubleshooting)
  attr(g, "system") <- list(
    ostype=.Platform$OS.type,
    arch=R.version$arch,
    ptrsize=.Machine$sizeof.pointer,
    endian=.Platform$endian
  )

  class(g) <- c("RecordedPlot", class(g))
  
  g
}


#' Automatically replays a recorded plot
#'
#' This is identical to the \code{\link[grDevices:replayPlot]{print}()}
#' method available in \pkg{grDevices}, but if replaying the plot gives
#' an error it will also try to replay it after coercing the data structure
#' to match the architecture of the current machine.  This will make it
#' possible to, for instance, replay a plot generated on a 32-bit machine
#' on a 64-bit machine.
#' 
#' @param x A recorded plot of class \code{recordedplot}.
#'
#' @return Returns \code{x} invisibly.
#'
#' @seealso Internally, \code{\link{as.architecture}()} is used
#' to coerce to the current architecture.
#'
#' @export
#' @keywords internal
setMethodS3("print", "RecordedPlot", function(x, ...) {
  ## First, try without coercion
  res <- tryCatch({
    replayPlot(x)
  }, error = identity)

  ## If that didn't work, then try to coerce
  if (inherits(res, "simpleError")) {
    x <- as.architecture(x)
    replayPlot(x)
  }
  
  invisible(x)
}, private = TRUE)

## Look for existing generic functions also in imported namespaces.
## This will affect whether setGenericS3() creates a generic function
## or not.
options("R.methodsS3:checkImports:setGenericS3"=TRUE)

#########################################################################/**
# @RdocFunction eps
#
# @title "EPS graphics device"
#
# \description{
#  Device driver for Encapsulated Postscript. This driver is the same as
#  the postscript driver where some arguments have different default values.
# }
#
# @synopsis
#
# \arguments{
#   \item{file}{Default file name (pattern).}
#   \item{width, height}{The width and height of the figure.}
#   \item{horizontal}{If @FALSE, an horizontal EPS file is created,
#         otherwise a portrait file is created.}
#   \item{paper}{A @character string specifying the paper type. Overrides
#         the default of \code{postscript()}.}
#   \item{onefile}{Not used.}
#   \item{...}{Other arguments accepted by \code{postscript()}.}
# }
#
# \value{
#   A plot device is opened; nothing is returned.
# }
#
# \examples{\dontrun{
#   eps("foo.eps", width=7, height=7)
#
#   # is identical to
#
#   postscript("foo.eps", width=7, height=7, onefile=TRUE, horizontal=FALSE)
#
#   # and
#
#   dev.print(eps, "foo.eps", ...)
#
#   # is identical to
#
#   dev.print(postscript, "foo.eps", onefile=TRUE, horizontal=FALSE, paper="special", ...)
# }}
#
# \keyword{device}
#
# \seealso{
#   This is just a convenient wrapper for @see "grDevices::postscript"
#   with the proper arguments set to generate an EPS file.
#
#   It is recommended to use @see "toEPS" instead.
# }
#
# @author
#
# @keyword device
# @keyword internal
#*/#########################################################################
eps <- function(file="Rplot%03d.eps", width=7, height=7, horizontal=FALSE, paper="special", onefile=FALSE, ...) {
  onefile <- FALSE
  postscript(file=file, width=width, height=height, horizontal=horizontal, paper=paper, onefile=onefile, ...)
}

#' Parse a GPar structure from its raw byte format
#'
#' @param con A binary connection.
#' @param ptrsize Address pointer size - either 4L or 8L.
#'
#' @return A named list.
#'
#' @details
#' This is just an internal function used to parse the GPar byte sequence
#' that is part of an recordedplot object into a names list structure.
#'
#' @importFrom R.utils hpaste
#' @importFrom grDevices rgb
#' @keyword internal
parse_GPar <- local({
  hpaste <- R.utils::hpaste
  
  readBin2 <- function(con, what, size, n=1L, ...) {
    pos <- seek(con) + 1L  ## Index with offset 1 (not 0 as the file itself)
    pad <- switch(what, integer=pos %% 4, double=pos %% 8, pos %% 4)
    whatT <- switch(what, char="raw", color="raw", pad="raw", what)
    value <- readBin(con=con, what=whatT, size=size, n=n, ...)
    calls <- sys.calls()
    names <- sapply(calls, FUN=function(call) as.character(call[[1]]))
    names <- grep("parse_", names, value=TRUE)
    call <- names[2]
    pos2 <- seek(con)  ## Index with offset 1 (not 0 as the file itself)
    if (what == "char") {
      value <- rawToChar(value, multiple=FALSE)
    }
    valueT <- if (is.character(value)) sQuote(value) else value
    cat(sprintf("%s @ [%d,%d] (%d) => %s: {%s}\n", call, pos, pos2, pad, what, hpaste(valueT)))
    if (what == "double") {
  #    stopifnot(pad == 0)
    }
    if (what == "pad") {
      to_pad <- 8 - pos %% 8
  #    if(n != to_pad) stop(sprintf("pos=%d, to_pad=%d, n=%d", pos, to_pad, n))
    }
  #  attr(value, "pos") <- c(pos, pos2)
    value
  } ## readBin2()
  
  parse_unsigned_short <- function(con, n=1L) readBin2(con, what="integer", size=2L, signed=FALSE, n=n)
  parse_unsigned_int <- function(con, n=1L) {
    ## WORKAROUND: readBin(.., size=4L, signed=FALSE) is not supported
    value <- readBin2(con, what="integer", size=4L, signed=TRUE, n=n)
    if (value < 0) value <- value + as.integer(2^24)
    attr(value, "note") <- "should be: unsigned int"
    value
  }
  
  parse_byte <- function(con, n=1L) readBin2(con, what="integer", size=1L, signed=FALSE, n=n)
  
  parse_pad <- function(con, n=1L) {
    pos <- seek(con)
    value <- readBin2(con, what="pad", size=1L, n=n)
    stopifnot(all(value == 0))
    attr(value, "pos") <- pos
    value
  }
  
  parse_int <- function(con, n=1L) readBin2(con, what="integer", size=4L, signed=TRUE, n=n)
  
  parse_double <- function(con, size=8L, n=1L) readBin2(con, what="double", size=size, signed=TRUE, n=n)
  
  parse_unsigned_char <- function(con, n=1L) readBin2(con, what="character", size=1L, signed=FALSE, n=n)
  
  parse_char <- function(con, n=1L) readBin2(con, what="char", size=1L, n=n)
  
  nenum <- 0L
  parse_enum <- function(con) {
    nenum <<- nenum + 1L
    parse_int(con)
  }
  
  parse_Rboolean <- function(con) {
    value <- parse_enum(con)
    value <- as.logical(value)
    value
  }
  
  parse_rcolor <- function(con) {
    value <- readBin2(con, what="color", size=1L, n=4L)
    do.call(rgb, args=c(as.list(as.integer(value)), maxColorValue=255L))
  }
  
  parse_R_GE_lineend <- function(con) parse_enum(con)-1L
  
  parse_R_GE_linejoin <- function(con) parse_enum(con)-1L
  
  parse_GUnit <- function(con) parse_enum(con)
  
  parse_GTrans <- function(con) {
    list(
      ax = parse_double(con),
      ay = parse_double(con),
      bx = parse_double(con),
      by = parse_double(con)
    )
  }

  MAX_LAYOUT_ROWS <- 200L
  MAX_LAYOUT_COLS <- 200L
  MAX_LAYOUT_CELLS <- 10007L
  
  function(con, ptrsize=4L) {
    parse_pad64 <- function(con) raw(0L)
    if (ptrsize == 8L) {
      parse_pad64 <- function(con) parse_pad(con, n=4L)
    }
  
    list(  
      state = parse_int(con),             ## pos  0-3   0L
      valid = parse_Rboolean(con),        ## pos  4-7   0L
      adj = parse_double(con),            ## pos  8-15  0.1
      ann = parse_Rboolean(con),          ## pos 16-19  0L
      bg = parse_rcolor(con),             ## pos 20-23  {aa,bb,cc,ff}
      bty = parse_char(con),              ## pos 24     'n'
      PAD = parse_pad(con, n=3),          ## pos 25-27
      PAD64 = parse_pad64(con),
      cex_ = parse_double(con),           ## pos 32-39  1.0
      lheight = parse_double(con),        ## pos 40-47  2.1
      col = parse_rcolor(con),            ## pos 48-51  {11,22,33,44}
      PAD64 = parse_pad64(con),
      crt = parse_double(con),            ## pos 56-63  0.8
      din = parse_double(con, n=2),       ## pos 64-79  (0,0)
      err = parse_int(con),               ## pos 80-83  0L
      fg = parse_rcolor(con),             ## pos 84-87  {cc,dd,ee,ff}
      family = parse_char(con, n=201),    ## pos 88-288 'serif'
      PAD = parse_pad(con, n=3),
      font = parse_int(con),              ## pos 292-295 1L
      gamma = parse_double(con),          ## pos 296-303 1.0
      lab = parse_int(con, n=3),          ## pos 304-315 (3L,3L,5L)
      las = parse_int(con),               ## pos 316     2L 
      lty = parse_byte(con, n=4),         ## pos 320-323
      PAD64 = parse_pad64(con),
      lwd = parse_double(con),            ## pos 328-335      3.14
      lend = parse_R_GE_lineend(con),     ## pos 336          1L
      ljoin = parse_R_GE_linejoin(con),   ## pos 340          0L
      lmitre = parse_double(con),         ## pos 344-351      2.3
      mgp = parse_double(con, n=3),       ## pos 352,360,368 (2.5,1.5,0.5)
      mkh = parse_double(con),            ## pos 376-383      0.001
      pch = parse_int(con),               ## pos 384-387      2L
      PAD64 = parse_pad64(con),
      ps = parse_double(con),             ## pos 392          5.0
      smo = parse_int(con),               ## pos 400          2L ???
      PAD64 = parse_pad64(con),           ## pos 404          0L ???
      srt = parse_double(con),            ## pos 408          78.9
      tck = parse_double(con),            ## pos 416          0.6
      tcl = parse_double(con),            ## pos 424          0.7
      xaxp = parse_double(con, n=3),      ## pos 432,440,448  (0.1,0.2,4.0)
      xaxs = parse_char(con),             ## pos 456          'i'
      xaxt = parse_char(con),             ## pos 457          'n'
      PAD = parse_pad(con, n=2),
      xlog = parse_Rboolean(con),         ## pos 451-454      1L
      xpd = parse_int(con),               ## pos 455-458      1L
      oldxpd = parse_int(con),
      yaxp = parse_double(con, n=3),        
      yaxs = parse_char(con),                
      yaxt = parse_char(con),                
      PAD = parse_pad(con, n=2),
      ylog = parse_Rboolean(con),        
      cexbase = parse_double(con),        
      cexmain = parse_double(con),        
      cexlab = parse_double(con),        
      cexsub = parse_double(con),        
      cexaxis = parse_double(con),        
      fontmain = parse_int(con),        
      fontlab = parse_int(con),        
      fontsub = parse_int(con),        
      fontaxis = parse_int(con),        
      colmain = parse_rcolor(con),        
      collab = parse_rcolor(con),        
      colsub = parse_rcolor(con),        
      colaxis = parse_rcolor(con),        
      layout = parse_Rboolean(con),       ## pos 576  1
      numrows = parse_int(con),           ## pos 580  3
      numcols = parse_int(con),           ## pos 584  2
      currentFigure = parse_int(con),     ## pos 588  7
      lastFigure = parse_int(con),        ## pos 592  6
      PAD64 = parse_pad64(con),
      heights = parse_double(con, n=MAX_LAYOUT_ROWS),
      widths = parse_double(con, n=MAX_LAYOUT_COLS),
      cmHeights = parse_int(con, n=MAX_LAYOUT_ROWS),
      cmWidths = parse_int(con, n=MAX_LAYOUT_COLS),
      order = parse_unsigned_short(con, n=MAX_LAYOUT_CELLS),
      rspct = parse_int(con),                
      respect = parse_unsigned_char(con, n=MAX_LAYOUT_CELLS),
      mfind = parse_int(con),                
      PAD = parse_pad(con, n=3),
      fig = parse_double(con, n=4),       ## pos 35432
      fin = parse_double(con, n=2),       ## pos 35464, 35472 (3.5, 2.333)
      fUnits = parse_GUnit(con),          ## pos 35480        6L
      PAD64 = parse_pad64(con),
      plt = parse_double(con, n=4),       ## pos 35488, 35496, 35504, 35512
      pin = parse_double(con, n=2),       ## pos 35520, 35528 (1.77893, 1.77893
      pUnits = parse_GUnit(con),                
      defaultFigure = parse_Rboolean(con),        
      defaultPlot = parse_Rboolean(con),        
      PAD64 = parse_pad64(con),
      mar = parse_double(con, n=4),       ## pos 35552, 35560, 35568, 35576
      mai = parse_double(con, n=4),       ## pos 35584, 35592, 35600, 35608
      mUnits = parse_GUnit(con),          ## pos 35616
      PAD64 = parse_pad64(con),
      mex = parse_double(con),            ## pos 35624  1.34
      oma = parse_double(con, n=4),        
      omi = parse_double(con, n=4),        
      omd = parse_double(con, n=4),        
      oUnits = parse_GUnit(con),        
      pty = parse_char(con),                
      PAD = parse_pad(con, n=3),
      usr = parse_double(con, n=4),        
      logusr = parse_double(con, n=4),
      new = parse_Rboolean(con),        
      devmode = parse_int(con),        
      xNDCPerChar = parse_double(con),        
      yNDCPerChar = parse_double(con),        
      xNDCPerLine = parse_double(con),        
      yNDCPerLine = parse_double(con),        
      xNDCPerInch = parse_double(con),        
      yNDCPerInch = parse_double(con),        
      fig2dev = parse_GTrans(con),        
      inner2dev = parse_GTrans(con),        
      ndc2dev = parse_GTrans(con),        
      win2fig = parse_GTrans(con),        
      scale = parse_double(con)
    )
  } ## parse_GPar()
}) ## local()


as.GPar <- function(x, ...) {
  gpar <- R.devices:::gpar
  stopifnot(inherits(x, "recordedplot"))
  arch <- architecture(x)
  gpar_raw <- gpar(x)
  con <- rawConnection(gpar_raw, open="rb")
  on.exit(close(con))
  gpar <- parse_GPar(con, ptrsize=arch$ptrsize)
  gpar <- gpar[names(gpar) != "PAD"]
  gpar <- gpar[names(gpar) != "PAD64"]
  gpar
} ## as.GPar()


readGPar <- function(pathname) {
  n <- file.size(pathname)
  raw <- base::readBin2(pathname, what="raw", n=n)
  ptrsize <- if (n == 35992L) 8L else if (n == 35956L) 4L else stop(n)
  con <- rawConnection(raw, open="rb")
  on.exit(close(con))
  parse_GPar(con, ptrsize=ptrsize)
} ## readGPar()


writeGPar <- function(g, pathname=sprintf("gpar,%s.bin", attr(g, "R.version")$arch)) {
  gpar <- R.devices:::gpar
  writeBin(as.raw(gpar(g)), con=pathname)
  pathname
} ## writeGPar()

#########################################################################/**
# @RdocPackage R.devices
#
# \description{
#   @eval "getDescription(R.devices)"
# }
#
# \section{To get started}{
#   \itemize{
#    \item Vignette '\href{../doc/index.html}{R.devices overview}'
#    \item @see "toEPS", @see "toPDF", @see "toPNG", ... - evaluate
#          graphics code and atomically save plot to a figure file.
#   }
# }
#
# \section{How to cite this package}{
#   To cite this package, please use:
# \preformatted{
# @eval "paste(capture.output(toBibtex(citation(package='R.devices'))), collapse='\n')"
# }
# }
#
# \author{
#  @eval "packageDescription('R.devices')$Author".
# }
#
# \section{License}{
#  @eval "packageDescription('R.devices')$License".
# }
#*/#########################################################################

###########################################################################/**
# @RdocFunction jpeg2
#
# @title "A JPEG device for Bitmap Files via GhostScript"
#
# \description{
#   @get "title".
# }
#
# @synopsis
#
# \arguments{
#  \item{filename}{The name of the file to be produced.}
#  \item{width, height}{The width and height (in pixels) of the result image.}
#  \item{res}{The resolution of the image.}
#  \item{type}{The output type. See @see "grDevices::dev2bitmap" for details.}
#  \item{...}{Additional arguments passed to \code{bitmap()}.}
# }
#
# \value{
#   Returns nothing.
# }
#
# @author
#
# @keyword device
# @keyword internal
#*/###########################################################################
jpeg2 <- function(filename, width=480, height=480, res=144, type="jpeg", ...) {
  bitmap(file=filename, type=type, width=width/res, height=height/res, res=res, ...)
}

###########################################################################/**
# @RdocFunction devOptions
# @alias getDevOption
#
# @title "Gets the default device options"
#
# \description{
#  @get "title" as given by predefined devices options adjusted for
#  the default arguments of the device function.
# }
#
# @synopsis
#
# \arguments{
#   \item{type}{A @character string or a device @function specifying
#      the device to be queried.  May also be a @vector, for querying
#      device options for multiple devices.}
#   \item{custom}{If @TRUE, also the default settings specific to this
#      function is returned. For more details, see below.}
#   \item{special}{A @logical.  For more details, see below.}
#   \item{inherits}{If @TRUE, the global option is used if the
#      type-specific is not set (or @NULL).}
#   \item{drop}{If @TRUE and only one device type is queried, then
#      a @list is returned, otherwise a @matrix.}
#   \item{options}{Optional named @list of settings.}
#   \item{...}{Optional named arguments for setting new defaults.
#      For more details, see below.}
#   \item{reset}{If @TRUE, the device options are reset to R defaults.}
# }
#
# \value{
#   If \code{drop=TRUE} and a single device is queries, a named @list is
#   returned, otherwise a @matrix is returned.
#   If a requested device is not implemented/supported on the current system,
#   then "empty" results are returned.
#   If options were set, that is, if named options were specified via
#   \code{...}, then the list is returned invisibly, otherwise not.
# }
#
# \details{
#  If argument \code{special} is @TRUE, then the 'width' and 'height'
#  options are adjusted according to the rules explained for
#  argument 'paper' in @see "grDevices::pdf", @see "grDevices::postscript",
#  and @see "grDevices::xfig".
# }
#
# \section{Setting new defaults}{
#  When setting device options, the \code{getOption("devOptions")[[type]]}
#  option is modified.  This means that for such options to be effective,
#  any device function needs to query also such options, which for instance
#  is done by @see "devNew".
#
#  Also, for certain devices (eps, pdf, postscript, quartz, windows and x11),
#  builtin R device options are set.
# }
#
# @examples "../incl/devOptions.Rex"
#
# @author
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devOptions <- function(type=NULL, custom=TRUE, special=TRUE, inherits=FALSE, drop=TRUE, options=list(), ..., reset=FALSE) {
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Local setups
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  devList <- list(
    ## Global device options
    "*"=NA_character_
  )
  ## All known and supported graphics devices on this system
  devList <- c(devList, devAll())
  knownTypes <- names(devList)


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Local functions
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (.Platform$OS.type == "windows") {
    # To please R CMD check
    windows.options <- NULL; rm(list="windows.options")
    x11.options <- grDevices::windows.options
    X11.options <- x11.options
  }


  # Global options; only used for its ability to reset them
  "*.options" <- function(..., reset=FALSE) {
    if (reset) {
      devOptions(type="*", sep=",", path="figures", field=NULL, force=TRUE)
    }
  }


  # A template for a dummy device options function.
  getNnnOptions <- function(type, fallbacks=FALSE, ...) {
    optList <- list(
      "*"="*.options",
      eps="ps.options",
      jpeg2="ps.options",
      pdf="pdf.options",
      png2="ps.options",
      postscript="ps.options",
      quartz="quartz.options",
      windows="windows.options",
      x11="X11.options",
      X11="x11.options"
    )

    if (fallbacks) {
      optList <- c(optList, list(
        "CairoPDF"="pdf.options",
        "CairoPS"="ps.options",
        "CairoX11"="x11.options"
      ))
    }

    dummy <- function(...) { list(); }

    if (!is.element(type, names(optList))) {
      return(dummy)
    }

    key <- optList[[type]]

    # Sanity check
    stop_if_not(length(key) == 1L)

    # Does the nnn.function() already exists?
    envir <- getNamespace("grDevices")
    if (exists(key, envir=envir, mode="function")) {
      fcn <- get(key, envir=envir, mode="function")
      return(fcn)
    } else if (exists(key, mode="function")) {
      fcn <- get(key, mode="function")
      return(fcn)
    }

    # If not, create either a real one or a dummy one...
    typeC <- capitalize(type)
    keyE <- sprintf(".%senv", typeC)
    if (exists(keyE, envir=envir, mode="environment")) {
      # A real one
      envir <- get(keyE, envir=envir, mode="environment")
      fcn <- function(..., reset=FALSE) {
        keyO <- sprintf(".%s.Options", typeC)
        opts <- get(keyO, envir=envir, mode="list")
        if (reset) {
          keyD <- sprintf("%s.default", keyO)
          opts <- get(keyD, envir=envir, mode="list")
          assign(keyD, value=opts, envir=envir)
        }
        opts
      }
    } else {
      # A dummy
      fcn <- dummy
    }

    fcn
  } # getNnnOptions()


  # See argument 'paper' in help("pdf"), help("postscript"), and
  # help("xfig").
  getSpecialDimensions <- function(options=list(), sizes=names(paperSizes), ...) {
    paperSizes <- list(
      a4        = c( 8.27, 11.69),
      a4r       = c(11.69,  8.27),
      executive = c( 7.25, 10.5 ),
      legal     = c( 8.5 , 14   ),
      letter    = c( 8.5 , 11   ),
      USr       = c(11   , 8.5  )
    )

    paper <- tolower(options$paper)
    if (paper == "default") {
      paper <- getOption("papersize", "a4")
    }

    if (paper != "special") {
      paperSizes <- paperSizes[sizes]
      dim <- paperSizes[[paper]]

      # Replace "special" 0:s with NA:s, to indicate they are missing
      dim[dim == 0L] <- NA_real_
    } else {
      dim <- c(options$width, options$height)
    }

    dim
  } # getSpecialDimensions()


  getDevOptions <- function(type, ...) {
    opts <- getOption("devOptions", list())
    opts <- opts[[type]]
    if (is.null(opts)) {
      opts <- list()
    }
    opts
  } # getDevOptions()


  setDevOptions <- function(.type, ..., reset=FALSE) {
    devOpts <- getOption("devOptions", list())
##    str(list(.type=.type, devOpts=devOpts))
    oopts <- opts <- devOpts[[.type]]
    if (is.null(opts)) opts <- list()
    # Sanity check
    stop_if_not(is.list(opts))
##    str(list(opts=opts))
    if (reset) {
      opts <- NULL
    } else {
      args <- list(...)
      if (length(args) > 0L) {
        for (key in names(args)) {
          value <- args[[key]]
##          str(list(key=key, value=value))
          opts[[key]] <- value
        }
      }
    }
    devOpts[[.type]] <- opts
    options(devOptions=devOpts)
    invisible(oopts)
  } # setDevOptions()


  getDeviceFunctions <- function(type, default=function() {}, ...) {
    # Find all device functions.  If non-existent, return the default
    devs <- devList[[type]]

    parts <- strsplit(devs, split="::", fixed=TRUE)
    devs <- lapply(parts, FUN=function(s) {
      if (length(s) > 1L) {
        # Device package may not be installed, e.g. Cairo and JavaGD.
        envir <- tryCatch({
          # Attach it as well, so that the device function is found
          # when needed inside devNew() at do.call().
          suppressWarnings({
            if (require(s[1L], character.only=TRUE, quietly=TRUE)) {
              # Return namespace
              getNamespace(s[1L])
            } else {
              emptyenv()
            }
          })
        }, error = function(ex) emptyenv())
        s <- s[-1L]
      } else {
        envir <- as.environment(-1L)
      }

      if (exists(s[1L], envir=envir, mode="function")) {
        get(s[1L], envir=envir, mode="function")
      } else {
        default
      }
    })

    devs
  } # getDeviceFunctions()


  findDeviceFunction <- function(fcn, ...) {
    types <- names(devList)

    ## (i) Scan primary device functions
    for (type in types) {
      devs <- getDeviceFunctions(type, default=NULL)
      if (identical(fcn, devs[[1]])) return(type)
    }

    ## (ii) Scan non-primary device functions
    for (type in types) {
      devs <- getDeviceFunctions(type, default=NULL)
      for (dev in devs[-1]) {
        if (identical(fcn, dev)) return(type)
      }
    }

    "<function>"
  } # findDeviceFunction()


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Validate arguments
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Argument 'type':
  if (is.null(type)) {
    type <- names(devList)
  }

  # Argument 'options':
  if (!is.list(options)) {
    throw("Argument 'options' must be a list: ", class(options)[1L])
  }
  nopts <- length(options)
  if (nopts > 0L && is.null(names(options))) {
    throw("Argument 'options' must be a named list.")
  }

  # Argument 'inherits':
  inherits <- as.logical(inherits)

  # Argument 'reset':
  reset <- as.logical(reset)

  # Additional arguments
  args <- list(...)
  nargs <- length(args)
  if (nargs > 0L && is.null(names(args))) {
    throw("Optional ('...') arguments must be named.")
  }

  # Append/overwrite 'options' with the named argument, e.g. field=NULL
  for (key in names(args)) {
    options[key] <- list(args[[key]])
  }
  nopts <- length(options)
  # Not needed anymore
  args <- nargs <- NULL

  # Argument 'type':
  if (missing(type) || length(type) == 0L) {
    if (nopts > 0L) {
      throw("Cannot set device options. Argument 'type' is missing or NULL. Should be one of: ", paste(sQuote(knownTypes), collapse=", "))
    }

    res <- devOptions(type=knownTypes, custom=custom, special=special, inherits=inherits, drop=drop, reset=reset)
    if (reset) {
      return(invisible(res))
    } else {
      return(res)
    }
  } else if (is.character(type)) {
    # Expand by regexp matching, iff any
    type <- .devTypeName(type, pattern=TRUE, knownTypes=knownTypes)
  }


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Vectorized call?
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (length(type) > 1L) {
    if (nopts > 0L) {
      throw("Cannot set device options for more than one devices at the time: ", hpaste(sQuote(type), collapse=", "))
    }

    # Support vector of 'type':s
    types <- type
    res <- lapply(types, FUN=devOptions, inherits=inherits, drop=TRUE)
    fields <- lapply(res, FUN=function(opts) names(opts))
    fields <- unique(unlist(fields, use.names=FALSE))
    opts <- lapply(res, FUN=function(x) x[fields])
    opts <- unlist(opts, use.names=FALSE, recursive=FALSE)
    dim(opts) <- c(length(fields), length(types))
    dimnames(opts) <- list(fields, types)
    opts <- t(opts)
    ##class(opts) <- c("DeviceOptions", class(opts))
    return(opts)
  }


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # A single device at this point
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (is.function(type)) {
    # Try to find name of device function
    type <- findDeviceFunction(fcn=type)
  }
  if (is.character(type)) {
    if (!is.element(type, knownTypes)) {
      throw(sprintf("Device type %s is not known/supported on this operating system/platform. Supported device types are: %s", sQuote(type), paste(sQuote(setdiff(knownTypes, "*")), collapse=", ")))
    }
  }

  if (!is.element(type, names(devList))) {
    throw("Cannot query/modify device options. Unknown device: ", type)
  }

  if (inherits) {
    throw("Argument inherits=TRUE is not yet supported.")
  }


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Locate the nnn.options() function for this type of device
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  nnn.options <- getNnnOptions(type)


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Locate the list of device functions used by this type of device
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # If non-existent, return a dummy
  devs <- getDeviceFunctions(type, default=function() {})


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Reset?
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (reset) {
    res <- devOptions(type=type, special=special, drop=drop, reset=FALSE)
    setDevOptions(type, reset=TRUE)

    # Only for certain devices...
    nnn.options(reset=TRUE)

    return(invisible(res))
  }


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Assign user arguments, iff possible
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (nopts > 0L) {
    do.call(setDevOptions, args=c(list(.type=type), options))

    # Only for certain devices...
    # (a) Try to set all options at once
    ok <- tryCatch({
      do.call(nnn.options, args=options)
      TRUE
    }, error = function(ex) FALSE)

    # (b) Otherwise, try to set them one by one
    if (!ok) {
      for (kk in seq_along(options)) {
        res <- tryCatch({
          do.call(nnn.options, args=options[kk])
        }, error = function(ex) {})
      }
    }
  }


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Get builtin device options, if available
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Only for certain devices...
  nnn.options <- getNnnOptions(type, fallbacks=TRUE)
  opts <- nnn.options()


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Get (nested) device formals
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  defArgs <- lapply(rev(devs), FUN=formals)
  defArgs <- Reduce(c, defArgs)
  # Drop '...'
  defArgs <- defArgs[names(defArgs) != "..."]
  # Drop missing
  # Drop overridden values
  defArgs <- defArgs[!duplicated(names(defArgs), fromLast=TRUE)]


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Merge arguments that either are not in the predefined set of
  # device options, or ones that replaced the default value.
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  optsT <- c(opts, defArgs)


  # Include custom options specific to devOptions()?
  if (custom) {
    devOpts <- getDevOptions(type)
    optsT <- c(optsT, devOpts)
  }


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Build the concensus of all options
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # (a) Keep all non-duplicated options
  dups <- duplicated(names(optsT))
  idxsA <- which(!dups)
  optsA <- optsT[idxsA]

  # (b) Among duplicates, keep those with values
  idxsB <- which(dups)
  idxsB <- idxsB[!sapply(optsT[idxsB], FUN=is.symbol)]
  optsB <- optsT[idxsB]

  opts <- c(optsA, optsB)

  # Drop overridden values
  opts <- opts[!duplicated(names(opts), fromLast=TRUE)]


  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Adjust for special cases?
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if (special) {
    if (is.element(type, c("eps", "postscript", "CairoPS"))) {
      sizes <- c("a4", "executive", "legal", "letter")
      dim <- getSpecialDimensions(opts, sizes)
    } else if (type == "xfig") {
      sizes <- c("a4", "legal", "letter")
      dim <- getSpecialDimensions(opts, sizes)
    } else if (is.element(type, c("pdf", "CairoPDF"))) {
      sizes <- c("a4", "a4r", "executive", "legal", "letter", "USr")
      dim <- getSpecialDimensions(opts, sizes)
    } else {
      dim <- NULL
    }

    if (!is.null(dim)) {
      opts$width <- dim[1L]
      opts$height <- dim[2L]
    }
  }

  # Return a list?
  if (!drop) {
    names <- names(opts)
    dim(opts) <- c(1L, length(opts))
    dimnames(opts) <- list(type, names)
    ## Possibly, class(opts) <- c("DeviceOptions", class(opts))
  }

  # Return invisibly?
  if (nopts > 0L) {
    invisible(opts)
  } else {
    opts
  }
} # devOptions()


getDevOption <- function(type, name, default=NULL, inherits=TRUE, ..., old=TRUE) {
  # BACKWARD COMPATIBILITY:
  if (isTRUE(old)) old <- sprintf("devEval/args/%s", name)
  if (is.character(old)) {
    .importOldGlobalOption(name, from=old, remove=TRUE)
  }

  # Get options
  if (is.character(type) && (length(type) != 1L || type != "*")) {
    value <- devOptions(type=type, ...)[[name]]
  } else {
    value <- NULL
  }

  # Fallback to global options?
  if (is.null(value) && inherits) {
    value <- devOptions(type="*", ...)[[name]]
  }

  # Use default value?
  if (is.null(value)) value <- default

  value
} # getDevOption()


# BACKWARD COMPATIBILITY for old-style R options().
.importOldGlobalOption <- function(name, from, remove=TRUE, ...) {
  # Nothing to do?
  if (!is.element(from, names(options()))) return()

  # (a) Get old-style option value
  opt <- options(from)

  # (b) Assign to global device options instead
  names(opt) <- name
  do.call(devOptions, args=list(type="*", options=opt))

  # (c) Delete old-style option and never look back
  if (remove) setOption(from, NULL)
} # .importOldGlobalOption()

#########################################################################/**
# @RdocFunction nulldev
#
# @title "A \"null\" graphics device voiding all output"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{file}{The output file (ignored; forced to its default).}
#   \item{...}{All other arguments are also ignored.}
# }
#
# \value{
#   A plot device is opened; nothing is returned.
# }
#
# \details{
#   The null graphics device opens a regular \link[grDevices]{pdf} device
#   and directs its output to the null file, which is \code{/dev/null} unless
#   on Windows where it is \code{NUL}.
# }
#
# \examples{\dontrun{
#   nulldev()
#   plot(1:3)
#   dev.off()
# }}
#
# \keyword{device}
#
# @author
#
# @keyword device
# @keyword internal
#*/#########################################################################
nulldev <- function(file = nullfile(), ...) {
  type <- getOption("R.devices.nulldev", "pdf")
  do.call(type, args = list(nullfile()))
}

###########################################################################/**
# @RdocFunction devIsOpen
#
# @title "Checks if zero or more devices are open or not"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{which}{An index (@numeric) @vector or a label (@character) @vector.}
#   \item{...}{Not used.}
# }
#
# \value{
#   Returns a named @logical @vector with @TRUE if a device is open,
#   otherwise @FALSE.
# }
#
# @examples "../incl/deviceUtils.Rex"
#
# @author
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devIsOpen <- function(which=dev.cur(), ...) {
  # Nothing to do?
  if (length(which) == 0L) {
    res <- logical(0L)
    names(res) <- character(0L)
    return(res)
  }

  devList <- .devList()
  devs <- devList[which]

  labels <- names(devs)
  isOpen <- sapply(devs, FUN=function(dev) !is.null(dev) && nzchar(dev))
  isOpen <- isOpen & !is.na(labels)
  isOpen
} # devIsOpen()





###########################################################################/**
# @RdocFunction devList
#
# @title "Lists the indices of the open devices named by their labels"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{interactiveOnly}{If @TRUE, only open interactive/screen devices
#     are returned.}
#   \item{dropNull}{If @TRUE, the "null" device (device index 1) is
#     not returned.}
#   \item{...}{Not used.}
# }
#
# \value{
#   Returns a named @integer @vector.
# }
#
# @author
#
# \seealso{
#   \code{\link[grDevices:dev]{dev.list}()}.
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devList <- function(interactiveOnly=FALSE, dropNull=TRUE, ...) {
  devList <- .devList()

  # Return only opened devices
  isOpen <- sapply(devList, FUN=function(dev) (dev != ""))
  names(isOpen) <- names(devList)
  idxs <- which(isOpen)

  # Exclude the "null" device?
  if (dropNull) {
    idxs <- idxs[-1L]
  }

  # Include only interactive devices?
  if (interactiveOnly) {
    types <- unlist(devList[idxs])
    keep <- devIsInteractive(types)
    idxs <- idxs[keep]
  }

  idxs
} # devList()



###########################################################################/**
# @RdocFunction devGetLabel
#
# @title "Gets the labels of zero or more devices"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{which}{An index (@numeric) @vector or a label (@character) @vector.}
#   \item{...}{Not used.}
# }
#
# \value{
#   Returns a @character @vector.
#   If a device does not exist, an error is thrown.
# }
#
# @author
#
# \seealso{
#   @see "devSetLabel" and @see "devList".
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devGetLabel <- function(which=dev.cur(), ...) {
  devList <- devList(dropNull=FALSE)
  if (is.numeric(which)) {
    devs <- devList[match(which, devList)]
  } else {
    devs <- devList[which]
  }
  labels <- names(devs)
  unknown <- which[is.na(labels)]
  if (length(unknown) > 0L) {
    known <- names(devList(dropNull=FALSE))
    if (length(known) == 0L) known <- "<none>"
    throw(sprintf("Cannot get device label. No such device: %s (known devices: %s)", paste(sQuote(unknown), collapse=", "), paste(sQuote(known), collapse=", ")))
  }
  labels
} # devGetLabel()



###########################################################################/**
# @RdocFunction devSetLabel
#
# @title "Sets the label of a device"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{which}{An index (@numeric) or a label (@character).}
#   \item{label}{A @character string.}
#   \item{...}{Not used.}
# }
#
# \value{
#   Returns nothing.
# }
#
# @author
#
# \seealso{
#   @see "devGetLabel" and @see "devList".
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devSetLabel <- function(which=dev.cur(), label, ...) {
  # Argument 'which':
  if (length(which) != 1L) {
    throw("Argument 'which' must be a scalar: ", paste(which, collapse=", "))
  }

  devList <- .devList()
  if (is.numeric(which)) {
    idx <- which
  } else {
    idx <- .devListIndexOf(which)
  }

  # Unknown devices?
  if (devList[[idx]] == "") {
    known <- names(devList(dropNull=FALSE))
    if (length(known) == 0L) known <- "<none>"
    throw(sprintf("Cannot set device label. No such device: %s (known devices: %s)", paste(sQuote(which), collapse=", "), paste(sQuote(known), collapse=", ")))
  }

  # Update the label
  if (is.null(label))
    label <- ""
  names(devList)[idx] <- label

  assign(".Devices", devList, envir=baseenv())
}





###########################################################################/**
# @RdocFunction devSet
#
# @title "Activates a device"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{which}{An index (@numeric) or a label (@character).
#     If neither, then a label corresponding to the checksum of
#     \code{which} as generated by @see "digest::digest" is used.}
#   \item{...}{Not used.}
# }
#
# \value{
#   Returns what \code{\link[grDevices:dev]{dev.set}()} returns.
# }
#
# @author
#
# \seealso{
#   @see "devOff" and @see "devDone".
#   Internally, \code{\link[grDevices:dev]{dev.set}()} is used.
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devSet <- function(which=dev.next(), ...) {
  args <- list(...)

  # Argument 'which':
  if (!is.numeric(which) || length(which) != 1L) {
    if (length(which) != 1L || !is.character(which)) {
      # To please R CMD check
      requireNamespace("digest") || throw("Package not loaded: digest")
      which <- digest::digest(which)
    }

    if (is.character(which)) {
      args$label <- which
      idx <- .devListIndexOf(which, error=FALSE)
      # If not existing, open the next available one
      if (is.na(idx)) {
        which <- .devNextAvailable()
      } else {
        devList <- devList(dropNull=FALSE)
        which <- devList[[idx]]
      }
    }
  }

  # Argument 'which':
  if (length(which) != 1L) {
    throw("Argument 'which' must be a scalar: ", paste(which, collapse=", "))
  }

  if (which < 2L || which > 63L) {
    throw("Cannot set device: Argument 'which' is out of range [2,63]: ", which)
  }

  if (devIsOpen(which)) {
    # Active already existing device
    return(dev.set(which))
  }

  # Identify set devices that needs to be opened inorder for
  # the next device to get the requested index
  if (which == 2L) {
    toBeOpened <- c()
  } else {
    toBeOpened <- setdiff(2:(which-1L), dev.list())
  }

  len <- length(toBeOpened)
  if (len > 0L) {
    tempfiles <- sapply(toBeOpened, FUN=function(...) tempfile())

    # Make sure to close all temporary devices when exiting function
    on.exit({
      for (kk in seq_along(toBeOpened)) {
        dev.set(toBeOpened[kk])
        dev.off()
        if (file.exists(tempfiles[kk])) file.remove(tempfiles[kk])
      }
    }, add=TRUE)

    # Create a dummy temporary postscript device (which is non-visible)
    for (kk in seq_along(toBeOpened)) {
      postscript(file=tempfiles[kk])
    }
  }

  # Open the device
  res <- do.call(devNew, args=args)

  invisible(res)
} # devSet()




###########################################################################/**
# @RdocFunction devOff
#
# @title "Closes zero or more devices"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{which}{An index (@numeric) @vector or a label (@character) @vector.}
#   \item{...}{Not used.}
# }
#
# \value{
#   Returns \code{\link[grDevices:dev]{dev.cur}()}.
# }
#
# @author
#
# \seealso{
#   @see "devDone".
#   Internally, \code{\link[grDevices:dev]{dev.off}()} is used.
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devOff <- function(which=dev.cur(), ...) {
  # Nothing to do?
  if (length(which) == 0L) return(dev.cur())

  # Only close each device once
  which <- unique(which)
  if (length(which) > 1L) {
    lapply(which, FUN=devOff)
    return(dev.cur())
  }

  # Nothing to do?
  if (!devIsOpen(which)) {
    return(dev.cur())
  }

  # Identify device
  which <- devSet(which)

  # Reset the label
  devSetLabel(which, label=NULL)

  # Close device
  dev.off(which)

  return(dev.cur())
} # devOff()




###########################################################################/**
# @RdocFunction devDone
#
# @title "Closes zero or more open devices except screen (interactive) devices"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{which}{An index (@numeric) @vector or a label (@character) @vector.}
#   \item{...}{Not used.}
# }
#
# \value{
#   Returns (invisibly) \code{\link[grDevices:dev]{dev.cur}()}.
# }
#
# @author
#
# \seealso{
#   @see "devOff".
#   @see "grDevices::dev.interactive".
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devDone <- function(which=dev.cur(), ...) {
  # Nothing to do?
  if (length(which) == 0L) return(dev.cur())

  # Only close each device once
  which <- unique(which)
  if (length(which) > 1L) {
    lapply(which, FUN=devDone)
    return(dev.cur())
  }

  # Nothing to do?
  if (!devIsOpen(which)) {
    return(invisible(dev.cur()))
  }

  # Do nothing?
  if (is.numeric(which) && length(which) == 1L && which <= 1L) {
    return(invisible(dev.cur()))
  }

  which <- devSet(which)
  if (which != 1L) {
    type <- tolower(names(which))
    type <- gsub(":.*", "", type)

    knownInteractive <- deviceIsInteractive()
    knownInteractive <- tolower(knownInteractive)
    isOnScreen <- (is.element(type, knownInteractive))
    if (!isOnScreen)
      devOff(which)
  }

  return(invisible(dev.cur()))
} # devDone()


###########################################################################/**
# @RdocFunction devIsInteractive
#
# @title "Checks whether a device type is interactive or not"
#
# \description{
#  @get "title".
# }
#
# @synopsis
#
# \arguments{
#   \item{types}{A @character @vector.}
#   \item{...}{Not used.}
# }
#
# \value{
#   Returns a @logical @vector with @TRUE if the device type is interactive,
#   otherwise @FALSE.
# }
#
# @author
#
# \seealso{
#   Internally, @see "grDevices::deviceIsInteractive" is used.
# }
#
# @keyword device
# @keyword utilities
#*/###########################################################################
devIsInteractive <- function(types, ...) {
  # Known interactive devices
  knownInteractive <- grDevices::deviceIsInteractive()
  knownInteractive <- c(knownInteractive, "CairoWin", "CairoX11")
##  knownInteractive <- c(knownInteractive, "Cairo")

  # Return all known?
  if (missing(types)) return(knownInteractive)

  # Nothing to do?
  if (length(types) == 0L) {
    res <- logical(0L)
    names(res) <- character(0L)
    return(res)
  }

  if (length(types) > 1L) {
    res <- sapply(types, FUN=devIsInteractive)
    if (is.character(types)) names(res) <- types
    return(res)
  }


  # Sanity check
  stop_if_not(length(types) == 1L)

  # Investigate one type below
  type0 <- type <- types

  if (is.function(type)) {
    for (name in knownInteractive) {
      if (exists(name, mode="function")) {
        dev <- get(name, mode="function")
        if (identical(dev, type)) {
          res <- TRUE
          names(res) <- name
          return(res)
        }
      }
    }
    res <- FALSE
    names(res) <- NA_character_
  } else {
    type <- as.character(type)
    # Device type aliases?
    type <- .devTypeName(type)
    type <- tolower(type)

    # An known one?
    res <- is.element(type, tolower(knownInteractive))
    names(res) <- type0
  }

  res
} # devIsInteractive()



devAll <- local({
  # Memoize results, because this function is called many many
  # times either directly or indirectly by various functions.
  .devAll <- NULL

  isFALSE <- function(x) is.logical(x) && length(x) == 1L && !is.na(x) && !x

  base_capabilities <- local({
    res <- base::capabilities()
    function(force=FALSE) {
      if (force) res <<- base::capabilities()
      res
    }
  })

  supports <- function(type, pkg="grDevices") {
    capabilities <- function() character(0L)
    if (isPackageInstalled(pkg)) {
      if (pkg == "Cairo") {
        ns <- getNamespace("Cairo")
        Cairo.capabilities <- get("Cairo.capabilities", envir=ns, mode="function")
        capabilities <- function() {
          res <- Cairo.capabilities()
          names <- sprintf("Cairo%s", toupper(names(res)))
          names <- gsub("WIN", "Win", names, fixed=TRUE)
          names(res) <- names
          res
        }
      } else if (pkg == "grDevices") {
        capabilities <- base_capabilities
      }
    }
    x <- capabilities()
    !isFALSE(x[type])
  } # supports()


  function(force=FALSE, ...) {
    res <- .devAll
    if (force || is.null(res)) {
      if (force) base_capabilities(force=TRUE)

      # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      # Setup all possible devices
      # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      res <- list()

      ## grDevices
      res <- c(res, list(
        bmp          = "grDevices::bmp",
        jpeg         = "grDevices::jpeg",
        pdf          = "grDevices::pdf",
        pictex       = "grDevices::pictex",
        png          = "grDevices::png",
        postscript   = "grDevices::postscript",
        quartz       = "grDevices::quartz",
        svg          = c("grDevices::svg", "needs::cairo"),
        cairo_pdf    = c("grDevices::cairo_pdf", "needs::cairo"),
        cairo_ps     = c("grDevices::cairo_ps", "needs::cairo"),
        tiff         = "grDevices::tiff",
        win.metafile = "grDevices::win.metafile",
        xfig         = "grDevices::xfig"
      ))
      res <- c(res, list(
        windows    = "grDevices::windows",
        x11        = "grDevices::x11",
        X11        = "grDevices::X11"
      ))

      # R.devices
      res <- c(res, list(
        nulldev = c("R.devices::nulldev",
                    "grDevices::png", "grDevices::postscript"),
        eps     = c("R.devices::eps",
                    "grDevices::postscript"),
        favicon = c("R.devices::favicon",
                    "grDevices::png"),
        jpeg2   = c("R.devices::jpeg2",
                    "grDevices::bitmap", "grDevices::postscript"),
        png2    = c("R.devices::png2",
                    "grDevices::bitmap", "grDevices::postscript")
      ))

      ## Cairo package
      res <- c(res, list(
        CairoPDF  = c("Cairo::CairoPDF",
                      "grDevices::pdf"),
        CairoPS   = c("Cairo::CairoPS",
                      "grDevices::postscript"),
        CairoPNG  = c("Cairo::CairoPNG",
                      "grDevices::png"),
        CairoJPEG = c("Cairo::CairoJPEG",
                      "grDevices::jpeg"),
        CairoTIFF = c("Cairo::CairoTIFF",
                      "grDevices::tiff"),
        CairoSVG  = c("Cairo::CairoSVG",
                      "grDevices::svg")
      ))
      res <- c(res, list(
        CairoWin  = c("Cairo::CairoWin",
                      "grDevices::windows"),
        CairoX11  = c("Cairo::CairoX11",
                      "grDevices::x11")
      ))

      ## RStudio devices
      res <- c(res, list(
        RStudioGD  = c("tools:rstudio::RStudioGD",
                       "grDevices::png")
      ))
      
      ## JavaGD
      ## JavaGD=c("JavaGD::JavaGD")


      # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      # Drop devices this system is not capable of
      # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      for (type in names(res)) {
        # Assume it is supported, unless...
        supported <- TRUE

        for (fcn in res[[type]]) {
          pattern <- "^(.+)::(.+)$"
          pkg <- gsub(pattern, "\\1", fcn)
          name <- gsub(pattern, "\\2", fcn)

          if (pkg == "needs") {
            if (!supports(name)) {
              supported <- FALSE
              break
            }
            next
          } else if (pkg == "tools:rstudio") {
            supported <- FALSE
            pos <- match(pkg, search())
            if (!is.na(pos)) {
              env <- pos.to.env(pos)
              supported <- exists("RStudioGD", envir=env, mode="function")
            }
            break
          }

          if (!isPackageInstalled(pkg)) {
            supported <- FALSE
            break
          }

          ns <- getNamespace(pkg)
          if (!exists(name, envir=ns, mode="function")) {
            supported <- FALSE
            break
          }

          if (!supports(type, pkg=pkg)) {
            supported <- FALSE
            break
          }
        } # for (fcn ...)

        # Drop any 'needs::*'
        res[[type]] <- grep("^needs::", res[[type]], value=TRUE, invert=TRUE)

        # Not supported?
        if (!supported) res[[type]] <- NULL
      } # for (type ...)


      # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      # Drop devices based on 'R_R_DEVICES_TYPES_DROP'
      # This is can be used to emulate 'R CMD check' on other OSes
      # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      drop <- Sys.getenv("R_R_DEVICES_TYPES_DROP", "")
      drop <- unlist(strsplit(drop, split = ",", fixed = TRUE))
      res <- res[setdiff(names(res), drop)]

      # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      # Order by name
      # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      o <- order(names(res))
      res <- res[o]

      # Memoize
      .devAll <<- res
    } # if (force ...)

    res
  }
}) # devAll()



# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# BEGIN: Local functions
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
.devList <- function() {
  if (exists(".Devices")) {
    devList <- get(".Devices")
  } else {
    devList <- list("null device")
  }

  labels <- names(devList)
  if (is.null(labels)) {
    labels <- paste("Device", seq_along(devList), sep=" ")
    names(devList) <- labels
    assign(".Devices", devList, envir=baseenv())
  } else {
    # Update the names
    labels <- names(devList)
    idxs <- which(nchar(labels) == 0L)
    if (length(idxs) > 0L) {
      labels[idxs] <- paste("Device", idxs, sep=" ")
    }
    names(devList) <- labels
  }

  devList
} # .devList()

## Gets the devList() index of a device by label
.devListIndexOf <- function(labels, error=TRUE) {
  # Nothing to do?
  if (length(labels) == 0L) {
    res <- integer(0L)
    names(res) <- character(0L)
    return(res)
  }

  devList <- devList(dropNull=FALSE)
  idxs <- match(labels, names(devList))
  names(idxs) <- labels

  # Sanity check
  if (error) {
    if (any(is.na(idxs)))
      throw("No such device: ", paste(labels[is.na(idxs)], collapse=", "))
  }

  idxs
} # .devListIndexOf()


.devNextAvailable <- function() {
  # All open devices
  devList <- dev.list()

  if (length(devList) == 0L)
    return(2L)

  devPossible <- seq(from=2L, to=max(devList)+1L)
  devFree <- setdiff(devPossible, devList)

  devFree[1L]
} # .devNextAvailable()


.devTypeName <- function(types, pattern=FALSE, knownTypes=names(devAll()), mustWork=TRUE, ...) {
  # Nothing todo?
  if (!is.character(types)) {
    return(types)
  }

  names <- names(types)
  if (is.null(names)) {
    names(types) <- types
  }

  # Device type aliases
  aliases <- c(
    ## Aliases first
    "{ps}"           = "{postscript}",
    "{jpg}"          = "{jpeg}",
    "{win.metafile}" = "{wmf}",
    ## Actual aliases 
    "{bmp}"          = "bmp",
    "{eps}"          = "eps",
    "{emf}"          = "emf",
    "{CairoWin}"     = "CairoWin",
    "{CairoX11}"     = "CairoX11",
    "{favicon}"      = "favicon",
    "{jpeg}"         = "jpeg|CairoJPEG",
    "{nulldev}"      = "nulldev",
    "{quartz}"       = "quartz",
    "{pdf}"          = "pdf|cairo_pdf|CairoPDF",
    "{pictex}"       = "pictex",
    "{png}"          = "png|cairo_png|CairoPNG|png2",
    "{postscript}"   = "postscript|cairo_ps|CairoPS",
    "{RStudioGD}"    = "RStudioGD",
    "{svg}"          = "svg|CairoSVG",
    "{tiff}"         = "tiff|CairoTIFF",
    "{windows}"      = "windows",
    "{wmf}"          = "wmf",
    "{x11}"          = "x11|X11|CairoX11",
    "{xfig}"         = "xfig"
  )
  for (name in names(aliases)) {
    alias <- aliases[[name]]
    option <- sprintf("R.devices.alias.%s", gsub("[{}]", "", name))
    alias <- getOption(option, alias)
    types <- gsub(name, alias, types, fixed = TRUE)
  }

  # Select a working device type out of alternatives, e.g. "png|png2"
  idxs <- grep("|", types, fixed = TRUE)
  for (idx in idxs) {
    type <- types[[idx]]
    alts <- unlist(strsplit(type, split = "|", fixed = TRUE))
    alts <- alts[alts %in% knownTypes]
    if (length(alts) == 0L && mustWork) {
      stop("None of the alternative device types are supported: ", sQuote(type))
    }
    if (length(alts) > 1L) alts <- alts[1]
    types[[idx]] <- alts
  }

  # Match to known set of device types by regular expression?
  if (pattern) {
    types <- as.list(types)
    for (kk in seq_along(types)) {
      typeKK <- types[[kk]]
      pattern <- sprintf("^%s$", typeKK)
      idxs <- grep(pattern, knownTypes)
      if (length(idxs) > 0L) {
        typesKK <- knownTypes[idxs]
        names(typesKK) <- rep(typeKK, times=length(typesKK))
        types[[kk]] <- typesKK
      } else {
        names(typeKK) <- typeKK
        names(types[[kk]]) <- typeKK
      }
    } # for (kk ...)
    names <- unlist(lapply(types, FUN = base::names), use.names = FALSE)
    types <- unlist(types, use.names = FALSE)
    names(types) <- names
  }

  # Common aliases
  names <- names(types)
  types[types == "jpg"] <- "jpeg"
  types[types == "ps"] <- "postscript"
  names(types) <- names

  types
} # .devTypeName()

.devTypeExt <- function(types, ...) {
  if (is.function(types)) {
    types <- .devTypeNameFromFunction(types)
  }
  types <- as.character(types)
  exts <- types

  ## Cairo package
  pattern <- "^Cairo(JPEG|PDF|PNG|PS|SVG|TIFF)$"
  idxs <- grep(pattern, exts)
  exts[idxs] <- tolower(gsub(pattern, "\\1", exts[idxs]))

  ## cairo_* devices
  pattern <- "^cairo_(pdf|ps)$"
  exts <- gsub(pattern, "\\1", exts)

  ## Recognize types of any case. Always return in lower case.
  exts <- tolower(exts)

  # Common type-to-extension conversions
  exts[exts == "win.metafile"] <- "wmf"
  exts[exts == "png2"] <- "png"
  exts[exts == "jpeg"] <- "jpg"
  exts[exts == "jpeg2"] <- "jpg"
  exts[exts == "postscript"] <- "ps"

  exts
} # .devTypeExt()


.devTypeNameFromFunction <- function(fcn, knownTypes=R.devices:::devAll(), ...) {
  stop_if_not(length(fcn) == 1, is.function(fcn))
  knownFcns <- lapply(knownTypes, FUN = function(x) {
    eval(parse(text = x[1]), enclos = baseenv())
  })
  name <- NULL
  for (name in names(knownFcns)) {
    if (identical(fcn, knownFcns[[name]])) return(name)
  }
  NA_character_
} # .devTypeNameFromFunction()


.devEqualTypes <- (function() {
   # Recorded known results
   known <- list()

   function(type, other, args=list()) {
     # (a) Same types?
     if (identical(unname(type), unname(other))) return(TRUE)

     # (b) A known equality?
     if (is.character(type) && is.character(other)) {
       res <- known[[type]][other]
       if (is.logical(res)) return(res)
     }

     # (c) Comparing to a device function?
     if (is.function(type) && is.character(other)) {
       if (!exists(other, mode="function")) return(FALSE)
       otherT <- get(other, mode="function")
       if (identical(unname(type), unname(otherT))) return(TRUE)
     } else if (is.function(other) && is.character(type)) {
       if (!exists(type, mode="function")) return(FALSE)
       typeT <- get(type, mode="function")
       if (identical(unname(typeT), unname(other))) return(TRUE)
     }

     # (d) Check if temporarily opening the requested type actually
     # creates a device matching the existing one.
     typeT <- tryCatch({
       local({
         do.call(type, args=args)
         on.exit(dev.off(), add=TRUE)
       })
       names(dev.cur())
     }, error=FALSE)
     if (identical(typeT, FALSE)) return(FALSE)

     if (is.function(other)) {
       other <- tryCatch({
         local({
           do.call(other, args=args)
           on.exit(dev.off(), add=TRUE)
           names(dev.cur())
         })
       }, error=FALSE)
       if (identical(other, FALSE)) return(FALSE)
     }

     res <- (typeT == other)

     # Record result to avoid opening next?
     if (is.character(type) && is.character(other)) {
       known[[type]][other] <<- res
     }

     res
  }
})() # .devEqualTypes()


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# END: Local functions
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

#!/usr/bin/env Rscript

precheck <- function() {
  ## WORKAROUND: Remove checked pkgs that use file links, which otherwise
  ## produce warnings which are promoted to errors by revdepcheck.
  unlink("revdep/checks/aroma.affymetrix", recursive = TRUE)
}

revdepcheck.extras::run()

###########################################################################/**
# @RdocDocumentation "Non-documented objects"
#
# \description{
#   This page contains aliases for all "non-documented" objects that 
#   \code{R CMD check} detects in this package. 
#
#   Almost all of them are \emph{generic} functions that have specific 
#   document for the corresponding method coupled to a specific class. 
#   Other functions are re-defined by \code{setMethodS3()} to 
#   \emph{default} methods. Neither of these two classes are non-documented
#   in reality.
#   The rest are deprecated methods.
# }
#
# @author
#
# @keyword internal
#*/###########################################################################

###########################################################################/**
# @RdocFunction capabilitiesX11
#
# @title "Check whether current R session supports X11 or not"
#
# \description{
#   @get "title".
#
#   Contrary to \code{capabilities("X11")} which only checks for X11
#   support on startup [1], this function checks whether X11 is supported
#   when it is called.  This is done by querying a temporary R session.
# }
#
# @synopsis
#
# \arguments{
#  \item{reset}{If @TRUE, any previously obtained results are ignored,
#   otherwise not.}
#  \item{...}{(optional) @character strings of command-line options
#   to \code{Rscript}.}
# }
#
# \value{
#   Returns @TRUE if X11 is supported, otherwise @FALSE.
# }
#
# @author
#
# \seealso{
#  @see base::capabilities
# }
#
# \references{
#  [1] R-devel thread 'capabilities("X11"): Force refresh from within R?
#      (... and minor documentation issue)' on 2015-05-06,
#      \url{https://stat.ethz.ch/pipermail/r-devel/2015-May/071100.html}\cr
# }
#
# @keyword device
# @keyword internal
#*/###########################################################################
capabilitiesX11 <- local({
  capableOfX11 <- NA
  function(reset=FALSE, ...) {
    if (is.na(capableOfX11) || reset) {
      bin <- file.path(R.home("bin"), "Rscript")
      cmd <- "cat(capabilities('X11'))"
      args <- c(..., "-e", dQuote(cmd))
      value <- system2(bin, args=args, stdout=TRUE)
      capableOfX11 <<- isTRUE(as.logical(value))
    }
    capableOfX11
  }
})

# analysisPipelines

<details>

* Version: 1.0.2
* GitHub: https://github.com/Mu-Sigma/analysis-pipelines
* Source code: https://github.com/cran/analysisPipelines
* Date/Publication: 2020-06-12 08:00:02 UTC
* Number of recursive dependencies: 125

Run `revdep_details(, "analysisPipelines")` for more info

</details>

## In both

*   checking package dependencies ... NOTE
    ```
    Packages which this enhances but not available for checking:
      'SparkR', 'reticulate'
    ```

*   checking LazyData ... NOTE
    ```
      'LazyData' is specified without a 'data' directory
    ```

# aroma.core

<details>

* Version: 3.2.2
* GitHub: https://github.com/HenrikBengtsson/aroma.core
* Source code: https://github.com/cran/aroma.core
* Date/Publication: 2021-01-05 05:10:12 UTC
* Number of recursive dependencies: 48

Run `revdep_details(, "aroma.core")` for more info

</details>

## In both

*   checking package dependencies ... NOTE
    ```
    Packages suggested but not available for checking:
      'sfit', 'expectile', 'HaarSeg', 'mpcbs'
    ```

# eoffice

<details>

* Version: 0.2.1
* GitHub: NA
* Source code: https://github.com/cran/eoffice
* Date/Publication: 2020-11-18 21:40:02 UTC
* Number of recursive dependencies: 89

Run `revdep_details(, "eoffice")` for more info

</details>

## In both

*   checking LazyData ... NOTE
    ```
      'LazyData' is specified without a 'data' directory
    ```

# inpdfr

<details>

* Version: 0.1.11
* GitHub: https://github.com/frareb/inpdfr
* Source code: https://github.com/cran/inpdfr
* Date/Publication: 2020-01-16 12:00:02 UTC
* Number of recursive dependencies: 119

Run `revdep_details(, "inpdfr")` for more info

</details>

## In both

*   checking dependencies in R code ... NOTE
    ```
    Namespace in Imports field not imported from: ‘ca’
      All declared Imports should be used.
    ```

# PointedSDMs

<details>

* Version: 1.0.6
* GitHub: https://github.com/PhilipMostert/PointedSDMs
* Source code: https://github.com/cran/PointedSDMs
* Date/Publication: 2022-06-15 08:20:12 UTC
* Number of recursive dependencies: 131

Run `revdep_details(, "PointedSDMs")` for more info

</details>

## In both

*   checking package dependencies ... NOTE
    ```
    Package suggested but not available for checking: ‘INLA’
    ```

*   checking Rd cross-references ... NOTE
    ```
    Package unavailable to check Rd xrefs: ‘INLA’
    ```

# CRAN submission R.devices 2.17.1

on 2022-06-21

I've verified this submission has no negative impact on any of the 12 reverse package dependencies available on CRAN.

Thanks in advance


## Notes not sent to CRAN

### R CMD check validation

The package has been verified using `R CMD check --as-cran` on:

| R version     | GitHub | R-hub    | mac/win-builder |
| ------------- | ------ | -------- | --------------- |
| 3.4.x         | L      |          |                 |
| 3.5.x         | L      |          |                 |
| 3.6.x         | L      |          |                 |
| 4.1.x         | L      |          |                 |
| 4.2.x         | L M W  | L M M1 W | M1 W            |
| devel         | L   W  | L        |    W            |

*Legend: OS: L = Linux, M = macOS, M1 = macOS M1, W = Windows*


R-hub checks:

```r
res <- rhub::check(platform = c(
  "debian-clang-devel", "debian-gcc-patched", "linux-x86_64-centos-epel",
  "macos-highsierra-release-cran", "macos-m1-bigsur-release",
  "windows-x86_64-release"))
print(res)
```

gives

```
── R.devices 2.17.1: OK

  Build ID:   R.devices_2.17.1.tar.gz-dc868d984f094c3dbe4530e093f557db
  Platform:   Debian Linux, R-devel, clang, ISO-8859-15 locale
  Submitted:  5m 12.1s ago
  Build time: 5m 10s

0 errors ✔ | 0 warnings ✔ | 0 notes ✔

── R.devices 2.17.1: OK

  Build ID:   R.devices_2.17.1.tar.gz-9aa1dae224b647bbbd9d483408dfdc91
  Platform:   Debian Linux, R-patched, GCC
  Submitted:  5m 12.1s ago
  Build time: 4m 37.8s

0 errors ✔ | 0 warnings ✔ | 0 notes ✔

── R.devices 2.17.1: NOTE

  Build ID:   R.devices_2.17.1.tar.gz-a73b58130ca243919cb7c4455c88e359
  Platform:   CentOS 8, stock R from EPEL
  Submitted:  5m 12.2s ago
  Build time: 4m 10.6s

❯ checking package dependencies ... NOTE
  Package suggested but not available for checking: ‘Cairo’

0 errors ✔ | 0 warnings ✔ | 1 note ✖

── R.devices 2.17.1: OK

  Build ID:   R.devices_2.17.1.tar.gz-719d56058e79427facee24a9f6f1e0cd
  Platform:   macOS 10.13.6 High Sierra, R-release, CRAN's setup
  Submitted:  5m 12.2s ago
  Build time: 1m 39.5s

0 errors ✔ | 0 warnings ✔ | 0 notes ✔

── R.devices 2.17.1: OK

  Build ID:   R.devices_2.17.1.tar.gz-965506b7e14e464281df9ed7d853e993
  Platform:   Apple Silicon (M1), macOS 11.6 Big Sur, R-release
  Submitted:  5m 12.2s ago
  Build time: 57.1s

0 errors ✔ | 0 warnings ✔ | 0 notes ✔

── R.devices 2.17.1: OK

  Build ID:   R.devices_2.17.1.tar.gz-03a8cf8c98d447d6ae62c7536ef72881
  Platform:   Windows Server 2008 R2 SP1, R-release, 32/64 bit
  Submitted:  5m 12.2s ago
  Build time: 3m 27.6s

0 errors ✔ | 0 warnings ✔ | 0 notes ✔
```

# Data types

* Rboolean
  - Definition: `typedef enum { FALSE = 0, TRUE /*, MAYBE */ } Rboolean;`
  - Source: R_ext/Boolean.h

* rcolor
  - Definition: `typedef unsigned int rcolor;`
  - Source: R_ext/GraphicsEngine.h

* R_GE_lineend
  - Definition:
    ```c
      typedef enum {
        GE_ROUND_CAP  = 1,
        GE_BUTT_CAP   = 2,
        GE_SQUARE_CAP = 3
      } R_GE_lineend;
    ```
  - Source: R_ext/GraphicsEngine.h

* R_GE_linejoin
  - Definition:
    ```c
      typedef enum {
        GE_ROUND_JOIN = 1,
        GE_MITRE_JOIN = 2,
        GE_BEVEL_JOIN = 3
      } R_GE_linejoin;
    ```
  - Source: R_ext/GraphicsEngine.h

* GTrans:
  - Definition:
    ```c
      typedef struct {
      	double ax;
      	double bx;
      	double ay;
      	double by;
      } GTrans;
    ```
  - Source: Graphics.h


* GUnit:
  - Definition:
    ```c
typedef enum {
 DEVICE	= 0,	/* native device coordinates (rasters) */
 NDC	= 1,	/* normalised device coordinates x=(0,1), y=(0,1) */
 INCHES = 13,	/* inches x=(0,width), y=(0,height) */
 NIC	= 6,	/* normalised inner region coordinates (0,1) */
 OMA1	= 2,	/* outer margin 1 (bottom) x=NIC, y=LINES */
 OMA2	= 3,	/* outer margin 2 (left) */
 OMA3	= 4,	/* outer margin 3 (top) */
 OMA4	= 5,	/* outer margin 4 (right) */
 NFC	= 7,	/* normalised figure region coordinates (0,1) */
 NPC	= 16,	/* normalised plot region coordinates (0,1) */
 USER	= 12,	/* user/data/world coordinates;
		 * x,=(xmin,xmax), y=(ymin,ymax) */
 MAR1	= 8,	/* figure margin 1 (bottom) x=USER(x), y=LINES */
 MAR2	= 9,	/* figure margin 2 (left)   x=USER(y), y=LINES */
 MAR3	= 10,	/* figure margin 3 (top)    x=USER(x), y=LINES */
 MAR4	= 11,	/* figure margin 4 (right)  x=USER(y), y=LINES */

	/* possible, units (for specifying dimensions) */
	/* all of the above, plus ... */

 LINES = 14,	/* multiples of a line in the margin (mex) */
 CHARS = 15	/* multiples of text height (cex) */
} GUnit;
    ```
  - Source: Rgraphics.h



typedef struct {
    /* Plot State */
    /*
       When the device driver is started this is 0
       After the first call to plot.new/perps it is 1
       Every graphics operation except plot.new/persp
       should fail if state = 0
       This is checked at the highest internal function
       level (e.g., do_lines, do_axis, do_plot_xy, ...)
    */

    int	state;		/* plot state: 1 if GNewPlot has been called
			   (by plot.new or persp) */
    Rboolean valid;	/* valid layout ?  Used in GCheckState & do_playDL */

    /* GRZ-like Graphics Parameters */
    /* ``The horror, the horror ... '' */
    /* Marlon Brando - Appocalypse Now */

    /* General Parameters -- set and interrogated directly */

    double adj;		/* String adjustment */
    Rboolean ann;	/* Should annotation take place */
    rcolor bg;		/* **R ONLY** Background color */
    char bty;		/* Box type */
    double cex;		/* Character expansion */
    double lheight;     /* Line height
			   The height of a line of text is:
			   ps * cex * lheight */
    rcolor col;		/* Plotting Color */
    double crt;		/* Character/string rotation */
    double din[2];	/* device size in inches */
    int	err;		/* Error repporting level */
    rcolor fg;		/* **R ONLY** Foreground Color */
    char family[201];  /* **R ONLY** Font family
			   Simple name which is mapped by device-specific
			   font database to device-specific name.
			   Only used if not "".
			   Default is "".
			   Ignored by some devices. */
    int	font;		/* Text font */
    double gamma;	/* Device Gamma Correction */
    int	lab[3];		/* Axis labelling */
			/* [0] = # ticks on x-axis */
			/* [1] = # ticks on y-axis */
			/* [2] = length of axis labels */
    int	las;		/* Label style (rotation) */
    int	lty;		/* Line texture */
    double lwd;		/* Line width */
    R_GE_lineend lend;  /* **R ONLY** Line end style */
    R_GE_linejoin ljoin;/* **R ONLY** Line join style */
    double lmitre;      /* **R ONLY** Line mitre limit */
    double mgp[3];	/* Annotation location */
			/* [0] = location of axis title */
			/* [1] = location of axis label */
			/* [2] = location of axis line */
    double mkh;		/* Mark size in inches */
    int	pch;		/* Plotting character */
    /* Note that ps is never changed, so always the same as dev->startps.
       However, the ps in the graphics context is changed */
    double ps;		/* Text & symbol pointsize */
    int	smo;		/* Curve smoothness */
    double srt;		/* String Rotation */
    double tck;		/* Tick size as in S */
    double tcl;		/* Tick size in "lines" */
    double xaxp[3];	/* X Axis annotation */
			/* [0] = coordinate of lower tick */
			/* [1] = coordinate of upper tick */
			/* [2] = num tick intervals */
			/* almost always used internally */
    char xaxs;		/* X Axis style */
    char xaxt;		/* X Axis type */
    Rboolean xlog;	/* Log Axis for X */
    int	xpd;		/* Clip to plot region indicator */
    int	oldxpd;
    double yaxp[3];	/* Y Axis annotation */
    char yaxs;		/* Y Axis style */
    char yaxt;		/* Y Axis type */
    Rboolean ylog;	/* Log Axis for Y */

    /* Annotation Parameters */

    double cexbase;	/* Base character size */
    double cexmain;	/* Main title size */
    double cexlab;	/* xlab and ylab size */
    double cexsub;	/* Sub title size */
    double cexaxis;	/* Axis label size */

    int	fontmain;	/* Main title font */
    int	fontlab;	/* Xlab and ylab font */
    int	fontsub;	/* Subtitle font */
    int	fontaxis;	/* Axis label fonts */

    rcolor colmain;	/* Main title color */
    rcolor collab;	/* Xlab and ylab color */
    rcolor colsub;	/* Subtitle color */
    rcolor colaxis;	/* Axis label color */

    /* Layout Parameters */

    Rboolean layout;	/* has a layout been specified */

    int	numrows;
    int	numcols;
    int	currentFigure;
    int	lastFigure;
    double heights[MAX_LAYOUT_ROWS];
    double widths[MAX_LAYOUT_COLS];
    int	cmHeights[MAX_LAYOUT_ROWS];
    int	cmWidths[MAX_LAYOUT_COLS];
    unsigned short order[MAX_LAYOUT_CELLS];
    int	rspct;		/* 0 = none, 1 = full, 2 = see respect */
    unsigned char respect[MAX_LAYOUT_CELLS];

    int	mfind;		/* By row/col indicator */

    /* Layout parameters which can be set directly by the */
    /* user (e.g., par(fig=c(.5,1,0,1))) or otherwise are */
    /* calculated automatically */
    /* NOTE that *Units parameters are for internal use only */

    double fig[4];	/* (current) Figure size (proportion) */
			/* [0] = left, [1] = right */
			/* [2] = bottom, [3] = top */
    double fin[2];	/* (current) Figure size (inches) */
			/* [0] = width, [1] = height */
    GUnit fUnits;	/* (current) figure size units */
    double plt[4];	/* (current) Plot size (proportions) */
			/* [0] = left, [1] = right */
			/* [2] = bottom, [3] = top */
    double pin[2];	/* (current) plot size (inches) */
			/* [0] = width, [1] = height */
    GUnit pUnits;	/* (current) plot size units */
    Rboolean defaultFigure;	/* calculate figure from layout ? */
    Rboolean defaultPlot;	/* calculate plot from figure - margins ? */

    /* Layout parameters which are set directly by the user */

    double mar[4];	/* Plot margins in lines */
    double mai[4];	/* Plot margins in inches */
			/* [0] = bottom, [1] = left */
			/* [2] = top, [3] = right */
    GUnit mUnits;	/* plot margin units */
    double mex;		/* Margin expansion factor */
    double oma[4];	/* Outer margins in lines */
    double omi[4];	/* outer margins in inches */
    double omd[4];	/* outer margins in NDC */
			/* [0] = bottom, [1] = left */
			/* [2] = top, [3] = right */
    GUnit oUnits;	/* outer margin units */
    char pty;		/* Plot type */

    /* Layout parameters which can be set by the user, but */
    /* almost always get automatically calculated anyway */

    double usr[4];	/* Graphics window */
			/* [0] = xmin, [1] = xmax */
			/* [2] = ymin, [3] = ymax */

    /* The logged usr parameter;  if xlog, use logusr[0:1] */
    /* if ylog, use logusr[2:3] */

    double logusr[4];

    /* Layout parameter: Internal flags */

    Rboolean new;	/* Clean plot ? */
    int	devmode;	/* creating new image or adding to existing one */

    /* Coordinate System Mappings */
    /* These are only used internally (i.e., cannot be */
    /* set directly by the user) */

    /* The reliability of these parameters relies on */
    /* the fact that plot.new is the */
    /* first graphics operation called in the creation */
    /* of a graph  (unless it is a call to persp) */

    /* udpated per plot.new */

    double xNDCPerChar;	/* Nominal character width (NDC) */
    double yNDCPerChar;	/* Nominal character height (NDC) */
    double xNDCPerLine;	/* Nominal line width (NDC) */
    double yNDCPerLine;	/* Nominal line height (NDC) */
    double xNDCPerInch;	/* xNDC -> Inches */
    double yNDCPerInch;	/* yNDC -> Inches */

    /* updated per plot.new and if inner2dev changes */

    GTrans fig2dev;	/* Figure to device */

    /* udpated per DevNewPlot and if ndc2dev changes */

    GTrans inner2dev;	/* Inner region to device */

    /* udpated per device resize */

    GTrans ndc2dev;	/* NDC to raw device */

    /* updated per plot.new and per plot.window */

    GTrans win2fig;	/* Window to figure mapping */

    /* NOTE: if user has not set fig and/or plt then */
    /* they need to be updated per plot.new too */

    double scale;       /* An internal "zoom" factor to apply to ps and lwd */
                        /* (for fit-to-window resizing in Windows) */
} GPar;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  DevEvalProduct.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{getMimeType.DevEvalFileProduct}
\alias{getMimeType.DevEvalFileProduct}
\alias{DevEvalFileProduct.getMimeType}
\alias{getMimeType,DevEvalFileProduct-method}
\alias{DevEvalFileProduct.getMime}
\alias{getMime.DevEvalFileProduct}
\alias{getMime,DevEvalFileProduct-method}

\alias{getMimeType}
\alias{getMime}

\title{Gets the MIME type}

\description{
 Gets the MIME type.
}

\usage{
\method{getMimeType}{DevEvalFileProduct}(this, default="", ...)
}

\arguments{
  \item{default}{The value returned, if the MIME type could not be inferred.}
  \item{...}{Not used.}
}

\value{
 Returns a \code{\link[base]{character}} string.
}

\author{Henrik Bengtsson}

\seealso{
  For more information see \code{\link{DevEvalFileProduct}}.
}
\keyword{internal}
\keyword{methods}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  devNew.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devNew}
\alias{devNew}


\title{Opens a new device}

\description{
 Opens a new device.
}

\usage{
devNew(type=getOption("device"), ..., scale=1, aspectRatio=1, par=NULL, label=NULL)
}

\arguments{
  \item{type}{A \code{\link[base]{character}} string specifying the type of device to be
    opened. This string should match the name of an existing device
    \code{\link[base]{function}}.}
  \item{...}{Additional arguments passed to the device \code{\link[base]{function}}, e.g.
    \code{width} and \code{height}.  If not given, the are inferred
    from \code{\link{devOptions}}().}
  \item{scale}{A \code{\link[base]{numeric}} scalar factor specifying how much the
    width and the height should be rescaled.}
  \item{aspectRatio}{A \code{\link[base]{numeric}} ratio specifying the aspect ratio
    of the image.  See below.}
  \item{par}{An optional named \code{\link[base]{list}} of graphical settings applied,
    that is, passed to \code{\link[graphics]{par}}, immediately after
    opening the device.}
  \item{label}{An optional \code{\link[base]{character}} string specifying the label of the
    opened device.}
}

\value{
  Returns the device index of the opened device.
}

\section{Width and heights}{
  The default width and height of the generated image is specific to
  the type of device used.  There is not straightforward programmatic
  way to infer these defaults; here we use \code{\link{devOptions}}(), which
  in most cases returns the correct defaults.
}

\section{Aspect ratio}{
  The aspect ratio of an image is the height relative to the width.
  If argument \code{height} is not given (or \code{\link[base]{NULL}}), it is
  calculated as \code{aspectRatio*width} as long as they are given.
  Likewise, if argument \code{width} is not given (or \code{\link[base]{NULL}}), it is
  calculated as \code{width/aspectRatio} as long as they are given.
  If neither \code{width} nor \code{height} is given, then \code{width}
  defaults to \code{devOptions(type)$width}.
  If both \code{width} and \code{height} are given, then
  \code{aspectRatio} is ignored.
}

\author{Henrik Bengtsson}

\seealso{
  \code{\link{devDone}}() and \code{\link{devOff}}().
  For simplified generation of image files, see \code{\link{devEval}}().
}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  deviceUtils.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devGetLabel}
\alias{devGetLabel}


\title{Gets the labels of zero or more devices}

\description{
 Gets the labels of zero or more devices.
}

\usage{
devGetLabel(which=dev.cur(), ...)
}

\arguments{
  \item{which}{An index (\code{\link[base]{numeric}}) \code{\link[base]{vector}} or a label (\code{\link[base]{character}}) \code{\link[base]{vector}}.}
  \item{...}{Not used.}
}

\value{
  Returns a \code{\link[base]{character}} \code{\link[base]{vector}}.
  If a device does not exist, an error is thrown.
}

\author{Henrik Bengtsson}

\seealso{
  \code{\link{devSetLabel}}() and \code{\link{devList}}().
}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  png2.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{png2}
\alias{png2}


\title{A PNG device for Bitmap Files via GhostScript}

\description{
  A PNG device for Bitmap Files via GhostScript.
}

\usage{
png2(filename, width=480, height=480, res=144, type="png256", ...)
}

\arguments{
 \item{filename}{The name of the file to be produced.}
 \item{width, height}{The width and height (in pixels) of the result image.}
 \item{res}{The resolution of the image.}
 \item{type}{The output type. See \code{\link[grDevices]{dev2bitmap}} for details.}
 \item{...}{Additional arguments passed to \code{bitmap()}.}
}

\value{
  Returns nothing.
}

\author{Henrik Bengtsson}



\keyword{device}
\keyword{internal}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  deviceUtils.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devList}
\alias{devList}


\title{Lists the indices of the open devices named by their labels}

\description{
 Lists the indices of the open devices named by their labels.
}

\usage{
devList(interactiveOnly=FALSE, dropNull=TRUE, ...)
}

\arguments{
  \item{interactiveOnly}{If \code{\link[base:logical]{TRUE}}, only open interactive/screen devices
    are returned.}
  \item{dropNull}{If \code{\link[base:logical]{TRUE}}, the "null" device (device index 1) is
    not returned.}
  \item{...}{Not used.}
}

\value{
  Returns a named \code{\link[base]{integer}} \code{\link[base]{vector}}.
}

\author{Henrik Bengtsson}

\seealso{
  \code{\link[grDevices:dev]{dev.list}()}.
}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  capturePlot.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{capturePlot}
\alias{capturePlot}


\title{Captures a plot such that it can be redrawn later/elsewhere}

\description{
  Captures a plot such that it can be redrawn later/elsewhere.

  \emph{This feature is only supported in R (>= 3.3.0).}
}

\usage{
capturePlot(expr, envir=parent.frame(), type=nulldev, ...)
}

\arguments{
  \item{expr}{The \code{\link[base]{expression}} of graphing commands to be evaluated.}
  \item{envir}{The \code{\link[base]{environment}} where \code{expr} should be evaluated.}
  \item{type}{The type of graphics device used in the background.
   The choice should not matter since the result should be identical
   regardless.  All graphics is captured but any output is also voided
   by sending the output to a "null" file.}
 \item{...}{Additional arguments passed to the graphics device.}
}

\value{
  Returns a \code{recordedplot} object, which can be
  \code{\link[grDevices]{replayPlot}()}:ed.  If replayed in an
  interactive session, the plot is displayed in a new window.
  For conveniency, the object is also replayed when \code{print()}:ed.
}

\details{
  Note that plot dimensions/aspect ratios are not recorded.  This
  means that one does not have to worry about those when recording
  the plot.  Instead, they are specified when setting up the graphics
  device(s) in which the recorded plot is replayed (see example).
}

\section{Replaying / replotting on a different architecture}{
 In order to replay a \code{recordedplot} object, it has to be replayed
 on an architecture that is compatible with the one who created the
 object.
 If this is not the case, then \code{\link[grDevices]{replayPlot}()}
 will generate an \emph{Incompatible graphics state} error.
 The \code{\link{as.architecture}()} function of this package tries
 to coerce between different architectures, such that one can replay
 across architectures using \code{replayPlot(as.architectures(g))}.
 For convenience, the recorded plot returned by \code{capturePlot()}
 is automatically coerced when \code{print()}:ed.
}

\examples{
if (getRversion() >= "3.3.0") {
oopts <- R.devices::devOptions("*", path=file.path(tempdir(), "figures"))

g <- capturePlot({
  plot(1:10)
})

## Display
print(g)

## Display with a 2/3 height-to-width aspect ratio
toDefault(aspectRatio=2/3, { print(g) })

## Redraw to many output formats using whatever PNG, EPS, and PDF
## device outputs available
devEval(c("{png}", "{eps}", "{pdf}"), aspectRatio=2/3, print(g))

R.devices::devOptions("*", path=oopts$path)
} ## if (getRversion() >= "3.3.0")
}

\author{Henrik Bengtsson}

\seealso{
  Internally \code{\link[grDevices]{recordPlot}()} is used.
}

\references{
 [1] Paul Murrell et al.,
     \emph{Recording and Replaying the Graphics Engine Display List},
     December 2015.
     \url{https://www.stat.auckland.ac.nz/~paul/Reports/DisplayList/dl-record.html}\cr
}


\keyword{device}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  deviceUtils.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devOff}
\alias{devOff}


\title{Closes zero or more devices}

\description{
 Closes zero or more devices.
}

\usage{
devOff(which=dev.cur(), ...)
}

\arguments{
  \item{which}{An index (\code{\link[base]{numeric}}) \code{\link[base]{vector}} or a label (\code{\link[base]{character}}) \code{\link[base]{vector}}.}
  \item{...}{Not used.}
}

\value{
  Returns \code{\link[grDevices:dev]{dev.cur}()}.
}

\author{Henrik Bengtsson}

\seealso{
  \code{\link{devDone}}().
  Internally, \code{\link[grDevices:dev]{dev.off}()} is used.
}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  DevEvalProduct.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{as.character.DevEvalProduct}
\alias{as.character.DevEvalProduct}
\alias{DevEvalProduct.as.character}
\alias{as.character,DevEvalProduct-method}
\alias{as.character.DevEvalFileProduct}
\alias{as.character}

\title{Gets a character representation of the product}

\description{
  Gets a character representation of the product.
}

\usage{
\method{as.character}{DevEvalProduct}(x, ...)
}

\arguments{
 \item{...}{Not used.}
}

\value{
  Returns a \code{\link[base]{character}} string.
}

\author{Henrik Bengtsson}

\seealso{
  For more information see \code{\link{DevEvalProduct}}.
}
\keyword{internal}
\keyword{methods}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  deviceUtils.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devSet}
\alias{devSet}


\title{Activates a device}

\description{
 Activates a device.
}

\usage{
devSet(which=dev.next(), ...)
}

\arguments{
  \item{which}{An index (\code{\link[base]{numeric}}) or a label (\code{\link[base]{character}}).
    If neither, then a label corresponding to the checksum of
    \code{which} as generated by \code{\link[digest]{digest}} is used.}
  \item{...}{Not used.}
}

\value{
  Returns what \code{\link[grDevices:dev]{dev.set}()} returns.
}

\author{Henrik Bengtsson}

\seealso{
  \code{\link{devOff}}() and \code{\link{devDone}}().
  Internally, \code{\link[grDevices:dev]{dev.set}()} is used.
}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  devOptions.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devOptions}
\alias{devOptions}

\alias{getDevOption}

\title{Gets the default device options}

\description{
 Gets the default device options as given by predefined devices options adjusted for
 the default arguments of the device function.
}

\usage{
devOptions(type=NULL, custom=TRUE, special=TRUE, inherits=FALSE, drop=TRUE,
  options=list(), ..., reset=FALSE)
}

\arguments{
  \item{type}{A \code{\link[base]{character}} string or a device \code{\link[base]{function}} specifying
     the device to be queried.  May also be a \code{\link[base]{vector}}, for querying
     device options for multiple devices.}
  \item{custom}{If \code{\link[base:logical]{TRUE}}, also the default settings specific to this
     function is returned. For more details, see below.}
  \item{special}{A \code{\link[base]{logical}}.  For more details, see below.}
  \item{inherits}{If \code{\link[base:logical]{TRUE}}, the global option is used if the
     type-specific is not set (or \code{\link[base]{NULL}}).}
  \item{drop}{If \code{\link[base:logical]{TRUE}} and only one device type is queried, then
     a \code{\link[base]{list}} is returned, otherwise a \code{\link[base]{matrix}}.}
  \item{options}{Optional named \code{\link[base]{list}} of settings.}
  \item{...}{Optional named arguments for setting new defaults.
     For more details, see below.}
  \item{reset}{If \code{\link[base:logical]{TRUE}}, the device options are reset to R defaults.}
}

\value{
  If \code{drop=TRUE} and a single device is queries, a named \code{\link[base]{list}} is
  returned, otherwise a \code{\link[base]{matrix}} is returned.
  If a requested device is not implemented/supported on the current system,
  then "empty" results are returned.
  If options were set, that is, if named options were specified via
  \code{...}, then the list is returned invisibly, otherwise not.
}

\details{
 If argument \code{special} is \code{\link[base:logical]{TRUE}}, then the 'width' and 'height'
 options are adjusted according to the rules explained for
 argument 'paper' in \code{\link[grDevices]{pdf}}, \code{\link[grDevices]{postscript}},
 and \code{\link[grDevices]{xfig}}.
}

\section{Setting new defaults}{
 When setting device options, the \code{getOption("devOptions")[[type]]}
 option is modified.  This means that for such options to be effective,
 any device function needs to query also such options, which for instance
 is done by \code{\link{devNew}}().

 Also, for certain devices (eps, pdf, postscript, quartz, windows and x11),
 builtin R device options are set.
}

\examples{
# Tabulate some of the default settings for known devices
print(devOptions()[,c("width", "height", "bg", "fg", "pointsize")])
}

\author{Henrik Bengtsson}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  DevEvalProduct.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{getDataURI.DevEvalFileProduct}
\alias{getDataURI.DevEvalFileProduct}
\alias{DevEvalFileProduct.getDataURI}
\alias{getDataURI,DevEvalFileProduct-method}
\alias{getDataURI}
\alias{DevEvalFileProduct.getData}
\alias{getData.DevEvalFileProduct}
\alias{getData,DevEvalFileProduct-method}

\alias{getData}

\title{Gets content as a Base64-encoded data URI}

\description{
 Gets content as a Base64-encoded data URI.
}

\usage{
\method{getDataURI}{DevEvalFileProduct}(this, mime=getMimeType(this), ...)
}

\arguments{
  \item{mime}{The MIME type to be embedded in the data URI.}
  \item{...}{Not used.}
}

\value{
 Returns a \code{\link[base]{character}} string.
}

\author{Henrik Bengtsson}

\seealso{
  For more information see \code{\link{DevEvalFileProduct}}.
}
\keyword{internal}
\keyword{methods}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  eps.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{eps}
\alias{eps}


\title{EPS graphics device}

\description{
 Device driver for Encapsulated Postscript. This driver is the same as
 the postscript driver where some arguments have different default values.
}

\usage{
eps(file="Rplot\%03d.eps", width=7, height=7, horizontal=FALSE, paper="special",
  onefile=FALSE, ...)
}

\arguments{
  \item{file}{Default file name (pattern).}
  \item{width, height}{The width and height of the figure.}
  \item{horizontal}{If \code{\link[base:logical]{FALSE}}, an horizontal EPS file is created,
        otherwise a portrait file is created.}
  \item{paper}{A \code{\link[base]{character}} string specifying the paper type. Overrides
        the default of \code{postscript()}.}
  \item{onefile}{Not used.}
  \item{...}{Other arguments accepted by \code{postscript()}.}
}

\value{
  A plot device is opened; nothing is returned.
}

\examples{\dontrun{
  eps("foo.eps", width=7, height=7)

  # is identical to

  postscript("foo.eps", width=7, height=7, onefile=TRUE, horizontal=FALSE)

  # and

  dev.print(eps, "foo.eps", ...)

  # is identical to

  dev.print(postscript, "foo.eps", onefile=TRUE, horizontal=FALSE, paper="special", ...)
}}

\keyword{device}

\seealso{
  This is just a convenient wrapper for \code{\link[grDevices]{postscript}}
  with the proper arguments set to generate an EPS file.

  It is recommended to use \code{\link{toEPS}}() instead.
}

\author{Henrik Bengtsson}



\keyword{device}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/capturePlot.R
\name{print.RecordedPlot}
\alias{print.RecordedPlot}
\title{Automatically replays a recorded plot}
\usage{
\method{print}{RecordedPlot}(x, ...)
}
\arguments{
\item{x}{A recorded plot of class \code{recordedplot}.}
}
\value{
Returns \code{x} invisibly.
}
\description{
This is identical to the \code{\link[grDevices:replayPlot]{print}()}
method available in \pkg{grDevices}, but if replaying the plot gives
an error it will also try to replay it after coercing the data structure
to match the architecture of the current machine.  This will make it
possible to, for instance, replay a plot generated on a 32-bit machine
on a 64-bit machine.
}
\seealso{
Internally, \code{\link{as.architecture}()} is used
to coerce to the current architecture.
}
\keyword{internal}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  DevEvalProduct.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{getPathname.DevEvalFileProduct}
\alias{getPathname.DevEvalFileProduct}
\alias{DevEvalFileProduct.getPathname}
\alias{getPathname,DevEvalFileProduct-method}
\alias{DevEvalFileProduct.getFilename}
\alias{getFilename.DevEvalFileProduct}
\alias{getFilename,DevEvalFileProduct-method}

\alias{DevEvalFileProduct.getPath}
\alias{getPath.DevEvalFileProduct}
\alias{getPath,DevEvalFileProduct-method}

\alias{DevEvalFileProduct.getExtension}
\alias{getExtension.DevEvalFileProduct}
\alias{getExtension,DevEvalFileProduct-method}

\alias{getPathname}
\alias{getFilename}
\alias{getPath}
\alias{getExtension}

\title{Gets the (relative) pathname, filename and path}

\description{
  Gets the (relative) pathname, filename and path.
}

\usage{
\method{getPathname}{DevEvalFileProduct}(this, relative=TRUE, ...)
\method{getFilename}{DevEvalFileProduct}(this, ...)
\method{getPath}{DevEvalFileProduct}(this, ...)
\method{getExtension}{DevEvalFileProduct}(this, ...)
}

\arguments{
 \item{...}{Not used.}
}

\value{
  Returns a \code{\link[base]{character}} string.
}

\author{Henrik Bengtsson}

\seealso{
  For more information see \code{\link{DevEvalFileProduct}}.
}
\keyword{internal}
\keyword{methods}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  DevEvalProduct.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{getFullname.DevEvalProduct}
\alias{getFullname.DevEvalProduct}
\alias{DevEvalProduct.getFullname}
\alias{getFullname,DevEvalProduct-method}
\alias{getFullname.DevEvalFileProduct}
\alias{DevEvalProduct.getName}
\alias{getName.DevEvalProduct}
\alias{getName,DevEvalProduct-method}

\alias{DevEvalProduct.getTags}
\alias{getTags.DevEvalProduct}
\alias{getTags,DevEvalProduct-method}

\alias{getFullname}
\alias{getName}
\alias{getTags}

\title{Gets the full name, name and tags}

\description{
  Gets the full name, name and tags consisting of a name and tags.
}

\usage{
\method{getFullname}{DevEvalProduct}(this, ...)
\method{getFullname}{DevEvalFileProduct}(this, ...)
\method{getName}{DevEvalProduct}(this, ...)
\method{getTags}{DevEvalProduct}(this, collapse=",", ...)
}

\arguments{
 \item{...}{Not used.}
}

\value{
  Returns a \code{\link[base]{character}} or a \code{\link[base]{character}} \code{\link[base]{vector}}.
}

\author{Henrik Bengtsson}

\seealso{
  For more information see \code{\link{DevEvalProduct}}.
}
\keyword{internal}
\keyword{methods}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  devEval.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devEval}
\alias{devEval}

\alias{devDump}
\alias{suppressGraphics}

\title{Opens a new graphics device, evaluate (graphing) code, and closes device}

\description{
 Opens a new graphics device, evaluate (graphing) code, and closes device.
}

\usage{
devEval(type=getOption("device"), expr, initially=NULL, finally=NULL,
  envir=parent.frame(), name=NULL, tags=NULL, sep=getDevOption(type, "sep",
  default = ","), ..., ext=NULL, filename=NULL, path=getDevOption(type, "path",
  default = "figures"), field=getDevOption(type, name = "field"),
  onIncomplete=c("remove", "rename", "keep"), force=getDevOption(type, "force",
  default = TRUE), which=dev.cur(), .exprAsIs=FALSE, .allowUnknownArgs=FALSE)
}

\arguments{
  \item{type}{Specifies the type of graphics device to be used.
   The device is created and opened using \code{\link{devNew}}().
   Multiple types may be specified.}
  \item{expr}{The \code{\link[base]{expression}} of graphing commands to be evaluated.}
  \item{initially, finally}{Optional \code{\link[base]{expression}}:s to be evaluated
   before and after \code{expr}. If \code{type} specifies multiple
   devices, these optional \code{\link[base]{expression}}:s are only evaluated ones.}
  \item{envir}{The \code{\link[base]{environment}} where \code{expr} should be evaluated.}
  \item{name, tags, sep}{The fullname name of the image is specified
    as the name with optional \code{sep}-separated tags appended.}
  \item{ext}{The filename extension of the image file generated, if any.
   By default, it is inferred from argument \code{type}.}
  \item{...}{Additional arguments passed to \code{\link{devNew}}().}
  \item{filename}{The filename of the image saved, if any.
   By default, it is composed of arguments \code{name}, \code{tags},
   \code{sep}, and \code{ext}.  See also below.}
  \item{path}{The directory where then image should be saved, if any.}
  \item{field}{An optional \code{\link[base]{character}} string specifying a specific
    field of the named result \code{\link[base]{list}} to be returned.}
  \item{onIncomplete}{A \code{\link[base]{character}} string specifying what to do with
    an image file that was incompletely generated due to an interrupt
    or an error.}
  \item{force}{If \code{\link[base:logical]{TRUE}}, and the image file already exists, then it is
    overwritten, otherwise not.}
  \item{which}{A \code{\link[base]{vector}} of devices to be copied.  Only applied if
    argument \code{expr} is missing.}
  \item{.exprAsIs, .allowUnknownArgs}{(Internal use only).}
}

\value{
  Returns a \code{\link{DevEvalFileProduct}} if the device generated an
  image file, otherwise an \code{\link{DevEvalProduct}}.
  If argument \code{field} is given, then the field of the
  \code{\link{DevEvalProduct}} is returned instead.
  \emph{Note that the default return value may be changed in future releases.}
}

\section{Generated image file}{
  If created, the generated image file is saved in the directory
  specified by argument \code{path} with a filename consisting of
  the \code{name} followed by optional comma-separated \code{tags}
  and a filename extension given by argument \code{ext}.

  By default, the image file is only created if the \code{expr}
  is evaluated completely.  If it is, for instance, interrupted
  by the user or due to an error, then any incomplete/blank image
  file that was created will be removed.  This behavior can be
  turned of using argument \code{onIncomplete}.
}

\examples{
# Plot to PNG using one whatever PNG device driver is available
res <- devEval("{png}", name="MyPlot", tags=c("10", "rnd"), aspectRatio=0.7, {
  plot(1:10)
})
print(res$pathname)
# [1] "figures/MyPlot,10,rnd.png"

str(res$dataURI)
# chr "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAeAA..."

## Plot to PDF using grDevices::pdf()
res <- devEval("pdf", name="MyPlot", tags=c("10", "rnd"), aspectRatio=0.7, {
  plot(1:10)
})
print(res$pathname)
# [1] "figures/MyPlot,10,rnd.pdf"


## Plot to EPS using R.devices::eps()
res <- devEval("eps", name="MyPlot", tags=c("10", "rnd"), aspectRatio=0.7, {
  plot(1:10)
})
print(res$pathname)
# [1] "figures/MyPlot,10,rnd.eps"
}

\author{Henrik Bengtsson}

\seealso{
  To change default device parameters such as the width or the height,
  \code{\link{devOptions}}().
  \code{\link{devNew}}().
}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  nulldev.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{nulldev}
\alias{nulldev}


\title{A \"null\" graphics device voiding all output}

\description{
 A \"null\" graphics device voiding all output.
}

\usage{
nulldev(file=nullfile(), ...)
}

\arguments{
  \item{file}{The output file (ignored; forced to its default).}
  \item{...}{All other arguments are also ignored.}
}

\value{
  A plot device is opened; nothing is returned.
}

\details{
  The null graphics device opens a regular \link[grDevices]{pdf} device
  and directs its output to the null file, which is \code{/dev/null} unless
  on Windows where it is \code{NUL}.
}

\examples{\dontrun{
  nulldev()
  plot(1:3)
  dev.off()
}}

\keyword{device}

\author{Henrik Bengtsson}



\keyword{device}
\keyword{internal}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  999.NonDocumentedObjects.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{Non-documented objects}
\alias{Non-documented objects}
\title{Non-documented objects}


\description{
  This page contains aliases for all "non-documented" objects that
  \code{R CMD check} detects in this package.

  Almost all of them are \emph{generic} functions that have specific
  document for the corresponding method coupled to a specific class.
  Other functions are re-defined by \code{setMethodS3()} to
  \emph{default} methods. Neither of these two classes are non-documented
  in reality.
  The rest are deprecated methods.
}

\author{Henrik Bengtsson}


\keyword{documentation}
\keyword{internal}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  capabilitiesX11.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{capabilitiesX11}
\alias{capabilitiesX11}


\title{Check whether current R session supports X11 or not}

\description{
  Check whether current R session supports X11 or not.

  Contrary to \code{capabilities("X11")} which only checks for X11
  support on startup [1], this function checks whether X11 is supported
  when it is called.  This is done by querying a temporary R session.
}

\usage{
capabilitiesX11(reset=FALSE, ...)
}

\arguments{
 \item{reset}{If \code{\link[base:logical]{TRUE}}, any previously obtained results are ignored,
  otherwise not.}
 \item{...}{(optional) \code{\link[base]{character}} strings of command-line options
  to \code{Rscript}.}
}

\value{
  Returns \code{\link[base:logical]{TRUE}} if X11 is supported, otherwise \code{\link[base:logical]{FALSE}}.
}

\author{Henrik Bengtsson}

\seealso{
 \code{\link[base]{capabilities}}()
}

\references{
 [1] R-devel thread 'capabilities("X11"): Force refresh from within R?
     (... and minor documentation issue)' on 2015-05-06,
     \url{https://stat.ethz.ch/pipermail/r-devel/2015-May/071100.html}\cr
}



\keyword{device}
\keyword{internal}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  999.package.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{R.devices-package}
\alias{R.devices-package}
\alias{R.devices}
\docType{package}

\title{Package R.devices}


\description{
  Functions for creating plots and image files in a unified way     regardless of output format (EPS, PDF, PNG, SVG, TIFF, WMF, etc.). Default     device options as well as scales and aspect ratios are controlled in a uniform     way across all device types. Switching output format requires minimal changes     in code. This package is ideal for large-scale batch processing, because it     will never leave open graphics devices or incomplete image files behind, even on     errors or user interrupts.
}

\section{To get started}{
  \itemize{
   \item Vignette '\href{../doc/index.html}{R.devices overview}'
   \item \code{\link{toEPS}}(), \code{\link{toPDF}}(), \code{\link{toPNG}}(), ... - evaluate
         graphics code and atomically save plot to a figure file.
  }
}

\section{How to cite this package}{
  To cite this package, please use:
\preformatted{
@Manual{,
  title = {R.devices: Unified Handling of Graphics Devices},
  author = {Henrik Bengtsson},
  year = {2022},
  note = {R package version 2.17.1},
  url = {https://github.com/HenrikBengtsson/R.devices},
}
}
}

\author{
 Henrik Bengtsson [aut, cre, cph].
}

\section{License}{
 LGPL (>= 2.1).
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recordedplot-methods.R
\name{architecture}
\alias{architecture}
\alias{as.architecture}
\title{Get the architecture of an object or coerce it into another}
\usage{
architecture(x, ...)

as.architecture(
  x,
  ostype = .Platform$OS.type,
  arch = R.version$arch,
  ptrsize = .Machine$sizeof.pointer,
  endian = .Platform$endian,
  ...
)
}
\arguments{
\item{x}{The object to be coerced.}

\item{...}{(optional) Additional arguments passed to the underlying method.}

\item{ostype}{A character string, e.g. \code{"unix"} and \code{"windows"}.}

\item{arch}{A character string, e.g. \code{"i386"}, \code{"i686"} and \code{"x86_64"}.}

\item{ptrsize}{The target pointer size - either \code{4L} or \code{8L}.}

\item{endian}{The target endianess - either \code{"little"} or \code{"big"}.}
}
\value{
\code{architecture()} returns a named list with
    character element \code{ostype} and \code{arch},
    integer element \code{ptrsize}, and character element \code{endian}.
    These elements take a missing values if they could not be inferred.

\code{as.architecture()} returns a coerced version of \code{x}.
    If no coercion was needed, then \code{x} itself is returned.
}
\description{
Get the architecture of an object or coerce it into another
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  DevEvalProduct.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{DevEvalFileProduct}
\docType{class}
\alias{DevEvalFileProduct}


\title{The DevEvalFileProduct class}

\description{
 Package:  R.devices \cr
\bold{Class DevEvalFileProduct}\cr

\code{character}\cr
\code{~~|}\cr
\code{~~+--}\code{\link[R.oo]{BasicObject}}\cr
\code{~~~~~~~|}\cr
\code{~~~~~~~+--}\code{\link[R.devices]{DevEvalProduct}}\cr
\code{~~~~~~~~~~~~|}\cr
\code{~~~~~~~~~~~~+--}\code{DevEvalFileProduct}\cr

\bold{Directly known subclasses:}\cr
\cr

public class \bold{DevEvalFileProduct}\cr
extends \link[R.devices]{DevEvalProduct}\cr



 A DevEvalFileProduct is a \code{\link{DevEvalProduct}} referring to a image
 file created by \code{\link{devEval}}().
}

\usage{
DevEvalFileProduct(filename=NULL, path=NULL, ...)
}

\arguments{
  \item{filename, path}{The filename and the optional path of the image
   file product.}
  \item{...}{Additional arguments passed to \code{\link{DevEvalProduct}}.}
}

\section{Fields}{
 The following (virtual; calculate on-the-fly) fields are available:
 \itemize{
  \item \code{pathname}: the (relative) pathname of the image file, e.g. 'figures/foo,a,b.png'.  This can be used to \emph{link to} image files in for instance HTML and Markdown documents.
  \item \code{path}, the (relative) path to the image file, e.g. 'figures/'
  \item \code{filename}: the filename ("basename") of the image file, e.g. 'foo,a,b.png'
  \item \code{fullname}: the fullname (the filename without the filename extension), e.g. 'foo,a,b'.  It is recommended to use this when including images in LaTeX documents that use the \code{\\usepackage{graphicx}} package.
  \item \code{name}: the part of the fullname before the first comma, e.g. 'foo'
  \item \code{tags}: the part of the fullname after the first comma, e.g. 'a,b'
  \item \code{dataURI}: the Base64-encoded Data URI representation of the image, e.g. 'data:image/png;base64,iVBORw0KGgoAAAA...'.  This can be used to \emph{include} ("inlining") image files in for instance self-contained HTML and Markdown documents.
  \item \code{data}: the character content of the image file.  This can be used to \emph{include} ("inlining") WebGL HTML-based image files in for instance self-contained HTML and Markdown documents.
 }
}

\seealso{
  In order to retrieve the Data URI, the \pkg{base64enc} package
  must be installed.
}

\author{Henrik Bengtsson}


\keyword{classes}
\keyword{internal}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  toNNN.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 \name{toNNN}
\alias{toNNN}

 \alias{toBMP}
 \alias{toEMF}
 \alias{toEPS}
 \alias{toPDF}
 %% \alias{toCairoPNG}
 \alias{toPNG}
 \alias{toSVG}
 \alias{toTIFF}
 \alias{toWMF}
 \alias{toFavicon}
 \alias{toNullDev}
 \alias{asDataURI}

 % Interactive/screen devices
 \alias{toDefault}
 \alias{toQuartz}
 \alias{toWindows}
 \alias{toX11}
 \alias{toCairoWin}
 \alias{toCairoX11}
 %% \alias{toJavaGD}
\alias{toRStudioGD}

 \title{Methods for creating image files of a specific format}

 \description{
  Methods for creating image files of a specific format.
 }

 \usage{
   toBMP(name, ...)
   toPDF(name, ...)
   toPNG(name, ...)
   toSVG(name, ...)
   toTIFF(name, ...)
   toEMF(name, ..., ext="emf")
   toWMF(name, ..., ext="wmf")

   toFavicon(..., name="favicon", ext="png",
             field=getDevOption("favicon", "field", default="htmlscript"))

   % Interactive/screen devices
   toDefault(name, ...)
   toQuartz(name, ...)
   toX11(name, ...)
   toWindows(name, ...)

   % Interactive/screen devices from other packages
   toCairoWin(name, ...)
   toCairoX11(name, ...)
   %% toJavaGD(name, ...)

   % Interactive/screen device for RStudio
   toRStudioGD(name, ..., .allowUnknownArgs = TRUE)
 }

 \arguments{
   \item{name}{A \code{\link[base]{character}} string specifying the name of the image file.}
   \item{..., .allowUnknownArgs}{Additional arguments passed to \code{\link{devEval}}(), e.g.
      \code{tags} and \code{aspectRatio}.}
   \item{ext, field}{Passed to \code{\link{devEval}}().}
 }

 \value{
   Returns by default the \code{\link{DevEvalProduct}}.
   For \code{toFavicon()} the default return value is a \code{\link[base]{character}} string.
 }

 \section{Windows Metafile Format}{
   Both \code{toEMF()} and \code{toWMF()} use the exact same graphics
   device (\code{win.metafile()}) and settings.  They only differ by
   filename extension.  The \code{win.metafile()} device function exists
   on Windows only; see the \pkg{grDevices} package for more details.
 }

 \author{Henrik Bengtsson}

 \seealso{
   These functions are wrappers for \code{\link{devEval}}().
   See \code{\link{devOptions}}() to change the default dimensions for
   a specific device type.
 }



\keyword{documentation}
\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  deviceUtils.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devIsInteractive}
\alias{devIsInteractive}


\title{Checks whether a device type is interactive or not}

\description{
 Checks whether a device type is interactive or not.
}

\usage{
devIsInteractive(types, ...)
}

\arguments{
  \item{types}{A \code{\link[base]{character}} \code{\link[base]{vector}}.}
  \item{...}{Not used.}
}

\value{
  Returns a \code{\link[base]{logical}} \code{\link[base]{vector}} with \code{\link[base:logical]{TRUE}} if the device type is interactive,
  otherwise \code{\link[base:logical]{FALSE}}.
}

\author{Henrik Bengtsson}

\seealso{
  Internally, \code{\link[grDevices]{deviceIsInteractive}} is used.
}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  withPar.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{withPar}
\alias{withPar}


\title{Evaluate an R expression with graphical parameters set temporarily}

\description{
 Evaluate an R expression with graphical parameters set temporarily.
}

\usage{
withPar(expr, ..., args=list(), envir=parent.frame())
}

\arguments{
  \item{expr}{The R expression to be evaluated.}
  \item{...}{Named options to be used.}
  \item{args}{(optional) Additional named options specified as a named \code{\link[base]{list}}.}
  \item{envir}{The \code{\link[base]{environment}} in which the expression should be evaluated.}
}

\value{
 Returns the results of the expression evaluated.
}

\details{
  Upon exit (also on errors), this function will reset \emph{all}
  (modifiable) graphical parameters to the state of options available
  upon entry.  This means any parameters \emph{modified} from evaluating
  \code{expr} will also be undone upon exit.
}

\author{Henrik Bengtsson}

\examples{
withPar({
  layout(1:4)

  withPar({
    plot(1:10)
    plot(10:1)
  }, pch=4)

  withPar({
    plot(1:10)
    plot(10:1)
  }, pch=0, bg="yellow")
}, mar=c(2,2,1,1))
}

\seealso{
  Internally, \code{\link[base]{eval}}() is used to evaluate the expression,
  and \code{\link[graphics]{par}} to set graphical parameters.
}



\keyword{IO}
\keyword{programming}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  DevEvalProduct.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{DevEvalProduct}
\docType{class}
\alias{DevEvalProduct}


\title{The DevEvalProduct class}

\description{
 Package:  R.devices \cr
\bold{Class DevEvalProduct}\cr

\code{character}\cr
\code{~~|}\cr
\code{~~+--}\code{\link[R.oo]{BasicObject}}\cr
\code{~~~~~~~|}\cr
\code{~~~~~~~+--}\code{DevEvalProduct}\cr

\bold{Directly known subclasses:}\cr
\link[R.devices]{DevEvalFileProduct}\cr

public class \bold{DevEvalProduct}\cr
extends \link[R.oo]{BasicObject}\cr



 A DevEvalProduct represents a handle to the "product" returned by
 \code{\link{devEval}}().
}

\usage{
DevEvalProduct(name=NULL, tags=NULL, type=NULL, ...)
}

\arguments{
  \item{name, tags}{The name and optional tags of the product.}
  \item{type}{The device type.}
  \item{...}{Not used.}
}

\section{Fields}{
 The following (virtual; calculate on-the-fly) fields are available:
 \itemize{
  \item \code{fullname}: the fullname of an image, e.g. 'foo,a,b'
  \item \code{name}: the part of the fullname before the first comma, e.g. 'foo'
  \item \code{tags}: the part of the fullname after the first comma, e.g. 'a,b'
 }
}

\author{Henrik Bengtsson}


\keyword{classes}
\keyword{internal}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  deviceUtils.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devIsOpen}
\alias{devIsOpen}


\title{Checks if zero or more devices are open or not}

\description{
 Checks if zero or more devices are open or not.
}

\usage{
devIsOpen(which=dev.cur(), ...)
}

\arguments{
  \item{which}{An index (\code{\link[base]{numeric}}) \code{\link[base]{vector}} or a label (\code{\link[base]{character}}) \code{\link[base]{vector}}.}
  \item{...}{Not used.}
}

\value{
  Returns a named \code{\link[base]{logical}} \code{\link[base]{vector}} with \code{\link[base:logical]{TRUE}} if a device is open,
  otherwise \code{\link[base:logical]{FALSE}}.
}

\examples{
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Use devices for conditional processing of code. 
# Close devices to rerun code.
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
cat("Currently opened device:\n")
print(devList())

# Alt A: Use device index counter (starting with the 16:th)
fig <- 15
if (!devIsOpen(fig <- fig + 1)) {
  devSet(fig)
  cat("Figure", fig, "\n")
  plot(1:10)
}
cat("Currently opened device:\n")
print(devList())

if (!devIsOpen(fig <- fig + 1)) {
  devSet(fig)
  cat("Figure", fig, "\n")
  plot(1:10)
}
cat("Currently opened device:\n")
print(devList())

# Alt B: Use device labels
if (!devIsOpen(label <- "part 1")) {
  devSet(label)
  cat("Part 1\n")
  plot(1:10)
}
cat("Currently opened device:\n")
print(devList())


if (!devIsOpen(label <- "part 2")) {
  devSet(label)
  cat("Part 2\n")
  plot(1:10)
}
cat("Currently opened device:\n")
print(devList())
}

\author{Henrik Bengtsson}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  deviceUtils.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{devSetLabel}
\alias{devSetLabel}


\title{Sets the label of a device}

\description{
 Sets the label of a device.
}

\usage{
devSetLabel(which=dev.cur(), label, ...)
}

\arguments{
  \item{which}{An index (\code{\link[base]{numeric}}) or a label (\code{\link[base]{character}}).}
  \item{label}{A \code{\link[base]{character}} string.}
  \item{...}{Not used.}
}

\value{
  Returns nothing.
}

\author{Henrik Bengtsson}

\seealso{
  \code{\link{devGetLabel}}() and \code{\link{devList}}().
}



\keyword{device}
\keyword{utilities}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  jpeg2.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{jpeg2}
\alias{jpeg2}


\title{A JPEG device for Bitmap Files via GhostScript}

\description{
  A JPEG device for Bitmap Files via GhostScript.
}

\usage{
jpeg2(filename, width=480, height=480, res=144, type="jpeg", ...)
}

\arguments{
 \item{filename}{The name of the file to be produced.}
 \item{width, height}{The width and height (in pixels) of the result image.}
 \item{res}{The resolution of the image.}
 \item{type}{The output type. See \code{\link[grDevices]{dev2bitmap}} for details.}
 \item{...}{Additional arguments passed to \code{bitmap()}.}
}

\value{
  Returns nothing.
}

\author{Henrik Bengtsson}



\keyword{device}
\keyword{internal}

H. Bengtsson, \emph{The R.oo package - Object-Oriented Programming with References Using Standard R Code}, In Kurt Hornik, Friedrich Leisch and Achim Zeileis, editors, Proceedings of the 3rd International Workshop on Distributed Statistical Computing (DSC 2003), March 20-22, Vienna, Austria. \url{http://www.ci.tuwien.ac.at/Conferences/DSC-2003/Proceedings/}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  DevEvalProduct.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{getType.DevEvalProduct}
\alias{getType.DevEvalProduct}
\alias{DevEvalProduct.getType}
\alias{getType,DevEvalProduct-method}
\alias{getType}

\title{Gets the type}

\description{
 Gets the type.
}

\usage{
\method{getType}{DevEvalProduct}(this, ...)
}

\arguments{
  \item{...}{Not used.}
}

\value{
 Returns a \code{\link[base]{character}} string or \code{\link[base]{NULL}}.
}

\author{Henrik Bengtsson}

\seealso{
  For more information see \code{\link{DevEvalProduct}}.
}
\keyword{internal}
\keyword{methods}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do not modify this file since it was automatically generated from:
% 
%  favicon.R
% 
% by the Rdoc compiler part of the R.oo package.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\name{favicon}
\alias{favicon}


\title{Favicon graphics device}

\description{
 Device driver for PNG favicons (Favorite icon) also known as
 shortcut icon, Web site icon, tab icon or bookmark icon.
 This driver is the same as the png driver where some arguments
 have different default values.
}

\usage{
favicon(filename="favicon.png", width=32L, height=width, bg="transparent",
  par=list(mar = c(0, 0, 0, 0)), ...)
}

\arguments{
  \item{file}{Default file name (pattern).}
  \item{width, height}{The width and height of the figure.}
  \item{par}{A named \code{\link[base]{list}} of graphical parameters to use.}
  \item{...}{Other arguments accepted by \code{png()}.}
}

\value{
  A plot device is opened; nothing is returned.
}

\examples{\dontrun{
  favicon(width=32L)

  # is identical to

  suppressWarnings({
    png("favicon.png", width=32L, height=32L, bg="transparent",
                       par=list(mar=c(0,0,0,0)))
  })
}}

\keyword{device}

\seealso{
  Internally, \code{\link[grDevices]{png}} is used.
  %% Add HTML Favicon script
  %% toFavicon(plot(1, col="blue", bg="yellow",
  %%           pch=21, cex=4, lwd=4, axes=FALSE))
  \if{html}{\out{
   <script>
    var link = document.createElement('link');
    link.rel = 'icon';
    link.href = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAMAAABEpIrGAAAADFBMVEX9/v0AAP/9/v3//wBEQjoBAAAABHRSTlP//wD//gy7CwAAAGJJREFUOI3N0rESwCAIA9Ag///PXdoiBk0HhmbNO49DMETQCexNCSyFgdlGoO5DYOr9ThLgPosA7osIQP0sHuDOog8UI/ALa988wzdwXJRctf4s+d36YPTJ6aMd8ux3+QO4ABTtB85yDAh9AAAAAElFTkSuQmCC";
    document.getElementsByTagName('head')[0].appendChild(link);
   </script>
  }}
}

\author{Henrik Bengtsson}



\keyword{device}
\keyword{internal}
